// Bacillus inherits from Gecko's Fission2D to specialize into my own stuff 
// Copyright (C) 1997-1998 Jan Kreft <Kreft@cardiff.ac.uk>
#import "random.h"
#import "Critter.h"
#import "CritterZoo.h"
#import "Bacillus.h"
#import "LandRoot.h"
#import "GeckoControlSwarm.h"
#import "Stat.h"
#import "ResourceQuad.h"
#import "BacillusDiffuser.h"
#import <stdlib.h>
#import <math.h>
#import <swarmobject.h>
#import <simtools.h>

@implementation Bacillus

// global vars valid for all Bacilli
static int printSteps;
static ResQuant maxAmount;
static int varUptakeSteps;
static BOOL wantToStop;

Bacillus* recycleBacilluss = 0;

-recycle
{
    [species tallyDeath: self];

    hook = recycleBacilluss;
    recycleBacilluss = self;
    [species decrementCensus];
    return nil;
}

-recycleClear
{
    Bacillus* togo;
    while (recycleBacilluss) {
	togo = recycleBacilluss;
	recycleBacilluss = recycleBacilluss->hook;
	[togo->energy drop];
	[togo drop];
    }
    return self;
}

-reuse
{
    Bacillus* result = recycleBacilluss;
    recycleBacilluss = recycleBacilluss->hook;
    result->hook    = nil;
    return result;
}

+create 
{
    id result;

    if (recycleBacilluss) {
	result = [recycleBacilluss reuse];
    } else {
	result = [Bacillus createBegin: globalZone];
	[result allocEnergy];
    }
    [result init];

    return result;					
}

-copymom {
    // create an agent matching mom in all heritable respects

    Bacillus* baby;
    Quad* newQuad;

    baby = [Bacillus create];
    /*
    baby-> = ;
    */
    baby->species   = species;
    // copies conversion matrix, not store levels, needed though reset later   
    [energy copyTo: baby->energy]; 
    baby->color     = color;

    [species incrementCensus];
    [species tallyBirth: self];

    baby->stepSize = stepSize;

    baby->breedsize   = breedsize;
    baby->kidsize     = kidsize; 
    baby->minsize     = minsize;
    baby->scaleradius = scaleradius;
    baby->scalev      = scalev;
    baby->shoveRadius = shoveRadius;

    // mutate params here if we also want diverse grandmas
    // vary hereditary params, mom and baby equal

    baby->maxUptakeRate  = maxUptakeRate;
    baby->kmUptake       = kmUptake;
    baby->kmOM           = kmOM;
    baby->delayTime      = delayTime;
    baby->convertBack    = convertBack;
    baby->yield          = yield;
    baby->maintenanceR   = maintenanceR;
    baby->density        = density;
    baby->committed1  = committed1; 
    baby->delayCount1 = delayCount1; 
    baby->committed2  = committed2; 
    baby->delayCount2 = delayCount2; 
    baby->committed3  = committed3; 
    baby->delayCount3 = delayCount3; 
    baby->committed4  = committed4; 
    baby->delayCount4 = delayCount4; 

    newQuad = [Quad create];
    [newQuad setData: baby Type: critter];
    [newQuad setSibling: nil];
    [newQuad setChild: nil];
    [newQuad setParent: nil];
    [baby setQuad: newQuad];

    return baby;
} // end copymom

-liveGrandma: (Species*) s {
  // initialize/reinit a default Bacillus which is as empty as possible

  // Conversions are done by -metabolize every step
  // Rate means: convert "Rate:" amount of A to B per timestep
  [energy addConvertFrom: sub To: cellMatter Rate: 0.0];
  [energy addConvertFrom: sub To: waste Rate: 0.0];
  [energy addConvertFrom: cellMatter To: waste Rate: 0.0];

  species = s;
  [self setColor: [species getColor]];

  committed1  = 0; 
  delayCount1 = -1; 
  committed2  = 0; 
  delayCount2 = -1; 
  committed3  = 0; 
  delayCount3 = -1; 
  committed4  = 0; 
  delayCount4 = -1; 

  [self addSpeciesParams];

  return self;
} // end liveGrandma

-setX: (GCoord) cx Y: (GCoord) cy Radius: (GCoord) r {
  // this is called in GeckoModelSwarm.m (-hangCritters: (Critter*)
  // grandma Num: (int) num) to make the very first Bacillus (Grandma)
  // with r being grandma->kidsize (kidvol = breedsize)

  static int totNumRuns = 1;
  static BOOL firstTime = 1;

  if (firstTime) {
    totNumRuns = [geckoControl getGeckoRuns];
    firstTime = 0;
  }

    x = cx;
    y = cy;
    radius = r;

    /////////////////////// random grandma switches \\\\\\\\\\\\\\\\\\\\


    // random size for grandma, but must be > minsize, don't change
    // starting point r
    /*
    do {
      radius = [self varyDValue: r bySigma: 0.4];
    } while (radius <= minsize);
    */

    [energy setStores: cellMatter 
            Amount: density * radius2volume(radius, scaleradius, scalev)];

    // this is a good place for initializations since also called at
    // start of second run and only called at start of runs

    printSteps = [geckoControl getStatCycle];
    maxAmount = 0.5;
    varUptakeSteps = 10000;
    wantToStop = 0;
    lastDryMass = [self getDryMass];
    [self update];

    [self getMcr: myquad->mcr];
    printLog("numRun %d\n", 1 + totNumRuns - [geckoControl getGeckoRuns]);
    return self;
}

-setBreedSize: (GCoord) bs {

    ResQuant kidvol, minvol;
    
    breedsize = bs;
    
    // Note that breedsize has the meaning of Volume in contrast to
    // standard Gecko where all sizes are radii (see Critter.h)

    // Note, this only applies to the mechanistic Donachie model:
    // (2 * breedsize) is the trigger volume, not the actual volume upon
    // reproduction due to delayTime, hence, kidvol is not the actual
    // value either

    kidvol   = breedsize; // since trigger vol = 2 * breedsize
    kidsize  = volume2radius(kidvol, scaleradius, scalev);
    minvol   = 0.2 * breedsize; // independent of kidvol which is variable
    minsize  = volume2radius(minvol, scaleradius, scalev);

    return self;
}
/*
-set: (GRate)  { = ; return self;}
*/
-setStepSize: (GRate) ss {stepSize = ss; return self;}
-setYield: (GRate) yd {yield = yd; return self;}
-setMaintenanceR: (GRate) m {maintenanceR = m * stepSize; return self;}
-setDensity: (GRate) d {density = d; return self;}
-setMaxUptakeRate: (GRate) mur {maxUptakeRate = mur * stepSize; return self;}
-setKmUptake: (GRate) k {kmUptake = k; return self;}
-setKmOM: (GRate) j {kmOM = j; return self;}
-setDelayTime: (int) dt {delayTime = (int)ceil(dt / stepSize); return self;}

-addSpeciesParams {
  /* Template:
  [species addParam: "" Vary: 0 DValue:  
          Message: M(set:) Owner: self];
  */

  [species addParam: "StepSize" Vary: 0 DValue: stepSize 
          Message: M(setStepSize:) Owner: self];

    [species addParam: "ConvertBack" Vary: 0 DValue: convertBack
 	Message: M(setConvertBack:) Owner: self];

    [species addParam: "BreedSize" Vary: 0 DValue: breedsize
	Message: M(setBreedSize:) Owner: self];
    [species addParam: "ShoveRadius" Vary: 0 DValue: shoveRadius
	Message: M(setShoveRadius:) Owner: self];
    [species addParam: "ScaleR" Vary: 0 DValue: scaleradius 
        Message: M(setScaleR:) Owner: self];
    [species addParam: "ScaleV" Vary: 0 DValue: scalev 
        Message: M(setScaleV:) Owner: self];

  [species addParam: "Yield" Vary: 0 DValue: yield
          Message: M(setYield:) Owner: self];
  [species addParam: "MaintenanceR" Vary: 0 DValue: maintenanceR 
          Message: M(setMaintenanceR:) Owner: self];
  [species addParam: "Density" Vary: 0 DValue: density
          Message: M(setDensity:) Owner: self];

    [species addParam: "MaxUptakeRate" Vary: 0 DValue: maxUptakeRate
	Message: M(setMaxUptakeRate:) Owner: self];
    [species addParam: "KmUptake" Vary: 0 DValue: kmUptake
	Message: M(setKmUptake:) Owner: self];
    [species addParam: "KmOM" Vary: 0 DValue: kmOM
	Message: M(setKmOM:) Owner: self];

    [species addParam: "DelayTime" Vary: 0 IValue: delayTime
	Message: M(setDelayTime:) Owner: self];

    return self;   
}

-(int) deviateFromIParam: (char*) ThisIParam bySigma: (double) sigma {
  // changing a value from the default, no change of the default itself!
  // the "I" method varies ints
  Param* param = [species getParam: ThisIParam];
  int result;
  double deviation;
  double cv = sigma; // coefficient of variation = relative standard deviation

  do {
    // restrict change to a range of 2 * relative standard deviation
    do deviation = GaussRand();
    while ((deviation >= 2.0) || (deviation <= -2.0));
    result = (int) (0.5 + [param getPartIValue] * (1.0 + cv * deviation));
  } while (!samesign(result, [param getPartIValue]));

  return result;   
}

-(double) deviateFromDParam: (char*) ThisDParam bySigma: (double) sigma {
  // changing a value from the default, no change of the default itself!
  // the "D" method varies doubles
  Param* param = [species getParam: ThisDParam];
  double result;
  double deviation;
  double cv = sigma; // coefficient of variation = relative standard deviation

  do {
    // restrict change to a range of 2 * relative standard deviation
    do deviation = GaussRand();
    while ((deviation >= 2.0) || (deviation <= -2.0));
    result = [param getPartDValue] * (1.0 + cv * deviation);
  } while (!samesign(result, [param getPartDValue]));

  return result;   
}

-(int) varyIValue: (int) iv bySigma: (double) sigma {
  // change the current value, new value may be inherited or not
  // the "I" method varies ints
  int val = iv;
  int result;
  double deviation;
  double cv = sigma; // coefficient of variation = relative standard deviation

  do {
    // restrict change to a range of 2 * relative standard deviation
    do deviation = GaussRand();
    while ((deviation >= 2.0) || (deviation <= -2.0));
    result = (int) (0.5 + val * (1.0 + cv * deviation));
  } while (!samesign(result, val));

  return result;   
}

-(double) varyDValue: (double) dv bySigma: (double) sigma {
  // change the current value, new value may be inherited or not
  // the "D" method varies doubles
  double val = dv;
  double result;
  double deviation;
  double cv = sigma; // coefficient of variation = relative standard deviation

  do {
    // restrict change to a range of 2 * relative standard deviation
    do deviation = GaussRand();
    while ((deviation >= 2.0) || (deviation <= -2.0));
    result = val * (1.0 + cv * deviation);
  } while (!samesign(result, val));

  return result;   
}

-metabolize {
    // Override from Critter to continually assimilate from compartment to
    // compartment of stores, as well as assimilate gullet into stores.
    // Cannot use Metab-metabolize, because I've altered the meaning of
    // the convert matrix--not convert from gullet to stores once, but 
    // convert from stores to stores, continually.
	
    int i, j;
    ResList* debits;
    ResQuant debit;

    // this verbatim from Critter-metabolize, so we don't screw that up.
    //lastr = radius; // save off what size we were...will reset at grow step

    // convert stores -> stores
    debits = [[ResList create] zero];    

    for (i=1; i<=NumRes; i++) {
        for (j=1; j<= NumRes; j++) {
            
            if (energy->convert[i][j] > 0) {
                
 	        // first conversions from store i to store j
 	        debit = energy->convert[i][j];
                energy->stores->list[j] += debit;

 	        // sum all losses from store i due to conversion
                debits->list[i] += debit;
                
	    }
	}
    }

    [energy->stores subtract: debits];
    [debits drop];

    // convert gullet -> stores
    energy->stores->list[1] += energy->gullet->list[1];
    [energy->gullet zero];

    return self;
} // end -metabolize

-setConversionRates {
  GRate maintR, overallR, anaR, cataR; 

  // overall substrate consumption rate = internal substrate mass at start
  overallR = (GRate)[energy getStores: sub];

  // growth independent substrate consumption = maintenance rate
  // as maintR is debited from cellMatter, not from sub, use yield to correct

////////////////////// maintenance switches \\\\\\\\\\\\\\\\\\\\\\\\\\\

  // 1. no maintenance, man, set value in Protocol to zero

  // 2. maintenance depends on surfArea, mas
  //maintR = maintenanceR * surfArea * yield;

  // 3. maintenance depends on dryMass, mad
  maintR = maintenanceR * dryMass * yield;

////////////////////// maintenance switches \\\\\\\\\\\\\\\\\\\\\\\\\\\

  // anabolism based on apparent maximal yield on glucose
  anaR = overallR * yield; 

  // catabolism is the remainder
  cataR = overallR * (1 - yield); 

  // set the conversion rates 
  [energy setConvertFrom: sub To: cellMatter Rate: anaR];
  [energy setConvertFrom: sub To: waste Rate: cataR];
  [energy setConvertFrom: cellMatter To: waste Rate: maintR];
  
  return self;  
}

-uptake {
    // The amount is added to self's gullet by the landscape Quad calling
    // Critter-feedRes. The feeding amount is:
    // feedwinArea * subConcn * amount

    // uptakeRate defined in Fission2D.h
  GRate maxR = maxUptakeRate / varUptakeSteps;
    DiffuseQuad* actualSite;
    ResQuant amount, fed;
    ResQuant feedwinArea;
    static Mcr* feedwin; // equiv-area square
    static BOOL firsttime = 1;
    static int lURound = -1;
    static int lWRound = -1;
    LandRoot* landRoot = [[geckoControl getModelSwarm] getLandRoot];

    if (firsttime) {
        firsttime = 0;
        feedwin = [Mcr create];
    }

    feedwinArea = radius2area(radius);
    [feedwin circleToSquare: x Y: y Radius: radius];

    actualSite = [landRoot firstQuadProducing: 1 Over: feedwin];
    if (!actualSite) {
      printLog("warning, actualSite nil\n"); 
      subConcn = 0;
      wantToStop = 1;
    } else {
      // weighted average of densities of all lattice cells covered by feedwin
      subConcn = [actualSite getDensityArea: feedwin];
      //subConcn = [actualSite getDensityAtX: x Y: y];
    }

    //printLogD("subConcn %g\n", subConcn);
    // subConcn = 0 if feedwin partly outside landscape, would
    // be better to test for this directly

    // this "if" begins the core of uptake
    if (subConcn > 0) { 

///////////// uptake switch \\\\\\\\\\\\\

      // 1. linear growth: fixed uptake rate per cell and per timestep
      //uptakeRate = maxR;

      // 2. Based on surfArea - Michaelis-Menten equation 
      // Since the MaxUptakeRate in the Protocol is based on
      // dryMass and 0.019792 micron^2 per fg dryMass is the surfArea of
      // max cells, we have to divide the protocol maxUptakeRate by this factor
      //uptakeRate = (maxR/0.019792) * surfArea * subConcn / (kmUptake + subConcn);

      // 3. Based on dryMass - Michaelis-Menten equation
      uptakeRate = maxR * dryMass * subConcn / (kmUptake + subConcn);

      // 4. Based on dryMass - Best equation
      //uptakeRate = maxR * dryMass * (subConcn + kmUptake + kmOM) * \
	//(1 - sqrt(1 - ((4 * subConcn * kmOM) / pow((subConcn + kmUptake + kmOM), 2.0) ))) / (2 * kmOM);

///////////// uptake switch \\\\\\\\\\\\\

      amount = uptakeRate / (feedwinArea * subConcn);

      if (amount > maxAmount) maxAmount = amount;

      // amount must be < 1, otherwise timesteps are too large
      if ((amount > 1.0) && (lURound != getCurrentTime())) {
	printLogD("bacillus requests amount > 1.0 = %g ", amount);
	printLogD("at subConcn = %g\n", subConcn);
	printLog("\t(only one warning per round)\n");
	lURound = getCurrentTime();
      }

    } else { // subConcn = 0
      uptakeRate = 0.0;
      amount = 0.0;
      if (lWRound != getCurrentTime()) {
	printLog("subConcn 0, only one warning per round\n");
	lWRound = getCurrentTime();
      }
      // since diffusion lattice is initialized after Bacillus,
      // subConcn is 0 at the first substep of timestep 0,
      // hence I need this test
      if (getCurrentTime() > 0) wantToStop = 1;
    }

    // finaly do the feeding
    fed = [landRoot feedCritter: self Resource: sub
                    Amount: amount
                    Area: feedwin];

    //printLogD("feedCritter Amount: %lf over ", amount);
    //[feedwin print];
    //printLogD(" returns fed = %lf\n", fed);

    //[actualSite printv]; // prints sub Densities over whole lattice
    //[landRoot printv]; // prints stats for landscape (whole lattice)
    //[actualSite printColorThresholds]; // Color table for sub

    return self;
} // end uptake

-(ResQuant) getSubConcn {
  return subConcn;
}

-(ResQuant) getDryMass {
  // multiple compartments: DryMass is sum of all stores

/*
    int i; 
    ResList* contents = [energy getStores];
    ResQuant mass = 0.0; 

    for (i=1; i<= NumRes; i++)
      mass += contents->list[i];
*/

  // simple version: DryMass is cellMatter
  ResQuant mass = [energy getStores: cellMatter]; 

  return mass;
}

-(ResQuant) getVolume {
  // dummy method, this method is called by GeckoModelSwarm to calc "Biomass"
  ResQuant biomass = [self getDryMass]; // or whatever should end up as "Biomass"
  return biomass;
}

-(ResQuant) getVol {
  // volume is sum of stores i (being dry masses) divided by their
  // respective "dry" densities, units g/ml, waste not considered
  // densities of E. coli after Domach et al (1984):
  // cytoplasm 0.26, cell wall 0.55, whole cell 0.29
  // densities of B. subtilis after Jeong et al. (1990):
  // cytoplasm 0.215, cell wall 0.38, whole cell 0.25

  ResQuant volume = [self getDryMass] / density; 
  return volume;
}

-update {
  // set variables once per step

  dryMass = [self getDryMass];
  vol = [self getVol];
  radius = volume2radius(vol, scaleradius, scalev);

///////////////// start surfin' switches \\\\\\\\\\\\\\\

  // 0. (con) surfArea constant => uptake rate per cell constant 

  // 1. surfArea calc based on sphereArea:

  // 1a. (coc) I'm a coccus: surfArea = sphereArea
  //surfArea = 4.0 * radius2area(radius);

  // 1b. (rod) I'm a rod with constant eccentricity of 4.4
  // surfArea = 1.318 * sphereArea, sphereArea = 4 * circleArea
  // 1.318 is ratio surfArea/sphereArea at eccentricity of 4.4
  surfArea = 5.272 * radius2area(radius);

  // 2. (fac) surfArea calc based on dryMass, factor treated as constant:
  // f for E. coli 2.86E6 cm^2/g after Joshi & Palsson (1988)
  // f for B. subtilis 1.515E6 cm^2/g after Jeong et al. (1990)
  // 0.019792 micron^2 per fg dryMass, E. coli growing at 1.23 per hour
  //surfArea = 0.019792 * dryMass;

//////////////////// end surfin' switches \\\\\\\\\\\\\\\

  // growthRate in terms of dry mass, not scaled to time unit of 1 min!
  growthRate = (dryMass - lastDryMass);
  [species tallyGrowth: self Amount: growthRate];

  return self;
}
	
-checkDeath {

    if (radius < minsize) {
        [self die]; 
        return nil;
    } else {
        return self;
    }       
}

-die {
    LandRoot* landRoot = [[geckoControl getModelSwarm] getLandRoot];

    [species tallyLossStarve: self Amount: dryMass]; 

    [landRoot payTaxX: x Y: y Resource: sub
               Amount: dryMass * convertBack];

    // note that recycle, called by delData, does the census
    [myquad delData];

    return nil;
}

-(ResList*) kidshare {

    // find initial stores for kid, multi resources, kids get 1/2 of
    // all internal resources (stores)

  int i; 
  ResList* mom; // mom's stores
  ResList* share;
  
  share = [[ResList create] zero];
  mom = [energy getStores];

  for (i=1; i<= NumRes; i++)
    share->list[i] = mom->list[i]/2.0;

  return share;
}

//////////////////////// Donachie switches \\\\\\\\\\\\\\\\\\\\\\\\\\\\
// descriptive Donachie model
-checkVolume {
  ResQuant breedVol;
  ResQuant gph; // generations per hour
  // specific uptake Rate, values must be consistently from the same timestep, hence lastDryMass
  // -step: setVarUptakeSteps -> uptake loop -> update (changes dryMass) -> checkVolume
  ResQuant specUptakeR = uptakeRate * varUptakeSteps / lastDryMass;

  //gph = specUptakeR * yield * 60 / (stepSize * log(2)); // no maintenance correction
  gph = (specUptakeR - maintenanceR) * yield * 60 / (stepSize * log(2)); // maint. correction
  //printLogD("gph %g\n", gph);

  if (gph > 0) {
    // breedsize in Protocol is (Vu of Donachie = 0.4 fl) * (some factor):
    // see lab book page 147
    breedVol = breedsize * pow(2.0,gph); 
    // if want to make vol measurements imprecise
    breedVol = [self varyDValue: breedVol bySigma: 0.1];
    //printLogD("min\t%g\t", getCurrentTime() * stepSize);
    //printLogD("breedVol\t%g\n", breedVol);
    if (vol >= breedVol) [self reproduce];
  }
  return self;
} // end checkVolume descriptive Donachie model
/*
// mechanistic Donachie model:
-checkVolume {
  // commit the cell to (another) division when any of the trigger
  // volumes (multiples of (2 * breedsize)) have been reached for the first
  // time, note that the cell could shrink afterwards, this does not
  // release the commitment, hence the name

  // committed[1,2,3,4] is a flag, delayCount[1,2,3,4] is the
  // corresponding counter that counts delayTime timesteps before
  // division, "only" 4 counters for now, if more would be needed, a
  // warning is produced

////////////////// eclipse switch \\\\\\\\\\\\\\\\\\\\\\\\\\\\

  // eclipse = minimum interval between commitments
  //int eclipse = 0;
  int eclipse = (int)ceil( log(2) / (maxUptakeRate * yield));

////////////////// eclipse switch \\\\\\\\\\\\\\\\\\\\\\\\\\\\

  ResQuant breedVol = 2 * breedsize;
  // if want to make vol measurements imprecise
  //breedVol = [self varyDValue: breedVol bySigma: 0.1];
  //printLogD("min\t%g\t", getCurrentTime() * stepSize);
  //printLogD("breedVol\t%g\n", breedVol);

  // Part 1: set # of commitments to # of times breedVol was reached
  // for the first time

  if (vol >= breedVol)     { 
    if (committed1 == 0) committed1 = 1; 
  }
  if (vol >= 2 * breedVol) {
    if ((committed2 == 0) && (delayCount1 >= eclipse)) committed2 = 1; 
  }
  if (vol >= 3 * breedVol) { 
    if ((committed3 == 0) && (delayCount2 >= eclipse)) committed3 = 1; 
  }
  if (vol >= 4 * breedVol) { 
    if ((committed4 == 0) && (delayCount3 >= eclipse)) committed4 = 1; 
  }

  if (vol >= 5 * breedVol) {
    printLog("warning, vol >= 5 * breedVol\n");
  }

  // Part 2: inc every counter once every timestep if committed is true

  if (committed1 == 1) delayCount1++; // at timestep of commitment
  if (committed2 == 1) delayCount2++; // delayCountx will be incremented
  if (committed3 == 1) delayCount3++; // from -1 to 0
  if (committed4 == 1) delayCount4++;

  //printLog("delayCount1_2_3_4_before_reset %d %d %d %d\n", \
  //   delayCount1, delayCount2, delayCount3, delayCount4);

  // Part 3: call reproduce if oldest counter has reached delayTime
  // and make the second oldest counter the oldest etc.

  if (delayCount1 >= delayTime) { 
    delayCount1 = delayCount2; 
    committed1 = committed2;

    delayCount2 = delayCount3; 
    committed2 = committed3;

    delayCount3 = delayCount4; 
    committed3 = committed4;

    delayCount4 = -1; 
    committed4 = 0;

    [self reproduce]; 
  }
  return self;
} // end checkVolume mechanistic Donachie
*/
-(BOOL) reproduce {
  // False if didn't.

  Bacillus* baby;
  GCoord distance;
  ResList* kidgets;
  double angle;
  Quad* myRoot;

  angle = 2.0 * M_PI * EvenRand();
    
  // record a birth--equivalent to generation counter, actually, and
  // firstbornon is the beginning of current generation....
  firstbornon = getCurrentTime();
  kidcount++;

  // mutate params here if we don't want diverse grandmas
  // but changes once every cell cycle, baby and mom equal
  //maxUptakeRate = [self varyDValue: maxUptakeRate bySigma: 0.1]; // vary
  maxUptakeRate = [self deviateFromDParam: "MaxUptakeRate" bySigma: 0.1] * stepSize; // deviate
  //printLogD("min\t%g\t", getCurrentTime() * stepSize);
  //printLogD("maxUptakeRate\t%g\n", maxUptakeRate / stepSize);

  // make the kid 
  baby         = [self copymom];
  distance     = [self getChildRadius];
  baby->x      = x + ( distance * cos(angle) );
  baby->y      = y + ( distance * sin(angle) );
  baby->wannax = cos(angle);
  baby->wannay = sin(angle);

  // transfer stores only, vol etc. follows from stores and must not
  // be transfered individually, instead update is called to take care
  kidgets = [self kidshare];
  [energy debit: kidgets];
  baby->energy = [baby->energy setStores:  kidgets];

  // update variables
  lastDryMass = lastDryMass / 2.0; // reset lastDryMass for mom and kid
  baby->lastDryMass = lastDryMass;
  [self update];
  [baby update];
  [self setConversionRates];
  [baby setConversionRates];

  // hang on quad tree
  [baby getMcr: baby->myquad->mcr];
  myRoot = [self getRoot];
  [myRoot addNode: baby->myquad];

  [kidgets drop];
    
  // now I move, too....
  x -= ( distance * cos(angle) );
  y -= ( distance * sin(angle) );
  [myquad growMcr];

  return 1;
} // end reproduce

-setVarUptakeSteps {
  // How many uptake steps do we need?
  // upper bound 10000, lower bound 1, depends on uptake params
  if (maxAmount > 0.8) {
    varUptakeSteps *= 10;
    if (varUptakeSteps > 10000) varUptakeSteps = 10000;
  } else if (maxAmount < 0.09) {
    varUptakeSteps /= 10;
    if (varUptakeSteps < 1) varUptakeSteps = 1;
  }
  return self;  
}

-subStep {
    BacillusDiffuser* actualSite;
    Quad* q;
    LandRoot* landRoot;
    Quad* mobileRoot;
    Mcr* landscape;
    int i;

    landRoot   = [[geckoControl getModelSwarm] getLandRoot];
    mobileRoot = [[geckoControl getModelSwarm] getMobileRoot];
    actualSite = [landRoot firstQuadProducing: 1 Over: [mobileRoot getMcr]];
    landscape  = [actualSite getMcr];
    
    for (i = 0; i < varUptakeSteps; i++) {
        q = [mobileRoot pickWin: landscape Root: mobileRoot 
                        Invisibles: 1];
        do {
            [q->data uptake];
            q = [q pickWin: landscape Root: mobileRoot Invisibles: 1];
        } while ((q)); 
            
        // goes second due to order in model swarm
        [actualSite controllerStep: varUptakeSteps];
    } 

    [actualSite setNewRound: 1];

    return self;  
}

-step {
  // One step in the lifecycle of a Bacillus.
  // Note: returns nil if dies--higher level routine deletes, because
  // higher level routine is traversing the tree.

  static int lastRound = -1;

  // exactly once per time step, we do our substepping loop.
  if (lastRound != (int) getCurrentTime()) {
    lastRound = getCurrentTime();
    [self setVarUptakeSteps];
    maxAmount = 0.0; // reset maxAmount to 0 for this round
    [self subStep]; // uptake/diffusion loop
  }

  // stores -> stores (all conversions among compartments) & gullet -> stores
  [self metabolize]; 
  [energy setStores: waste Amount: 0.0]; // waste is discarded without effect

  // must save off dryMass before update, also in reproduce & setX....
  lastDryMass = dryMass; 
  [self update];
  [self setConversionRates];

  if (![self checkDeath]) // death if radius < minsize
    return nil;

  if (wantToStop == 1) {
    [self printv]; 
    [geckoControl stopRunning]; 
    return self;
  }

  // print after update
  if (getCurrentTime() % printSteps == 0) {
    [self printv]; 
  }

  [self checkVolume]; // are we big enough to trigger division?
  [self move];

  return self;
} // end step

-printv {
  static int lastPRound = -1;
  int now = getCurrentTime();

  //LandRoot* landRoot = [[geckoControl getModelSwarm] getLandRoot];
  //DiffuseQuad* actualSite = [landRoot firstQuadProducing: 1 Over: [myquad getMcr]];

  // if want to print only for one cell per round
  if (lastPRound != now) {
    if (now == 10) [self screenshot]; // at time zero display not yet complete
    if ((now % (printSteps * 50) == 0) || (wantToStop == 1)) {
      [self printSubGradient];
      if (now > 10 ) [self screenshot]; // at time zero display not yet complete
    }

    printLogD("now\t%g\t", now * stepSize); [self printHet];
    //[energy printv];
    //[super println];
    //[super printv]; // super's printv also calls [energy printv]
    ////printLogD(" %g\n", );
    //printLogD("min %g ", now * stepSize);
    //printLogD("breedsize %g\n", breedsize);
    //printLogD("stepSize %g\n", stepSize);
    //printLogD("yield %g ", yield);
    //printLogD("maintenanceR %g ", maintenanceR / stepSize);
    //printLogD("density %g\n", density);
    //printLogD("maxAmount %g ", maxAmount);
    //printLog("varUptakeSteps %d\n", varUptakeSteps);
    //printLogD("maxUptakeRate %g ", maxUptakeRate / stepSize);
    //printLogD("min %g\n", getCurrentTime() * stepSize);
    //printLogD("kmUptake %g\n", kmUptake);
    //printLogD("kmOM %g\n", kmOM);
    //printLog("delayTime %d\n", delayTime);
    //printLogD("subConcn %g\n", subConcn);
    //printLogD("uptakeRate %g ", uptakeRate);
    //printLogD("growthRate %g\n", growthRate);
    //printLogD("dryMass %g ", dryMass);
    //printLogD("vol %g\n", vol);
    //printLogD("surfArea/sphereArea %g\n", surfArea/(4*radius2area(radius)));

    /*printLog("delayCount1_2_3_4_before_checkVolume %d %d %d %d\n", \
	     delayCount1, delayCount2, delayCount3, delayCount4); */

    //[actualSite printv]; // prints sub Densities over whole lattice
    //[landRoot printv]; // prints stats for landscape (whole lattice)
    //[actualSite printColorThresholds]; // Color table for sub

    //printLog("\n");

    lastPRound = now;
  }
  return self;
}

-screenshot {
  // make a screenshot of "Gecko" window as ppm
  [geckoControl->geckoDisplaySwarm->worldDisplay writeFrame];
  return self;
}

-printHet { // print SD of subConcn heterogeneity over all bacs
  int n = 0; // number of bacs
  ResQuant mean; // average subConcn over all bacs
  ResQuant sq; // sum of squared deviations
  ResQuant stdDev; // Standard Deviation of SubConcn: not a sample SD, hence SD(n) not SD(n-1)
  ResQuant sub; // subConcn
  ResQuant subSum = 0.0; // Sum of all subConcn
  ResQuant subSquaredSum = 0.0; // Sum of squared subConcn
  BacillusDiffuser* actualSite;
  Quad* q;
  LandRoot* landRoot;
  Quad* mobileRoot;
  Mcr* landscape;
  landRoot   = [[geckoControl getModelSwarm] getLandRoot];
  mobileRoot = [[geckoControl getModelSwarm] getMobileRoot];
  actualSite = [landRoot firstQuadProducing: 1 Over: [mobileRoot getMcr]];
  landscape  = [actualSite getMcr];
    
  q = [mobileRoot pickWin: landscape Root: mobileRoot 
		  Invisibles: 1];
  do {
    sub = [q->data getSubConcn];
    subSum += sub;
    subSquaredSum += pow(sub, 2.0);
    n++;
    q = [q pickWin: landscape Root: mobileRoot Invisibles: 1];
  } while ((q)); 

  mean = subSum / n;
  sq = subSquaredSum - (pow(subSum, 2.0) / n);
  //printLogD("%g\t", sq);
  if (sq < 1E-9) sq = 0.0; // sq can become negative, discard variance due to numerical imprecision
  stdDev = sqrt(sq / n);
  printLogD("%g", mean);
  printLogD("\t%g", stdDev);
  //printLogD("\t%g", stdDev / mean); // CV
  printLog("\t%d", n);
  printLog("\n");
  return self;  
}

-printHetLattice { // print SD of subConcn heterogeneity over total lattice
  int i, j;
  int num; // number of lattice cells
  ResQuant mean; // average subConcn in lattice
  ResQuant sq = 0.0; // sum of squared deviations
  ResQuant stdDev; // Standard Deviation of SubConcn: not a sample SD, hence SD(n) not SD(n-1)
  ResQuant sub; // subConcn
  LandRoot* landRoot = [[geckoControl getModelSwarm] getLandRoot];
  DiffuseQuad* actualSite = [landRoot firstQuadProducing: 1 Over: [myquad getMcr]];
  Param* cellSizeXPtr = [actualSite->siteDef getParam: "CellSizeX"];
  GCoord cellSizeX = [cellSizeXPtr getPartDValue];
  Param* cellSizeYPtr = [actualSite->siteDef getParam: "CellSizeY"];
  GCoord cellSizeY = [cellSizeYPtr getPartDValue];
  Param* cellsXPtr = [actualSite->siteDef getParam: "CellsX"];
  GCoord cellsX = [cellsXPtr getPartIValue];
  Param* cellsYPtr = [actualSite->siteDef getParam: "CellsY"];
  GCoord cellsY = [cellsYPtr getPartIValue];

  num = cellsX * cellsY;
  mean = [actualSite getReservoir] / (num * cellSizeX * cellSizeY);

  for (i=1; i <= cellsX; i++) {
    for (j=1; j <= cellsY; j++) {	
      sub = [actualSite getDensityCellX: i Y: j];
      sq += pow((sub - mean), 2.0);
    }
  }
  stdDev = sqrt(sq / num);
  printLogD("%g", mean);
  printLogD("\t%g", stdDev);
  printLogD("\t%g", stdDev / mean); // CV
  //printLog("\t%d", num);
  printLog("\n");
  return self;
}

-printSubGradient {
  // subConc matrix -> radial subConcGradient
  // matrix must be square and length uneven
  double now = getCurrentTime() * stepSize;
  int max, mid, dist, distCount;
  int i;
  double sum, growthPot, count;
  LandRoot* landRoot = [[geckoControl getModelSwarm] getLandRoot];
  DiffuseQuad* actualSite = [landRoot firstQuadProducing: 1 Over: [myquad getMcr]];
  Param* cellNumPtr = [actualSite->siteDef getParam: "CellsX"];
  Param* cellSizePtr = [actualSite->siteDef getParam: "CellSizeX"];
  GCoord cellSize = [cellSizePtr getPartDValue];

  // for comparision:
  //[actualSite printv]; // prints sub Densities over whole lattice

  // calc middle of matrix and length of radius
  max = [cellNumPtr getPartIValue];
  mid = (int)ceil(max / 2.0);
  dist = max - mid;
  distCount = 0;

  // print subConcn at center
  sum = [actualSite getDensityCellX: mid Y: mid];
  // note uptake switch: this function is valid for MM-dryMass
  growthPot = ((maxUptakeRate * sum / (kmUptake + sum)) - maintenanceR) * yield * 60 / stepSize;
  //printLog("subConcnGradient:\n");
  printLogD("%g\t", now);
  printLogD("%g\t", distCount * cellSize);
  printLogD("%g\t", sum);
  printLogD("%g\n", growthPot);

  // loop through matrix and inc distance from center, start next to center
  for (distCount = 1; distCount <= dist; distCount++) {
    sum = 0.0;
    count = 0.0;

    // One way to do it: only average the vertical and horizontal "diameter"
    /*
    sum += [actualSite getDensityCellX: mid Y: mid + distCount];
    sum += [actualSite getDensityCellX: mid Y: mid - distCount];
    sum += [actualSite getDensityCellX: mid + distCount Y: mid];
    sum += [actualSite getDensityCellX: mid - distCount Y: mid];
    count += 4;
    */

    // The other way: average the whole concentric "ring"
    // 1. top row
    for (i = mid - distCount; i <= mid + distCount; i++) {
      // edges occur twice in the entire loop
      if ((i == mid - distCount) || (i == mid + distCount)) {
	sum += 0.5 * [actualSite getDensityCellX: i   Y: mid + distCount];
	count += 0.5;	
      }
      else {
	sum += [actualSite getDensityCellX: i   Y: mid + distCount];
	count++;
      }
    }
    // 2. bottom row
    for (i = mid - distCount; i <= mid + distCount; i++) {
      // edges occur twice in the entire loop
      if ((i == mid - distCount) || (i == mid + distCount)) {
	sum += 0.5 * [actualSite getDensityCellX: i   Y: mid - distCount];
	count += 0.5;	
      }
      else {
	sum += [actualSite getDensityCellX: i   Y: mid - distCount];
	count++;
      }
    }
    // 3. left column
    for (i = mid - distCount; i <= mid + distCount; i++) {
      // edges occur twice in the entire loop
      if ((i == mid - distCount) || (i == mid + distCount)) {
	sum += 0.5 * [actualSite getDensityCellX: mid - distCount Y: i];
	count += 0.5;	
      }
      else {
	sum += [actualSite getDensityCellX: mid - distCount Y: i];
	count++;
      }
    }
    // 4. right column
    for (i = mid - distCount; i <= mid + distCount; i++) {
      // edges occur twice in the entire loop
      if ((i == mid - distCount) || (i == mid + distCount)) {
	sum += 0.5 * [actualSite getDensityCellX: mid + distCount Y: i];
	count += 0.5;	
      }
      else {
	sum += [actualSite getDensityCellX: mid + distCount Y: i];
	count++;
      }
    }

  sum = sum / count;
  // note uptake switch: this function is valid for MM-dryMass
  growthPot = ((maxUptakeRate * sum / (kmUptake + sum)) - maintenanceR) * yield * 60 / stepSize;
  printLogD("%g\t", now);
  printLogD("%g\t", distCount * cellSize);
  printLogD("%g\t", sum);
  printLogD("%g\n", growthPot);
  }
  return self;
}


// A rough example of rigging up "observeSpecific".  Observe growth rate,
// by distance from center of site.

// counter 0 = member count
//     sum 0 = growth
//     sum 1 = biomass in this bucket.

-statsFinalize
    // called at end of Species building on grandma
    // let's add one more trace--avg. growth overall
{
    GenericBucket* bucket;
    TraceGraph* statGraph = [species getSpecificGraph];
    SpecificStructure* stats = (SpecificStructure*) [species getSpecificStats];

    bucket = (GenericBucket*) stats->totalBucket;
    [statGraph trackDouble: &(bucket->derived[0]) Key: "Avg. Growth"
               Color: "green"];
    [statGraph setLabel: 0 To: "     dist < 5"];
    [statGraph setLabel: 1 To: " 5 < dist < 10"];
    [statGraph setLabel: 2 To: "10 < dist < 15"];
    [statGraph setLabel: 3 To: "15 < dist < 20"];
    [statGraph setLabel: 4 To: "20 < dist < 25"];
    return self;
}

-calcSpecific: (id) specStats
    // we have one derived stat--avg. growth at this distance range
{
    int i;
    double growsum, bucketbio, avg;
    int members;
    SpecificStructure* s = (SpecificStructure*) specStats;
    double statcycle = [geckoControl getStatCycle];
    
    [s->totalBucket clear];
    s->total = 0;
    s->dtotal = 0;
    for (i=0; i < [s getMaxBucket]; i++) {
        members = [s->bucket[i] getCount: 0];
        growsum = [s->bucket[i] getSum: 0];
        bucketbio = [s->bucket[i] getSum: 1];
        
        // average growth rate = growth per member-origbiomass-timeunit
        if (members > 0)
            avg = growsum / (members * (bucketbio - growsum) * statcycle) ;
        else avg = 0;
        [s->bucket[i] setDerived: 0 Value: avg];
        [s->totalBucket addData: s->bucket[i]];
        s->total  += [s->bucket[i] getTotal];
        s->dtotal += [s->bucket[i] getDoubleTotal];
    }
    members = [s->totalBucket getCount: 0];
    growsum = [s->totalBucket getSum: 0];
    bucketbio = [s->totalBucket getSum: 1];
    if (members > 0)
        avg = growsum / (members * (bucketbio - growsum) * statcycle) ;
    else avg = 0;
    [s->totalBucket setDerived: 0 Value: avg];
    
    return self;
}

-(int) getSpecificNumBuckets {
    return 5;
}

-(int) getSpecificBucket {
    BacillusDiffuser* actualSite;
    LandRoot* landRoot;
    Quad* mobileRoot;
    Mcr* landscape;
    GCoord centerx, centery, dist;
    
    landRoot   = [[geckoControl getModelSwarm] getLandRoot];
    mobileRoot = [[geckoControl getModelSwarm] getMobileRoot];
    actualSite = [landRoot firstQuadProducing: 1 Over: [mobileRoot getMcr]];
    landscape  = [actualSite getMcr];
    
    [landscape getMcrCenter: &centerx Y: &centery];	
    dist = distance(centerx, centery, x, y);
    if (dist < 5)
        return 0;
    else if (dist < 10)
        return 1;
    else if (dist < 15)
        return 2;
    else if (dist < 20)
        return 3;
    else 
        return 4;
}
-(int) getSpecificSums { return 2;}

-(double) getSpecificAmount: (int) anEvent
{
    if (anEvent == SPECIFIC_EVENT_ISALIVE)
        return dryMass;
    else
        return -1;
}

-(int) getSpecificSum: (int) anEvent
{
    if (anEvent == SPECIFIC_EVENT_GREW)
        return 0;
    else if (anEvent == SPECIFIC_EVENT_ISALIVE)
        return 1;
    else
        return -1;
}
-(int) getSpecificCounter: (int) anEvent
{
    if (anEvent == SPECIFIC_EVENT_ISALIVE)
        return 0;
    else
        return -1;
}
@end
