// Original, vfb, 970818

#import "GeckoControlSwarm.h"
#import "ResourceQuad.h"
#import "BacillusDiffuser.h"

@implementation BacillusDiffuser

-init 
{
    [super init];
    controller = nil;
    return self;
}

+create
{
    id result;
    result = [[BacillusDiffuser createBegin: globalZone] createEnd];
    [result init];
    return result;					
}

-copymom
    // self=mom, returns copy
    // this actually allocates all the workspace--grandmas nil
{
    BacillusDiffuser* newbie;
    
    newbie           		= [BacillusDiffuser create];
    newbie->controller		= controller;
    [self copyTo: newbie];
    return newbie;
}

-addSiteParams
{
    [siteDef addParam: "Controller" Vary: 0 SValue: ""
          Message: M(setController:) Owner: self];
    [super addSiteParams];
    return self;
}

-setController: (char*) s
{
    Parameterized* speciesList = [ 
        [[geckoControl getModelSwarm] getSpecies] first];
    Parameterized* def = [speciesList getParameterized: s];
    if ((!def) && (strcmp(s, ""))) {
        geckoError(InvalidCombination,
            "Cannot find BacillusDiffuser species '%s'.\n", s);  
    } else {
        controller = (Species*) def;
    }
    return self;
}

-stepResources 
  // IFF bacilli alive, let them run the show, else super.
{
  if (controller == nil) {
    // no bacilli => 1 step, #substeps = neededTotalSteps, adjust varDiffusionRate
    [super setSubSteps: [self calcNeededTotalSteps]];
    [super calcVarDiffusionRate: [self calcNeededTotalSteps]];
    [super stepResources];
    [super setNewRound: 1];
  } else if ([controller getCensus] == 0) {
      // no bacilli => 1 step, #substeps = neededTotalSteps, adjust varDiffusionRate
      [super setSubSteps: [self calcNeededTotalSteps]];
      [super calcVarDiffusionRate: [self calcNeededTotalSteps]];
      [super stepResources];
      [super setNewRound: 1];
    }
    // otherwise do nothing.
    return self;
}

-controllerStep: (int) us
{
  // Bacilli drive diffusion Steps and SubSteps
  // they call controllerStep #uptakeSteps times
    int save = [super getSubSteps];
    int uptakeSteps = us;
    int addSteps = 1;
    int needed = [self calcNeededTotalSteps];

    if (uptakeSteps >= needed) {
      // one (sub)Step per controllerStep call
      // uptakeSteps # of controllerStep calls
      [super calcVarDiffusionRate: uptakeSteps];
      [super setSubSteps: 1];
      [super stepResources];
    } 
    else {
      addSteps = needed / uptakeSteps;
      [super calcVarDiffusionRate: needed];
      [super setSubSteps: addSteps];
      [super stepResources];
    }

    [super setSubSteps: save];
    return self;
}

@end
