// vfb, redesign, 951221:
// vfb, redesign, 960909:  Species more than just a label.

#import <swarmobject.h>		
#import <activity.h>				  
#import "Base.h"
#import "Mcr.h"
#import "Quad.h"
#import "Metab.h"
#import "Colors.h"
#import "Species.h"

// All living agents in Gecko are Critters.  Please skip on down to
// "step" to see what activities Critters do each round.

@interface Critter: SwarmObject {
  // defining settings
 @public
  Metab* energy; 	  // stores, gullet, conversions, taxes, etc.
  Species* species;	  // basically a label
  GCoord breedsize;	  // breeding radius (reproductive maturity)
  GCoord kidsize;	  // kids' initial radius
  GCoord minsize;	  // die at minsize
  GCoord scaleradius;	  // all radii = scale * (radius from volume)
  GCoord scalev;	  // volume = scalev * (sphere volume)

  // working settings  
  int color;		  // color map index
  GCoord x, y;		  // center
  GCoord radius;	  // current size
  GCoord lastr;	  	  // size before metabolize this step.
  ResQuant wannax, wannay;   // which way I wanna go...interpretation varies
  
  // life history stats
  timeval_t birthday;	  // getCurrentTime() at birth
  timeval_t firstbornon;  // getCurrentTime() at first kid
  unsigned  atecount;	  // meals I've had
  int kidcount;		  // kids I've had

  // structural
  GArea scratchpad;	  // used in feeding or whatever
  Quad* myquad;		  // backpointer to my quadtree record		
  id    hook;		  // for linking them together arbitrarily (recycling)
}

// subroutines of create--public for use by subclasses
-init;				
-allocEnergy;
-recycle;			// add to recycling list -of this type-
-reuse;			        // scrounge recycling list -of this type-
-recycleClear;                  // clear out recycling list

+create;

-copymom;			// self=mom, returns baby

-liveGrandma: (Species*) s;	// initialize/reinit a default plant.
-addSpeciesParams;

-resetBreedSize: (GCoord) bs KidSize: (GCoord) ks MinSize: (GCoord) ms
      TaxRate: (GRate) tr TaxExp: (GRate) te;
-resetConvertFrom: (ResType) rfrom To: (ResType) rto Rate: (GRate) newrate;

// and again in single-argument form...for tcl'ing convenience, vfb 960617.
-setBreedSize: (GCoord) bs;
-setKidSize: (GCoord) ks;
-setMinSize: (GCoord) ms;
-setScaleR: (GCoord) sr;
-setScaleV: (GCoord) v;
-setTaxRate: (GRate) tr;
-setTaxExp: (GRate) te;
-setConvertRate: (GRate) nr;

-setX: (GCoord) cx Y: (GCoord) cy Radius: (GCoord) r;	 
-getX: (GCoord*) cx Y: (GCoord*) cy Radius: (GCoord*) r; 
-getX: (GCoord*) cx Y: (GCoord*) cy;	  		
-(GCoord) getX;
-(GCoord) getY;
-(GCoord) getGX; // same as getX and getY, but unique'd names
-(GCoord) getGY;
-(const char*) getInstanceName; 
-(char*) getName; 

-setColor: (int) colorid;	
-(int) getColor;	

-setQuad: (Quad*) q;		// set quad backpointer
-(Quad*) getQuad;

-(GCoord) getRadius;		// others perceive -me- as radius
-(GCoord) getMoveRadius;
-(GCoord) getChildRadius;	// distance at which to place kid
-(GCoord) getKidRadius;		// radius of the kid
-(GCoord) getMinRadius;		// minimum living size
-(GCoord) getBreedRadius;	// size I reproduce at
-(GCoord) getPerceiveRadius;	// usually = radius
-(GCoord) getScaleR;		// usually = 1
-(GCoord) getScaleV;		// default 1/pi
-(Species*) getSpecies;		

-getDirectionX: (GCoord*) dx Y: (GCoord*) dy;	// unit vector
-(double) getDirectionAngle;	
-setDirectionX: (GCoord) dx Y: (GCoord) dy;
-addDirectionX: (GCoord) dx Y: (GCoord) dy Weight: (double) weight;

-setScratchpad: (GArea) s;			
-(GArea) getScratchpad;	
-(GArea) getArea;		
-(ResQuant) getVolume;		
-(Quad*) getRoot;

// for the proto-Critter, these decisions are based on metabolism
-(BOOL) competesWith: (Critter*) other;	
-(BOOL) attacks: (Critter*) other;	
-(BOOL) eats: (ResType) res;	
-(BOOL) eatList: (ResList*) list;
-(BOOL) makes: (ResType) res;	
-(BOOL) makes: (ResType) res;	
-(BOOL) hiddenExposed: (Mcr*) exposed;	// dummy usually

-(void) getMcr: (Mcr*) m;  // calculate, don't return myquad
-feedRes: (ResType) res Amount: (ResQuant) dinner;
-stealRes: (ResType) res Amount: (ResQuant) dinner;
-(ResQuant) stealAllRes: (ResType) res;

// stat support--eminently overrideable
// statsFinalize is your chance to add any specific stats to the Species, or
// attempt to override the labels on graphs, or whatever.
-statsFinalize;			// called at end of Species building on grandma

-(int) getAgeBucket;
-(int) getDefaultAgeCount;	// 5, if have no brighter ideas...
-(int) getStageBucket;
-(int) getDefaultStages;	// 3, if have no brighter ideas...
-(int) getSizeBucket;
-(int) getDefaultSizes;		// 3, if have no brighter ideas...

// "specific" stats are defined and run by a species family (subclass of 
// critter), rather than the generics provided in the "Species" machinery.
-calcSpecific: (id) specStats;	// calls BucketDistrib calc by default.

// You can also return false if you want Species-observeSpecific to create
// traces for you for each derived in each bucket.
-(BOOL) observeSpecific: (id) specStats; // false--take default.

-(int) getSpecificBucket;	// 0, if have no brighter ideas...
-(int) getSpecificCounter: (int) anEvent; // -1, if not implemented
// getSpecificSum says which summer to add to, getSpecificAmount says
// how much to add to the summer
-(int) getSpecificSum: (int) anEvent; // -1, if not implemented
-(double) getSpecificAmount: (int) anEvent; // 0, if not implemented
-(int) getSpecificNumBuckets;	// 1, if have no brighter ideas...
-(int) getSpecificCounts;	// 1, if have no brighter ideas...
-(int) getSpecificSums;		// 1, if have no brighter ideas...
-(int) getSpecificDeriveds;	// 1, if have no brighter ideas...

// the step cycle activities--major candidates for subclass override
-metabolize;			
-resetRadius;
-(ResList*) kidshare;		// find initial stores for kid
-paytax;
-die;				// always an option....
-(BOOL) reproduce;		// false if didn't -choose- to.
-hunt;				
-move;				
-step;

-print;
-println;
-printv;

@end

