// Original, vfb, 970424.

#import "Base.h"
#import "ResourceQuad.h"
#import "LandRoot.h"
#import "random.h"
#import "Critter.h"
#import "GeckoControlSwarm.h"
#import "Diffuse.h"
#import <simtools.h>
#import <math.h>

@implementation DiffuseQuad 

-init 
{
    [super init];
    type		= diffuse;
    resource            = sunwater;
    replenishDay        = 0;
    initDensity         = 0.0;
    haveInitMcr         = 0;
    initCellLX		= 0;
    initCellLY		= 0;
    initCellUX		= 0;
    initCellUY		= 0;
    cellSizeX 		= 0.0;
    cellSizeY 		= 0.0;
    cellsX    		= 0;
    cellsY    		= 0;
    subSteps  		= 1;
    newRound            = 1;
    diffusionRule 	= 1;
    diffusionRate 	= 1.0;
    varDiffusionRate    = 1.0;
    lossRate      	= 0.0;
    permeableEdge 	= 1;
    externalDensity 	= 0.0;
    numColors 		= 10;
    highColorDensity 	= 1.0;
    lowColorDensity 	= 0.00001;
    
    lattice 		= 0;
    lattice2 		= 0;
    colorDensity 	= 0;
    colorIds 		= 0;
    cellArea 		= 0.0;
    myArea 		= 0.0;
    return self;
}

+create
{
    id result;
    result = [[DiffuseQuad createBegin: globalZone] createEnd];
    [result init];
    return result;					
}

-addSiteParams
{
    [siteDef addParam: "Resource" Vary: 0 IValue: resource
          Message: M(setResource:) Owner: self];
    [siteDef addParam: "ReplenishDay" Vary: 0 IValue: replenishDay
          Message: M(setReplenishDay:) Owner: self];
    [siteDef addParam: "InitDensity" Vary: 0 DValue: initDensity
          Message: M(setInitDensity:) Owner: self];
    [siteDef addParam: "CellSizeX" Vary: 0 DValue: cellSizeX
          Message: M(setCellSizeX:) Owner: self];
    [siteDef addParam: "CellSizeY" Vary: 0 DValue: cellSizeY
          Message: M(setCellSizeY:) Owner: self];
    [siteDef addParam: "CellsX" Vary: 0 IValue: cellsX
          Message: M(setCellsX:) Owner: self];
    [siteDef addParam: "CellsY" Vary: 0 IValue: cellsY
          Message: M(setCellsY:) Owner: self];
    [siteDef addParam: "McrLX" Vary: 0 DValue: mcr->lx
          Message: M(setMcrLX:) Owner: self];
    [siteDef addParam: "McrLY" Vary: 0 DValue: mcr->ly
          Message: M(setMcrLY:) Owner: self];
    [siteDef addParam: "SubSteps" Vary: 0 IValue: subSteps
          Message: M(setSubSteps:) Owner: self];
    [siteDef addParam: "DiffusionRule" Vary: 0 IValue: diffusionRule
          Message: M(setDiffusionRule:) Owner: self];
    [siteDef addParam: "DiffusionRate" Vary: 0 DValue: diffusionRate
          Message: M(setDiffusionRate:) Owner: self];
    [siteDef addParam: "LossRate" Vary: 0 DValue: lossRate
          Message: M(setLossRate:) Owner: self];
    [siteDef addParam: "PermeableEdge" Vary: 0 IValue: permeableEdge
          Message: M(setPermeableEdge:) Owner: self];
    [siteDef addParam: "ExternalDensity" Vary: 0 DValue: externalDensity
          Message: M(setExternalDensity:) Owner: self];
    [siteDef addParam: "NumColors" Vary: 0 IValue: numColors
          Message: M(setNumColors:) Owner: self];
    [siteDef addParam: "HighColorDensity" Vary: 0 DValue: highColorDensity
          Message: M(setHighColorDensity:) Owner: self];
    [siteDef addParam: "LowColorDensity" Vary: 0 DValue: lowColorDensity
          Message: M(setLowColorDensity:) Owner: self];
    
    return self;
}

-landGrandma: (Site*) s
    // one default parameterized DiffuseQuad--relies on defaults in init.
    // Grandma has no lattice--work areas created by copymom.
{
    siteDef = s;
    [self setResource: sunwater]; 
    [self setCellSizeX: 10.0];
    [self setCellSizeY: 10.0];
    [self setCellsX: 10];
    [self setCellsY: 10];

    [self addSiteParams];
    [self setMcrLX: 0 LY: 0
            UX: (cellsX * cellSizeX) 
            UY: (cellsY * cellSizeY)];

    return self;
}

-initWithJunk
    // debug routine--1.0 around the edges, diagonal inside
{
    int i, j;
    
    i=1;
    /*
    for (j=1; j <= cellsY; j++) {
        [self setDensity: 1.0 CellX: i Y: j]; 
    }
    i=cellsX;
    for (j=1; j <= cellsY; j++) {
        [self setDensity: 1.0 CellX: i Y: j];
    }
    j=1;
    for (i=1; i <= cellsX; i++) {
        [self setDensity: 1.0 CellX: i Y: j];
    }
    j=cellsY;
    for (i=1; i <= cellsX; i++) {
        [self setDensity: 1.0 CellX: i Y: j];
    }
    */
    // diagonal gradient
    for (i=1; i<=cellsX; i++)
        for (j=1; j<=cellsY; j++)
            if (i==j) {
                [self setDensity: 1.0/((double)(i)) CellX: i Y: j];
            }
    return self;            
}

-(GRate**) allocLattice
    // lattice[xi][yi]
{
    GRate** l;
    int i, j;
    
    l = (GRate**) [globalZone alloc: (cellsX+2) * sizeof(GRate*)];
    for (i=0; i < cellsX+2; i++)
        l[i] = (GRate*) [globalZone alloc: (cellsY+2) * sizeof(GRate)];
    for (i=0; i<cellsX+2; i++)
        for (j=0; j<cellsY+2; j++)
            l[i][j] = 0.0;
    return l;
}

-initLattice
    // Lattice is 1 bigger than cells in each direction for dummying
    // the environment.
{
    int i, j;

    lattice  = [self allocLattice];
    lattice2 = [self allocLattice];

    // if the edge is permeable, the outsides stay fixed in value.
    if (permeableEdge) {
        for (i=0; i < cellsX + 2; i++) {
            j = 0;
            [self setDensity: externalDensity CellX: i Y: j];
            [self setDensityLattice2: externalDensity CellX: i Y: j];
            j = cellsY + 1;
            [self setDensity: externalDensity CellX: i Y: j];
            [self setDensityLattice2: externalDensity CellX: i Y: j];
        }
        for (j=0; j < cellsY + 2; j++) {
            i = 0;
            [self setDensity: externalDensity CellX: i Y: j];
            [self setDensityLattice2: externalDensity CellX: i Y: j];
            i = cellsX + 1;
            [self setDensity: externalDensity CellX: i Y: j];
            [self setDensityLattice2: externalDensity CellX: i Y: j];
        }
    }
    
    return self;
}

-initColors
    
{
    int i, colorstep;
    GRate delta;
    GRate divisor;

    if (numColors > NumDiffuseColors)
        numColors = NumDiffuseColors; 
    
    // allocate the space
    colorDensity = [[self getZone] alloc: numColors * sizeof(GRate)];
    colorIds 	 = [[self getZone] alloc: numColors * sizeof(int)];
    
    // calculate the thresholds--discussed in .h file
    delta = highColorDensity / lowColorDensity;
    divisor = pow(delta, (1.0 / numColors));
    colorDensity[0] = highColorDensity / divisor;
    for (i = 1; i < numColors - 1; i++)
        colorDensity[i] = colorDensity[i-1] / divisor;
    colorDensity[numColors - 1] = 0.0;
    
    // dole out available colors equably
    colorstep = 1;
    if (numColors < NumDiffuseColors) { 
        colorstep = (int) (NumDiffuseColors / (0.5 + numColors));
        if ((colorstep * numColors) > NumDiffuseColors)
            colorstep--;
    }
    colorIds[0] = HighDiffuseColor;
    for (i = 1; i < numColors - 1; i++)
        colorIds[i] = colorIds[i-1] + colorstep;
    colorIds[numColors - 1] = LowDiffuseColor;
    
    return self;
}

-(int) getColorCellX: (int) xi Y: (int) yi
{
    return [self getColorDensity: [self getDensityCellX: xi Y: yi]];
}

-(int) getColorDensity: (GRate) d
{
    int i;
    
    for (i=0; i < numColors; i++) {
        if (d >= colorDensity[i]) {
            return colorIds[i];
        }
    }
    return 0;    
}

-copymom
    // self=mom, returns copy
    // this actually allocates all the workspace--grandmas nil
{
    DiffuseQuad* newbie;
    newbie           		= [DiffuseQuad create];
    [self copyTo: newbie];
    return newbie;
}

-copyTo: (DiffuseQuad*) newbie 
{
    newbie->siteDef  		= siteDef;
    newbie->resource 		= resource;
    newbie->replenishDay 	= replenishDay;
    newbie->initDensity		= initDensity;
    newbie->haveInitMcr		= haveInitMcr;
    newbie->initCellLX		= initCellLX;	
    newbie->initCellLY		= initCellLY;	
    newbie->initCellUX		= initCellUX;	
    newbie->initCellUY		= initCellUY;	
    newbie->cellSizeX 		= cellSizeX;
    newbie->cellSizeY 		= cellSizeY;
    newbie->cellsX  		= cellsX;
    newbie->cellsY 		= cellsY;	
    newbie->subSteps 		= subSteps;
    newbie->newRound            = newRound;
    newbie->diffusionRule 	= diffusionRule;
    newbie->diffusionRate 	= diffusionRate;
    newbie->varDiffusionRate    = varDiffusionRate;
    newbie->lossRate 		= lossRate;
    newbie->permeableEdge	= permeableEdge;
    newbie->externalDensity 	= externalDensity;
    newbie->numColors 		= numColors;
    newbie->highColorDensity 	= highColorDensity;
    newbie->lowColorDensity 	= lowColorDensity;
    
    // One blank lattice, with space for writing density (double*).
    [newbie initLattice];
    [newbie initColors];
    
    newbie->cellArea 		= (double) cellSizeX * cellSizeY;
    newbie->myArea 	= ((double) (cellsX * cellsY)) * newbie->cellArea;
    
    [newbie setMcrLX: mcr->lx LY: mcr->ly 
            UX: mcr->lx + (cellsX * cellSizeX) 
            UY: mcr->ly + (cellsY * cellSizeY)];
    return self;
}

-(GRate) getDensityCellX: (int) xi Y: (int) yi
    // CELL (xi, yi)
    // always read from lattice
{
    return lattice[xi][yi];
}

-(GRate) getDensityLattice2CellX: (int) xi Y: (int) yi
    // CELL (xi, yi)
    // read from lattice2
{
    return lattice2[xi][yi];
}

-setDensityLattice2:  (GRate) d CellX: (int) xi Y: (int) yi
    // CELL (xi, yi)
    // always write to lattice2
{
    lattice2[xi][yi] = d;
    return self;
}

-setDensity:  (GRate) d CellX: (int) xi Y: (int) yi
    // CELL (xi, yi)
    // write back to lattice 
{
    lattice[xi][yi] = d;
    return self;
}

-addResourceAmount: (ResQuant) amount X: (GCoord) x Y: (GCoord) y
    // AMOUNT, not density
{
    int xi, yi;
    GRate dhere, qhere;
    
    if ([self fromPointX: x PointY: y toCellX: &xi CellY: &yi]) {
        dhere = [self getDensityCellX: xi Y: yi];
        qhere = dhere * cellArea;
        qhere += amount;
        dhere = qhere / cellArea;
        [self setDensity: dhere CellX: xi Y: yi];
    }
    return self;
}

-(Mcr*) getMcrCellX: (int) xi Y: (int) yi
{
    Mcr* cellm;
    
    if ((xi < 1) || (xi > cellsX) || (yi < 1) || (yi > cellsY))
        return 0;
    
    cellm = [Mcr create]; 
    cellm->lx = mcr->lx + ((xi - 1) * cellSizeX);  
    cellm->ly = mcr->ly + ((yi - 1) * cellSizeY);  
    cellm->ux = mcr->lx + (xi * cellSizeX);  
    cellm->uy = mcr->ly + (yi * cellSizeY);  
    return cellm;
}

-(GRate) getDensityAtX: (GCoord) x Y: (GCoord) y
    // point density--lattice2 (truth at last step)
{
    int xi, yi;
    if (![self fromPointX: x PointY: y toCellX: &xi CellY: &yi])
        return 0.0;
    else {
        return [self getDensityLattice2CellX: xi Y: yi];
    }
}

-(GRate) getDensityArea: (Mcr*) m
    // weighted density--area m may cross lattice cells.
{
    GArea mutualArea, wholearea;
    int lxi, lyi, uxi, uyi;	// lattice cells of ll and ur corners
    int i, j;
    GArea densitysum;
    static BOOL firsttime = 1;
    static Mcr* clipm;
    Mcr* cellm;
    
    if (firsttime) {
        clipm = [Mcr create];
        firsttime = 0;
    }
    
    wholearea = [m areaMcr];
    
    // given nothing to do, so get out
    if (wholearea == 0) 
        return 0;
    
    // next case--m contained in one lattice cell?
    // start with already clipped to lattice mcr...
    if (![mcr intersectMcr: m Intersection: clipm]) 
        return 0; // not on the lattice.
        
    if (![self fromPointX: clipm->lx PointY: clipm->ly 
            toCellX: &lxi CellY: &lyi]) {
        return 0; // we -are- confused....
    }
    if (![self fromPointX: clipm->ux PointY: clipm->uy 
            toCellX: &uxi CellY: &uyi]) {
        return 0; // we -are- confused....
    }
    
    // OK, all in one cell--easy out.
    if ((lxi == uxi) && (lyi == uyi))
        return [self getDensityLattice2CellX: lxi Y: lyi];
    
    // OK, only the hard way is left.  Weighted average.
    densitysum = 0; 
    for (i=lxi; i<= uxi; i++) {
        for (j=lyi; j<=uyi; j++) {
            cellm = [self getMcrCellX: i Y: j];
            if ([cellm intersectMcr: m Intersection: clipm]) {
                mutualArea = [clipm areaMcr];
                densitysum += mutualArea * [self getDensityLattice2CellX: i Y: j];
            }
            [cellm drop];
        }
    }
    return (densitysum / wholearea);
}

-(BOOL) fromPointX: (GCoord) x PointY: (GCoord) y 
        toCellX: (int*) xi CellY: (int*) yi
{
    if (![mcr ptInsideX: x Y: y])
        return 0;
    *xi = (int)( (x - mcr->lx) / cellSizeX ) + 1;
    *yi = (int)( (y - mcr->ly) / cellSizeY ) + 1;
    return 1;
}

-(ResQuant) feedCritter: (Critter*) eater 
            Amount: (GRate) amount Area: (Mcr*) m
    // here we interpret "amount" as percentage
    
    // NOTE:  it's the CALLER's responsibility to apply all relevant
    // debits to the amount, including overlapping agents.  This routine
    // simply applies the areas.
{
    int lxi, lyi, uxi, uyi, i, j;
    GArea mutualArea;
    ResQuant meal, total;		
    double dhere, qhere;	// density (per area) and quantity here
    double percent = amount;

    static unsigned lastWarningRound = 0;
    static BOOL firsttime = 1;
    static Mcr* clipm;    
    Mcr* cellm;    
    
    total = 0;

    if (firsttime) {
        firsttime = 0;
        clipm = [Mcr create];
    }
    
    // don't cross?
    if (![mcr intersectMcr: m Intersection: clipm]) 
        return 0;

    // which cell(s)? these -should- get valid answers....
    if (![self fromPointX: clipm->lx PointY: clipm->ly toCellX: &lxi CellY: &lyi])
        return 0;
    if (![self fromPointX: clipm->ux PointY: clipm->uy toCellX: &uxi CellY: &uyi])
        return 0;

    // and do it
    for (i=lxi; i<=uxi; i++) {
        for (j=lyi; j<=uyi; j++) {
            cellm = [self getMcrCellX: i Y: j];
            if ([cellm intersectMcr: m Intersection: clipm]) {
                mutualArea = [clipm areaMcr];
                dhere = [self getDensityLattice2CellX: i Y: j];
                qhere = dhere * mutualArea;
                meal  = dhere * mutualArea * percent;
                if (meal > qhere)
                    meal = qhere;
                
                qhere -= meal;
                total += meal;
                dhere = [self getDensityCellX: i Y: j];
                dhere = ((dhere * cellArea) - meal) / cellArea;
                if (dhere < 0) {
                    if (lastWarningRound < getCurrentTime()) {
                        geckoWarning("agents ate too much, density < 0\n", " ");
                        printLogD("\t density %lf reset to 0", dhere);
                        printLog(" at cell (%d, %d)\n", i, j);
                        printLog("\t(only one warning per round)\n");
                        lastWarningRound = getCurrentTime();
                    }
                    dhere = 0;
                }
                [self setDensity: dhere CellX: i Y: j];
            }
            [cellm drop];
        }
    }
    [eater feedRes: resource Amount: total];
    reservoir -= total;

    fed += total;    
    
    return total;
}

-permDiffuse
    // based on Swarm's "second order approximation (141 4G4 141)" in Diffuse2d.m
    // But I don't understand why they implemented "evaporationRate" as they 
    // did.  Also, according to Jan Kreft, 0.5 is NOT part of the diffusion
    // constant itself, but simply the evening of the squares.
{
    int i, j;
    double delta, d;
    double sum;
    
    for (i=1; i <= cellsX; i++) {
        for (j=1; j <= cellsY; j++) {	
            sum = 0.0;
            
            sum += 4 *    [self getDensityCellX: i   Y: j+1];	// N
            sum += 4 *    [self getDensityCellX: i   Y: j-1];	// S
            sum += 4 *    [self getDensityCellX: i+1 Y: j  ];	// E
            sum += 4 * 	  [self getDensityCellX: i-1 Y: j  ];	// W
            
            sum +=        [self getDensityCellX: i+1 Y: j+1];	// NE
            sum +=        [self getDensityCellX: i+1 Y: j-1];	// SE
            sum +=        [self getDensityCellX: i-1 Y: j-1];	// SW
            sum +=        [self getDensityCellX: i-1 Y: j+1];	// NW
            
            sum -= 20.0 * [self getDensityCellX: i   Y: j  ]; 	// here
            delta = 0.5 * varDiffusionRate * sum / 20.0; // normal diffusion
            //delta = sum / 20.0; // diffusion @ infinite speed

            // now change my value by that smoothing
            d = [self getDensityCellX: i Y: j];  
            d += delta;

            // now do losses--in original, this said "d *= evaporationRate"
            d -= d * lossRate;
  
            [self setDensityLattice2: d CellX: i Y: j];  
        }
    } 
    return self;
}

-cornerDiffuseX: (int) x Y: (int)y
    VertX: (int) vx VertY: (int) vy
    HorizX: (int) hx HorizY: (int) hy
    DiagX: (int) dx DiagY: (int) dy
{
    GRate oldme, new, oldgifted, gift, delta;
    GRate weights = 9.0;
    GRate modrate = 0.5 * varDiffusionRate/weights;
    
    oldme = [self getDensityCellX: x Y: y];
    delta = 0;

    // vertical
    oldgifted = [self getDensityCellX: vx Y: vy];
    gift = modrate * 4.0 * (oldme - oldgifted);
    oldgifted = [self getDensityLattice2CellX: vx Y: vy];
    [self setDensityLattice2: oldgifted + gift CellX: vx Y: vy];
    delta -= gift;
    
    // horizontal
    oldgifted = [self getDensityCellX: hx Y: hy];
    gift = modrate * 4.0 * (oldme - oldgifted);
    oldgifted = [self getDensityLattice2CellX: hx Y: hy];
    [self setDensityLattice2: oldgifted + gift CellX: hx Y: hy];
    delta -= gift;
    
    // diagonal
    oldgifted = [self getDensityCellX: dx Y: dy];
    gift = modrate * (oldme - oldgifted);
    // in this case, need to add to diagonal's latest
    oldgifted = [self getDensityLattice2CellX: dx Y: dy];
    [self setDensityLattice2: oldgifted + gift CellX: dx Y: dy];
    delta -= gift;
    
    // me
    oldme = [self getDensityLattice2CellX: x Y: y];
    new = oldme + delta;
    [self setDensityLattice2: new CellX: x Y: y];
    
    return self;
}

-sideDiffuseX: (int) x Y: (int) y
    Side1X: (int) s1x Side1Y: (int) s1y
    Side2X: (int) s2x Side2Y: (int) s2y
    PerpX:  (int) px  PerpY:  (int) py
    Diag1X: (int) d1x Diag1Y: (int) d1y
    Diag2X: (int) d2x Diag2Y: (int) d2y
{
    GRate oldme, new, oldgifted, gift, delta;
    GRate weights = 14.0;
    GRate modrate = 0.5 * varDiffusionRate/weights;
    
    oldme = [self getDensityCellX: x Y: y];
    delta = 0;

    // side1
    oldgifted = [self getDensityCellX: s1x Y: s1y];
    gift = modrate * 4.0 * (oldme - oldgifted);
    oldgifted = [self getDensityLattice2CellX: s1x Y: s1y];
    [self setDensityLattice2: oldgifted + gift CellX: s1x Y: s1y];
    delta -= gift;
    
    // side2
    oldgifted = [self getDensityCellX: s2x Y: s2y];
    gift = modrate * 4.0 * (oldme - oldgifted);
    oldgifted = [self getDensityLattice2CellX: s2x Y: s2y];
    [self setDensityLattice2: oldgifted + gift CellX: s2x Y: s2y];
    delta -= gift;

    // perp
    oldgifted = [self getDensityCellX: px Y: py];
    gift = modrate * 4.0 * (oldme - oldgifted);
    // sides and diagonal, add to latest
    oldgifted = [self getDensityLattice2CellX: px Y: py];
    [self setDensityLattice2: oldgifted + gift CellX: px Y: py];
    delta -= gift;
    
    // diagonal1
    oldgifted = [self getDensityCellX: d1x Y: d1y];
    gift = modrate * (oldme - oldgifted);
    oldgifted = [self getDensityLattice2CellX: d1x Y: d1y];
    [self setDensityLattice2: oldgifted + gift CellX: d1x Y: d1y];
    delta -= gift;

    // diagonal2
    oldgifted = [self getDensityCellX: d2x Y: d2y];
    gift = modrate * (oldme - oldgifted);
    oldgifted = [self getDensityLattice2CellX: d2x Y: d2y];
    [self setDensityLattice2: oldgifted + gift CellX: d2x Y: d2y];
    delta -= gift;
    
    // me
    oldme = [self getDensityLattice2CellX: x Y: y];
    new = oldme + delta;
    [self setDensityLattice2: new CellX: x Y: y];
    return self;
}

-debitNextToEdge
    // Subtract off all (pos or neg) contributions already made from the
    // edges to the next-to-edges.  Those gifts (losses) are recalculated
    // and replaced by cornerDiffuseX... and sideDiffuseX...
    // Also, restore edges on lattice2 to their pre-permDiffuse levels.
{
    int i, j, iedge, jedge;
    GRate old, new, gifts, newnew;
    GRate modrate = 0.5 * varDiffusionRate/20.0;
    
    if ((cellsX <= 2) || (cellsY <= 2)) {
        // there is no interior 
        return self;
    }
    
    // Strategy: the corners have 5 neighbors on the edge, the internals
    // have 3 neighbors on the edge.  So we treat all as having three,
    // except the last on the row, which has 2 on this round, then 
    // 3 as normal going the next direction around.  We're going around
    // counter-clockwise from cell 2,2 (one in from 1,1, which is one in
    // from the unused false outside 0,0.)  
    
    // across the bottom
    for (i=2; i<=cellsX - 1; i++) { 
        j = 2;
        jedge = 1;
        
        old = [self getDensityCellX: i Y: j];
        new = [self getDensityLattice2CellX: i Y: j];

        iedge = i - 1; 		// diagonal below left
        gifts = modrate * (old - [self getDensityCellX: iedge Y: jedge]);
        iedge = i; 		// below 
        gifts += 4.0 * modrate * (old - [self getDensityCellX: iedge Y: jedge]);

        if (i < (cellsX-1)) {
            iedge = i + 1; 	// below right
            gifts += modrate * (old - [self getDensityCellX: iedge Y: jedge]);
        }
        newnew = new + gifts;
        [self setDensityLattice2: newnew CellX: i Y: j];  
    }
    
    // up the right side
    for (j=2; j<cellsY; j++) { 
        i = cellsX - 1;
        iedge = cellsX;
        
        old = [self getDensityCellX: i Y: j];
        new = [self getDensityLattice2CellX: i Y: j];

        jedge = j - 1; 		// diagonal below right
        gifts = modrate * (old - [self getDensityCellX: iedge Y: jedge]);
        jedge = j; 		// right
        gifts += 4.0 * modrate * (old - [self getDensityCellX: iedge Y: jedge]);

        if (j < (cellsY-1)) {
            jedge = j + 1; 	// above right
            gifts += modrate * (old - [self getDensityCellX: iedge Y: jedge]);
        }
        newnew = new + gifts;
        [self setDensityLattice2: newnew CellX: i Y: j];  
    }

    // across the top
    for (i=cellsX - 1; i>=2; i--) { 
        j = cellsY - 1;
        jedge = cellsY;
        
        old = [self getDensityCellX: i Y: j];
        new = [self getDensityLattice2CellX: i Y: j];

        iedge = i + 1; 		// diagonal above right
        gifts = modrate * (old - [self getDensityCellX: iedge Y: jedge]);
        iedge = i; 		// above
        gifts += 4.0 * modrate * (old - [self getDensityCellX: iedge Y: jedge]);

        if (i > 2) {
            iedge = i - 1; 	// above left
            gifts += modrate * (old - [self getDensityCellX: iedge Y: jedge]);
        }
        newnew = new + gifts;
        [self setDensityLattice2: newnew CellX: i Y: j];  
    }

    // down the left side
    for (j=cellsY-1; j>=2; j--) { 
        i = 2;
        iedge = 1;
        
        old = [self getDensityCellX: i Y: j];
        new = [self getDensityLattice2CellX: i Y: j];

        jedge = j + 1; 		// diagonal above left
        gifts = modrate * (old - [self getDensityCellX: iedge Y: jedge]);
        jedge = j; 		// left
        gifts += 4.0 * modrate * (old - [self getDensityCellX: iedge Y: jedge]);

        if (j > 2) {
            jedge = j - 1; 	// below left
            gifts += modrate * (old - [self getDensityCellX: iedge Y: jedge]);
        }
        newnew = new + gifts;
        [self setDensityLattice2: newnew CellX: i Y: j];  
    }

    // restore edges
    for (i=1; i<=cellsX; i++) {
        j = 1;
        old = [self getDensityCellX: i Y: j];
        [self setDensityLattice2: old CellX: i Y: j];
        j = cellsY;
        old = [self getDensityCellX: i Y: j];
        [self setDensityLattice2: old CellX: i Y: j];
    }
    for (j=1; j<=cellsY; j++) {
        i = 1;
        old = [self getDensityCellX: i Y: j];
        [self setDensityLattice2: old CellX: i Y: j];
        i = cellsX;
        old = [self getDensityCellX: i Y: j];
        [self setDensityLattice2: old CellX: i Y: j];
    }
    
    return self;   
}

-edgyDiffuse 
    // There's a wall instead of a permeable edge--calc edges with less
    // neighbors
{
    int i, j;
    
    // do middle part first--no problems
    [self permDiffuse];
   
    // now for the problems--first, next-to-edge remove edge contributions
    [self debitNextToEdge];

    // corners
    [self cornerDiffuseX: 1               Y: 1 
                   VertX: 1           VertY: 2
                  HorizX: 2          HorizY: 1
                   DiagX: 2           DiagY: 2];
    [self cornerDiffuseX: cellsX 	  Y: 1 
                   VertX: cellsX      VertY: 2
                  HorizX: cellsX - 1 HorizY: 1
                   DiagX: cellsX - 1  DiagY: 2];
    [self cornerDiffuseX: cellsX 	  Y: cellsY  
                   VertX: cellsX      VertY: cellsY - 1
                  HorizX: cellsX - 1 HorizY: cellsY
                   DiagX: cellsX - 1  DiagY: cellsY - 1];
    [self cornerDiffuseX: 1     	  Y: cellsY  
                   VertX: 1           VertY: cellsY - 1
                  HorizX: 2          HorizY: cellsY
                   DiagX: 2           DiagY: cellsY - 1];

    // bottom
    for (i=2; i<=cellsX - 1; i++) {
        j = 1;
        [self sideDiffuseX: i             Y: j
                    Side1X: i-1      Side1Y: j
                    Side2X: i+1      Side2Y: j
                     PerpX: i         PerpY: j+1
                    Diag1X: i-1      Diag1Y: j+1
                    Diag2X: i+1      Diag2Y: j+1];
    }            
    // top
    for (i=2; i<=cellsX - 1; i++) {
        j = cellsY;
        [self sideDiffuseX: i             Y: j
                    Side1X: i-1      Side1Y: j
                    Side2X: i+1      Side2Y: j
                     PerpX: i         PerpY: j-1
                    Diag1X: i-1      Diag1Y: j-1
                    Diag2X: i+1      Diag2Y: j-1];
    }            
    // left
    for (j=2; j<=cellsY - 1; j++) {
        i = 1;
        [self sideDiffuseX: i             Y: j
                    Side1X: i        Side1Y: j+1
                    Side2X: i        Side2Y: j-1
                     PerpX: i+1       PerpY: j
                    Diag1X: i+1      Diag1Y: j-1
                    Diag2X: i+1      Diag2Y: j+1];
    }            
    // right
    for (j=2; j<=cellsY - 1; j++) {
        i = cellsX;
        [self sideDiffuseX: i             Y: j
                    Side1X: i        Side1Y: j+1
                    Side2X: i        Side2Y: j-1
                     PerpX: i-1       PerpY: j
                    Diag1X: i-1      Diag1Y: j-1
                    Diag2X: i-1      Diag2Y: j+1];
    }            
    
    return self;
}

-syncLattice
    // make lattice2 match lattice
    // Lattice is 1 bigger than cells in each direction for dummying
    // the environment.
{
    int i, j;
    GRate d;

    for (i=0; i < cellsX + 2; i++) {
        for (j=0; j < cellsY + 2; j++) {
            d = [self getDensityCellX: i Y: j];
            [self setDensityLattice2: d CellX: i Y: j];
        }
    }
    return self;
}

-subStepResources 
{
    GRate** swap;
    
    // read from lattice, write to lattice2
    // never read from lattice2, so don't need to re-initialize it
    if (permeableEdge)
        [self permDiffuse];
    else
        [self edgyDiffuse];

    // swap freshly written lattice2->current lattice  
    swap     = lattice2;
    lattice2 = lattice;
    lattice  = swap;

    return self;   
}

-addInit
{
    int lx, ly, ux, uy, i, j;
    GRate was;
    
    if (haveInitMcr) {
        lx = initCellLX;
        ly = initCellLY;
        ux = initCellUX;
        uy = initCellUY;
    } else {
        lx = 1;
        ly = 1;
        ux = cellsX;
        uy = cellsY;
    }
    for (i=lx; i <= ux; i++) {
        for (j=ly; j <= uy; j++) {
            was = [self getDensityCellX: i Y: j];
            [self setDensity: was + initDensity CellX: i Y: j];
        }
    }
    return self;         
}

-stepResources 
    // Note:  lattice is the one we're removing resources from during the
    // course of a step, but lattice2 the one we're consulting for the 
    // density at a location.  At the end of this routine, and only at the
    // end of this routine, they must match.  Current reality is lattice,
    // but to keep the game fair, we look for density from lattice2.
{
    int i, j;
    GRate  d;	// density

    if ( (getCurrentTime() == 0) ||
         ((replenishDay > 0) && (getCurrentTime() % replenishDay == 0)) ) {
      if (newRound) {
	[self addInit];
	[self setNewRound: 0];
      }
    }

    startres = 0.0;
    
    for (i=0; i<subSteps; i++)
        [self subStepResources];
    
    // so, how much we got?
    for (i=1; i<=cellsX; i++) {
        for (j=1; j<=cellsY; j++) {
            d = [self getDensityCellX: i Y: j];
            startres += d;
        }
    }
    startres *= cellArea;
    reservoir = startres;
    [self syncLattice];
    return self;
}

// the following exist for the convenience of the Params scheme--not for
// calling directly.
-setCellSizeX: (GCoord) x {cellSizeX = x; return self;}
-setCellSizeY: (GCoord) x {cellSizeY = x; return self;}
-setCellsX: (int) i {cellsX =  i; return self;}
-setCellsY: (int) i {cellsY =  i; return self;}
-(int) getCellsX {return cellsX;}
-(int) getCellsY {return cellsY;}
-(int) getSubSteps { return subSteps; }
-setSubSteps: (int) i {subSteps =  i; return self;}
-(BOOL) setNewRound: (BOOL) f { return newRound = f;}
-setDiffusionRule: (int) i {diffusionRule =  i; return self;}

-setDiffusionRate: (GRate) x {
  GRate stepSize = 1.0;
  Species* species = [[[geckoControl getModelSwarm] getSpecies] first];
  Param* param = [species getParam: "StepSize"];

  if (!strcmp(param->label, "StepSize")) {
    stepSize = [param getPartDValue];
  }

  diffusionRate = x * stepSize / (cellSizeX * cellSizeY); 
  return self;
}

-(int) calcNeededTotalSteps {
  // varDiffusionRate must be <= 1, therefore
  // neededTotalSteps = 10^(order of magnitude of diffusionRate)
  int neededTotalSteps;

  if (diffusionRate > 1.0) {
  neededTotalSteps = (int) pow(10.0, (double) ((int) log10(diffusionRate) + 1));
  } else {
    neededTotalSteps = 1;
  }
  return neededTotalSteps;
}

-calcVarDiffusionRate: (int) ts {
  varDiffusionRate = diffusionRate / ts;
  return self;
}

-setLossRate: (GRate) x {lossRate =  x; return self;}
-setPermeableEdge: (int) i {permeableEdge =  i; return self;}
-(BOOL) isPermeable: (GRate*) externalD
{
    if (permeableEdge) {
        *externalD = externalDensity;
        return 1;
    } else return 0;
}
-setExternalDensity: (GRate) x {externalDensity =  x; return self;}
-setNumColors: (int) i {numColors =  i; return self;}
-setHighColorDensity: (GRate) x {highColorDensity = x ; return self;}
-setLowColorDensity: (GRate) x {lowColorDensity =  x; return self;}
-setICLX: (GCoord) d {initCellLX = d; return self;}
-setICLY: (GCoord) d {initCellLY = d; return self;}
-setICUX: (GCoord) d {initCellUX = d; return self;}
-setICUY: (GCoord) d {initCellUY = d; return self;}
-setHaveInitMcr {haveInitMcr = 1; return self;}
-setInitDensity: (GRate) d {initDensity = d; return self;}
-setReplenishDay: (int) rd {replenishDay = rd; return self;}

-(GCoord) getMcrLXCellX: (int) xi Y: (int) yi
{
    return (mcr->lx + ((xi - 1) * cellSizeX));
}
-(GCoord) getMcrLYCellX: (int) xi Y: (int) yi
{
    return (mcr->ly + ((yi - 1) * cellSizeY));
}
-(GCoord) getMcrUXCellX: (int) xi Y: (int) yi
{
    return (mcr->lx + (xi * cellSizeX));
}
-(GCoord) getMcrUYCellX: (int) xi Y: (int) yi
{
    return (mcr->ly + (yi * cellSizeY));
}

-print
{
    printLog(" produces %d ", resource);
    printLog("lattice %d x %d over ", cellsX, cellsY);
    [mcr print];
    return self;
}

-printv
{
    int i, j;
    GRate d;
    
    [super printv];
    
    // need to dump all the params....
    
    if (permeableEdge) {
        printLog("Lattice densities:\n");
        for (j=cellsY+1; j>=0; j--) {
            for (i=0; i <= cellsX+1; i++) {
                d = [self getDensityCellX: i Y: j];
                printLogD("%f ", d);
            }
            printLog("\n");
        }
	/*
        printLog("Lattice2 densities:\n");
        for (j=cellsY+1; j>=0; j--) {
            for (i=0; i <= cellsX+1; i++) {
                d = [self getDensityLattice2CellX: i Y: j];
                printLogD("%f ", d);
            }
            printLog("\n");
        }
	*/
    } else {
        printLog("Lattice densities:\n");
        for (j=cellsY; j>=1; j--) {
            for (i=1; i <= cellsX; i++) {
                d = [self getDensityCellX: i Y: j];
                printLogD("%f ", d);
            }
            printLog("\n");
        }
	/*
        printLog("Lattice2 densities:\n");
        for (j=cellsY; j>=1; j--) {
            for (i=1; i <= cellsX; i++) {
                d = [self getDensityLattice2CellX: i Y: j];
                printLogD("%f ", d);
            }
            printLog("\n");
        }
	*/
    }
    return self;
}

-printColorThresholds
{
    int i;
    printLog("Colors:\n");
    for (i=0; i < numColors; i++) {
        printLogD("\tdensity >= %lf", colorDensity[i]);
        printLog("\t%d\n", colorIds[i]);
    }
    
    return self;
}

@end

@implementation CappedDiffuseQuad

-init 
{
    [super init];
    maxFeedRate 	= 1.2;
    return self;
}

+create
{
    id result;
    result = [[CappedDiffuseQuad createBegin: globalZone] createEnd];
    [result init];
    return result;					
}
-addSiteParams
{
    [super addSiteParams];
    [siteDef addParam: "MaxFeedRate" Vary: 0 DValue: maxFeedRate
          Message: M(setMaxFeedRate:) Owner: self];
    return self;
}

-copymom
    // self=mom, returns copy
    // this actually allocates all the workspace--grandmas nil
{
    CappedDiffuseQuad* newbie;
    
    newbie           		= [CappedDiffuseQuad create];
    newbie->siteDef  		= siteDef;
    newbie->resource 		= resource;
    newbie->replenishDay 	= replenishDay;
    newbie->initDensity		= initDensity;
    newbie->haveInitMcr		= haveInitMcr;
    newbie->initCellLX		= initCellLX;	
    newbie->initCellLY		= initCellLY;	
    newbie->initCellUX		= initCellUX;	
    newbie->initCellUY		= initCellUY;	
    newbie->cellSizeX 		= cellSizeX;
    newbie->cellSizeY 		= cellSizeY;
    newbie->cellsX  		= cellsX;
    newbie->cellsY 		= cellsY;	
    newbie->subSteps 		= subSteps;
    newbie->newRound            = newRound;
    newbie->diffusionRule 	= diffusionRule;
    newbie->diffusionRate 	= diffusionRate;
    newbie->varDiffusionRate 	= varDiffusionRate;
    newbie->lossRate 		= lossRate;
    newbie->permeableEdge	= permeableEdge;
    newbie->externalDensity 	= externalDensity;
    newbie->numColors 		= numColors;
    newbie->highColorDensity 	= highColorDensity;
    newbie->lowColorDensity 	= lowColorDensity;
    newbie->maxFeedRate 	= maxFeedRate;
    
    // One blank lattice, with space for writing density (double*).
    [newbie initLattice];
    [newbie initColors];
    
    newbie->cellArea 		= (double) cellSizeX * cellSizeY;
    newbie->myArea 	= ((double) (cellsX * cellsY)) * newbie->cellArea;
    
    [newbie setMcrLX: mcr->lx LY: mcr->ly 
            UX: mcr->lx + (cellsX * cellSizeX) 
            UY: mcr->ly + (cellsY * cellSizeY)];
    return newbie;
}

-(ResQuant) feedCritter: (Critter*) eater 
            Amount: (GRate) amount Area: (Mcr*) m
    // here we interpret "amount" as percentage
    
    // NOTE:  it's the CALLER's responsibility to apply all relevant
    // debits to the amount, including overlapping agents.  This routine
    // simply applies the areas.
{
    int lxi, lyi, uxi, uyi, i, j;
    GArea mutualArea;
    ResQuant meal, total;		
    double dhere, qhere;	// density (per area) and quantity here
    double percent = amount;

    static BOOL firsttime = 1;
    static Mcr* clipm;    
    Mcr* cellm;    
    
    total = 0;

    if (firsttime) {
        firsttime = 0;
        clipm = [Mcr create];
    }
    
    // don't cross?
    if (![mcr intersectMcr: m Intersection: clipm]) 
        return 0;

    // which cell(s)? these -should- get valid answers....
    if (![self fromPointX: clipm->lx PointY: clipm->ly toCellX: &lxi CellY: &lyi])
        return 0;
    if (![self fromPointX: clipm->ux PointY: clipm->uy toCellX: &uxi CellY: &uyi])
        return 0;

    // and do it
    for (i=lxi; i<=uxi; i++) {
        for (j=lyi; j<=uyi; j++) {
            cellm = [self getMcrCellX: i Y: j];
            if ([cellm intersectMcr: m Intersection: clipm]) {
                mutualArea = [clipm areaMcr];
                dhere = [self getDensityCellX: i Y: j];
                qhere = dhere * cellArea;
                if (dhere > maxFeedRate)
                    dhere = maxFeedRate;
                meal = mutualArea * percent * dhere;
                if (meal > qhere)
                    meal = qhere;
                
                qhere -= meal;
                total += meal;
                dhere = qhere / cellArea;
                [self setDensity: dhere CellX: lxi Y: lyi];
            }
            [cellm drop];
        }
    }
    [eater feedRes: resource Amount: total];
    reservoir -= total;
    fed += total;        
    return total;
}

-setMaxFeedRate: (GRate) mfr {maxFeedRate = mfr; return self;}

@end
