// vfb, orig, 960408: kludgy, as don't want to do interface until
//                    Swarm is ready.
// vfb, 960919: Swarm still not ready, tired of waiting.

#import <swarmobject.h>		
#import <activity.h>	

// One parameter combination to try.  One ES parameter combination
// involves an arbitrary number of couplets, as there are an arbitrary 
// number of objects whose parameters need tweaking. (Sites, species...)
@interface ESCombo: Parameterized {
   @public
    double	sigma;
    double* 	fitness;
    int         numfit;		// length of fitness array
}
+grandma: (int) numTrials;
-addParam: (Param*) p;		// parameter parameter
-(int) countVariables;		// how many parameters are variable?
-vary;				// apply sigma to vary the variables
-apply;				// apply parameters to their owners
-setSigma: (double) s;
-(double) getSigma;
-copymom;
-printv;
@end
			  
// It's possible to carry out the ES algorithm with only the parent, last
// best, and current combinations.  But I keep all the combinations for a
// round, for tidy table generation, and to do the combinations round-robin
// style instead of sequentially.
@interface ESTable: Parameterized {
    // Indices as to which run we're running now.
    int curTrial;	// = -1 until we start, then 0
    int	curOffspring;
    int curGeneration;
    int N;		// number of parameters to vary (internal)
    
    // Parameters controlling the Evolution Strategy algorithm itself.
    int numTrials;	// #runs per parameter combination to find fitness
    int numOffspring;	// number of parameter combinations per round
    int tweakSigma;	// fraction of offspring for higher and lower mutation
    double initSigma;   // standard deviation of mutation, to be scaled 
    double alpha;	// used to evolve the sigmas
    double skipThreshold;// minimum percent of parental fitness to continue     
    double initFitness;	// so can skip calculating the first.
    double maxFitness;	// so can auto-stop if there exists a maximum.
    double distance;	// est. avg. %distance to optimum
    id  fitFunctionOwner;// who to send fitFunction message to
    SEL fitFunction;	// should return a double when sent to owner....
    
    // The list of combinations.  comboList[0] is the round parent.
    // The first "tweakSigma" offspring have sigma = sig(parent) * alpha.
    // The next "tweakSigma" offspring have sigma = sig(parent) / alpha.
    // Remaining offspring (numOffspring - 2*tweakSigma) have parent's sigma.
    ESCombo**	comboList;	// this one's a C array
}

// This automagically sets up a default ES run based on what information
// is on hand in geckoModelSwarm.  In other words, set which parameters you'd
// like to vary in each Species and Site before calling createEnd.  They default
// to none varying.
+create;

// these should not be reset after createEnd
-setNumTrials: (int) nt;
-setNumOffspring: (int) no;

// these two should be set before createEnd, but can be reset along the way
-setFitFunctionOwner: (id) ffo;		
-setFitFunction: (SEL) ff;

-createEnd;

// call this only after createEnd, but before start
-setInitFitness: (double) fit0;		// don't calculate first one--optional

// Routines to override what +create thought was good for you.
// For allocation purposes, had to decide how many trials/offspring already.
-setTweakSigma: (int) ts;		
-setAlpha: (double) a;
-setSkipThreshold: (double) percent;
-setMaxFitness: (double) fitmax;	

// call one of these, not both (distance resets sigma, sigma ignores distance
-setInitSigma: (double) is;
-setDistance: (double) percent;		// defaults to 25% off (a guess)

// We've fiddled all parameters, now set up and begin.
-start;

// Run the next ES scenario, if any left to do.  Returns 0 when done.
-(BOOL) nextScenario;

-calcFitness;

// Both means log and archive files, not both is just log file.
-printRulerBoth: (BOOL) both;
-printTableBoth: (BOOL) both;

-dump;

@end

extern ESTable* scenariosToDo;
extern id metaZone;	// zone for ES to run in.
