// vfb, orig, 960408: kludgy, as don't want to do interface until
//                    Swarm is ready.
// vfb, 960919: Swarm still not ready, tired of waiting.

typedef double PROBABILITY;
#import <stdio.h>
#import <math.h>
#import "random.h"
#import "Base.h"
#import "Critter.h"
#import "Grazer.h"
#import "Carnie.h"
#import "Stat.h"
#import "Param.h"
#import "GeckoControlSwarm.h"

#include "Evolve.h"

// Formats:  note all numbers stuck with 7 spaces.
#define FMTD100000 " %7.0f"
#define FMTD10000  " %7.1f"
#define FMTD1000   " %7.2f"
#define FMTD100    " %7.3f"
#define FMTD10     " %7.4f"
#define FMTD1      " %7.5f"
#define FMTD0      " %7.5f"
#define FMTI       " %7d"
#define BLANK8     "        "

id metaZone = nil;

// some globals for ESTable, so don't have to pass around strings.
char* buffer = NULL;
int buflen;

ESTable* scenariosToDo;

@implementation ESCombo

-addParam: (Param*) p
{
    switch ([p getParamType]) {
        case IntParam : 
            [super addParam: p->label 
                    Vary: p->varyFlag IValue: p->ival
                    Message: M(setPartIValue:) Owner: p];
                    break;
        case DoubleParam : 
            [super addParam: p->label 
                    Vary: p->varyFlag DValue: p->dval
                    Message: M(setPartDValue:) Owner: p];
                    break;
        case StringParam : 
            [super addParam: p->label 
                    Vary: p->varyFlag SValue: p->sval
                    Message: M(setSValue:) Owner: p];
                    break;
        default   : 
            [super addLabelParam: p->label];
    }
    
    return self;
}

-apply
    // Recall that each param in an ESCombo is a param of a param -- to
    // apply it, we apply the ESCombo param change to the source param
    // (probably in Species/Site), then apply the source param back to 
    // -its- source (Critter grandma or whatever.)
{
    id paramIndex;
    Param* param;
    IMP  mptr;
    
    paramIndex = [params begin: metaZone];
    while ( (param = [paramIndex next] ) ) {
        // apply ESCombo value to source param
        [param apply];	
    
        // apply source param to source value
        mptr = objc_msg_lookup(param->owner, M(apply) );
        if ( ! mptr )  {
            geckoError(InvalidCombination, "Can't apply ESCombo parameter\n");
            return 0;
        }
        mptr(param->owner, M(apply)); 
    }
    
    return self;
}

-init: (int) nt
{
    int i;
    [super init];
    fitness = [metaZone alloc: (nt+1) * sizeof(double)];
    for (i=0; i < nt+1; i++)
        fitness[i] = 0.0;
    numfit = nt + 1;
    
    return self;
}

+grandma: (int) numTrials 
{
    id obj;
    id listIndex;
    id paramIndex;
    id speciesList;
    id siteList;
    id paramList;
    Species* species;
    Site* site;
    Param* param;
    
    // allocate some space...   
    obj = [[ESCombo createBegin: metaZone] createEnd];
    [obj init: numTrials];
    
    // now copy all the parameter lists from the model
    siteList = [[geckoControl getModelSwarm] getSites];
    listIndex = [siteList begin: metaZone];
    while ( (site = [listIndex next] ) ) {
        paramList = [site getParams];
        [obj addLabelParam: [site getName]];
        
        paramIndex = [paramList begin: metaZone];
        while ( (param = [paramIndex next] ) ) {
            [obj addParam: param];
        }
    }

    speciesList = [[geckoControl getModelSwarm] getSpecies];
    listIndex = [speciesList begin: metaZone];
    while ( (species = [listIndex next] ) ) {
        paramList = [species getParams];
        [obj addLabelParam: [species getName]];
        
        paramIndex = [paramList begin: metaZone];
        while ( (param = [paramIndex next] ) ) {
            [obj addParam: param];
        }
    }

    return obj;
}

-(int) countVariables
    // how many parameters are variable?
{
    id paramIndex;
    Param* param;
    int count = 0;
   
    paramIndex = [params begin: metaZone];
    while ( (param = [paramIndex next] ) ) {
        if (param->varyFlag)
            count++;   
    }
     
    return count;
}

-setSigma: (double) s { sigma = s; return self; }
-(double) getSigma { return sigma; }

-copymom
{
    int i;
    ESCombo* child;
    id paramIndex;
    Param* param;
    
    child = [ESCombo create: metaZone];
    child->sigma = sigma;
    child->numfit = numfit;
    child->fitness = [metaZone alloc: numfit * sizeof(double)];
    for (i=0; i < numfit; i++)
        child->fitness[i] = 0.0;
    child->params = [List create: [self getZone]];
    paramIndex = [params begin: metaZone];
    while ( (param = [paramIndex next] ) ) {
        [child->params addLast: [param copymomZone: [self getZone]]];
    }
    return child;
}

void truncDouble(char* string, double *d)
{
    char* fmt;
    double newd;
    
    if (*d >= 100000) 
        fmt = FMTD100000;
    else if (*d >= 10000)
        fmt = FMTD10000;
    else if (*d >= 1000)
        fmt = FMTD1000;
    else if (*d >= 100)
        fmt = FMTD100;
    else if (*d >= 10)
        fmt = FMTD10;
    else if (*d >= 1)
        fmt = FMTD1;
    else if (*d >= 0)
        fmt = FMTD0;
    else if (*d >= -1)
        fmt = FMTD10;
    else if (*d >= -10)
        fmt = FMTD100;
    else if (*d >= -100)
        fmt = FMTD1000;
    else if (*d >= -1000)
        fmt = FMTD10000;
    else if (*d >= -10000)
        fmt = FMTD100000;
    sprintf(string, fmt, *d);   
    sscanf(string, "%lf", &newd);
    *d = newd;
}

-vary
{
    id paramIndex;
    Param* param;
    double d;
    int i;
    char* scratch;
    scratch = (char*)[metaZone alloc: 9];
    
    paramIndex = [params begin: metaZone];
    while ( (param = [paramIndex next] ) ) {
        if ([param getVary]) {
            switch ([param getParamType]) {
                case DoubleParam:
                    do {
                        d = [param getPartDValue] * (1.0 + sigma * GaussRand());
                        truncDouble(scratch, &d);
                    } while (!samesign(d, [param getPartDValue]));
                    [param setPartDValue: d];
                    break;
                case IntParam:
                    do {
                        i = (int) (0.5 + 
                            [param getPartIValue] * (1.0 + sigma * GaussRand()));
                    } while (!samesign(i, [param getPartIValue]));
                    [param setPartIValue: i];
                    break;
                default : break;
            }
        }
    }
    
    return self;
}

-printv
{
    int i;
    printLogD("ESCombo: sigma = %f\n", sigma);
    printLog("fitness: ");
    for (i=0; i < numfit; i++)
        printLogD("%f ", fitness[i]);
    printLog("\n");
    [params forEach: M(println)];
    return self;
}

@end

@implementation ESTable


-sigmaFromD
{
    double D;
    
    // for now, rather than keep a table, let's assume offspring close to 10
    double C10 = 1.54;
    
    // distance is per-parameter error, D is overall Euclidean distance
    D = sqrt(N * distance * distance);
 
    initSigma = C10 * D / N;

    return self;     
}

-addParams
{
    [super init];
    [self addParam: "NumOffspring" Vary: 0 IValue: numOffspring 
          Message: M(setNumOffspring:) Owner: self];
    [self addParam: "NumTrials" Vary: 0 IValue: numTrials 
          Message: M(setNumTrials:) Owner: self];
    [self addParam: "TweakSigma" Vary: 0 IValue: tweakSigma
          Message: M(setTweakSigma:) Owner: self];
        
    [self addParam: "InitSigma" Vary: 0 DValue: initSigma
          Message: M(setInitSigma:) Owner: self];
    [self addParam: "SkipThreshold" Vary: 0 DValue: skipThreshold
          Message: M(setSkipThreshold:) Owner: self];
    [self addParam: "MaxFitness" Vary: 0 DValue: maxFitness
          Message: M(setMaxFitness:) Owner: self];
    [self addParam: "InitFitness" Vary: 0 DValue: initFitness
          Message: M(setInitFitness:) Owner: self];
    [self addParam: "Distance" Vary: 0 DValue: distance
          Message: M(setDistance:) Owner: self];
    // currently have no object or selector parameters
    return self;
}

// This automagically sets up a default ES run based on what information
// is on hand in geckoModelSwarm.  In other words, set which parameters you'd
// like to vary in each Species and Site before calling this.  They default
// to none varying until told otherwise.
+create
{
    ESTable* obj;
    ESCombo* scratchGrandma;
    id speciesList, siteList;

    // error check 
    speciesList = [[geckoControl getModelSwarm] getSpecies];
    siteList = [[geckoControl getModelSwarm] getSites];
    if ((siteList == nil) || (speciesList == nil)) {
        geckoError(InvalidCombination, 
            "Model must be initialized before evolution strategy engine.\n"); 
        return self;
    }
    
    if (metaZone == nil)
        metaZone = [Zone create: globalZone];
    obj = [ESTable createBegin: metaZone];
    [obj setName: "EvolutionStrategy"];
    
    // defaults--reconsidered in createEnd
    obj->numTrials     =   5;
    obj->numOffspring  =  10;
    obj->tweakSigma    =   3;
    obj->distance      = .25;
    obj->skipThreshold = .90;
    obj->alpha         = 1.3;
    obj->maxFitness    = 0.0;
    obj->initFitness   = 0.0;
    
    scratchGrandma = [ESCombo grandma: 1];
    obj->N = [scratchGrandma countVariables];
    [obj sigmaFromD];
    [scratchGrandma drop];
    
    [obj addParams];
    
    return obj;
}

-createEnd
{
    IMP  mptr;

    mptr = objc_msg_lookup( fitFunctionOwner, fitFunction );
    if ( ! mptr )  {
        geckoError(InvalidCombination, "Fitness function not found.\n");  
        return self;
    }
    
    // check for mandatory items--can't even allocate space.
    if (!(numTrials)) {
        geckoError(InvalidCombination, 
            "Evolution strategy--no trials, nothing to do.\n"); 
        return self;
    }
    
    if (!(numOffspring)) {
        geckoError(InvalidCombination,
            "Evolution strategy--no offspring, nothing to do.\n"); 
        return self;
    }

    comboList    = [metaZone alloc: (numOffspring + 1) * sizeof(ESCombo*)];

    // comboList[0] is the parent of the round--first round, it's the given
    // note that numTrials and the parameters to vary are DONE now--never
    // going to be reconsidered for duration of Evolution Strategy run.
    comboList[0] = [ESCombo grandma: numTrials];
    comboList[0]->sigma = initSigma;
    
    // general book-keeping    
    curGeneration= 0;
    curOffspring = 0;
    curTrial     = 0;	
    
    scenariosToDo = self;
    return self;
}

// Routines to override what +create thought was good for you.
-setNumTrials: (int) nt { numTrials = nt; return self; }
-setNumOffspring: (int) no { numOffspring = no; return self; }
-setTweakSigma: (int) ts { tweakSigma = ts; return self;}
-setAlpha: (double) a {alpha = a; return self;}
-setSkipThreshold: (double) percent { skipThreshold = percent; return self; }
-setMaxFitness: (double) fitmax { maxFitness = fitmax; return self; }	
-setInitFitness: (double) fit { initFitness = fit; return self; }	
-setFitFunctionOwner: (id) ffo { fitFunctionOwner = ffo; return self; }
-setFitFunction: (SEL) ff { fitFunction = ff; return self; }

// call one of these, not both (distance resets sigma, sigma ignores distance
-setInitSigma: (double) is {initSigma = is; return self;}
-setDistance: (double) percent
{
    distance = percent;
    [self sigmaFromD];
    return self;
}

-comboInitFitness: (double) fit0
{
    int i;
    for (i=0; i < numTrials + 1; i++)
        comboList[0]->fitness[i] = fit0;
    return self;    
}

-makeTable
    // starting situation:  comboList[0] is the parent.
{
    int i;
    double mamasig = [comboList[0] getSigma];
    
    for (i=1; i < numOffspring + 1; i++) {
        comboList[i] = [comboList[0] copymom];
        
        if (i <= tweakSigma)
            [comboList[i] setSigma: (mamasig * alpha)];
        else if (i <= 2 * tweakSigma)
            [comboList[i] setSigma: (mamasig / alpha)];
        [comboList[i] vary];
    }
    return self;
}

-(BOOL) flunkOffspring
{
    double flunktest = skipThreshold * comboList[0]->fitness[numTrials];

    if (curTrial == 0)
        return 0;
    
    if (comboList[curOffspring]->fitness[0] < flunktest) 
        return 1;
    else
        return 0;
}

-(BOOL) nextTable
{
    static int lastparent   = 0;  // #rounds where parent won
    static int maxparent    = 3;  // max to run win parent

    double highscore;
    int highwho, i;
    ESCombo* swap;
        
    [self printRulerBoth: 1];
    [self printRulerBoth: 1];
    [self printRulerBoth: 1];
    
    highscore   = comboList[0]->fitness[numTrials];
    highwho     = 0;
    for (i=1; i<numOffspring+1; i++) {
        if (comboList[i]->fitness[numTrials] > highscore) {
            highscore = comboList[i]->fitness[numTrials];
            highwho   = i;
        }
    }
    
    // have we reached maxfitness?  no point running further, then.
    if (highscore == maxFitness) {
        printBoth("\n\n# Evolution Strategy - max fitness reached.\n");
        [comboList[highwho] apply];
        [geckoControl echoInput];
        return 0;
    }
    
    // parent was better than any of the offspring
    // proceed to next table, unless we're tired of this parent
    if (highwho == 0) {
        lastparent++;
        if (lastparent == maxparent) {
            return 0;
        }
        
        // replace sigma on parent
        comboList[0]->sigma /= (alpha * alpha);
        
    // we had a clear winner--reset lastparent
    } else {
        lastparent = 0;
        
        // shuffle for our dropping convenience.
        swap = comboList[0];
        comboList[0] = comboList[highwho];
        comboList[highwho] = swap;
        printBoth("\n\n# Evolution Strategy - round winner.\n");
        [comboList[0] apply];
        [geckoControl echoInput];
    }
    
    // reset counters...
    curGeneration++;
    curTrial = 0;
    curOffspring = 0;
    
    // drop all reject offspring
    for (i=1; i<numOffspring+1; i++) {
        [comboList[i] drop];
    }

    [self makeTable];
    [self printTableBoth: 1];

    // call back nextScenario to run.
    if (![self nextScenario])
        printBoth("*** Error starting up next round evolutionary strategy\n");
    return 1;
}

-(BOOL) nextScenario
    // load up next offspring, or next table.
{
    int i, t;
    double finalscore;
    int    scores;
    long seed;

    // find result from previous run, if any...
    [self calcFitness];
    
    // run numTrials worth of proto-parent first?
    if ( (curGeneration == 0) && 
         (curOffspring == 0) &&
         (comboList[0]->fitness[numTrials] == 0.0) ) {
        if (curTrial < (numTrials - 1)) {
            // already loaded, just run it again
            [comboList[curOffspring] apply];
            curTrial++;
            return 1;
        } else {
            // we have run numTrials worth of protoparent--
            // find the average, set it, then proceed normally
            for (i=0; i < numTrials; i++)
                comboList[0]->fitness[numTrials] +=
                    comboList[0]->fitness[i];
            comboList[0]->fitness[numTrials] /= (double) numTrials;
            curOffspring = 0;
            curTrial = 0;
            [self printTableBoth: 1];
        }
    }
    
    // load up next offspring?
    curOffspring++;
    if (curOffspring > numOffspring) {
        curTrial++;
        curOffspring = 1;
    }

    // increment again if this offspring already flunked out
    do {
        if ([self flunkOffspring]) {
            curOffspring++;
            if (curOffspring > numOffspring) {
                curTrial++;
                curOffspring = 1;
            }
        }
    } while ((curTrial < numTrials) && [self flunkOffspring]);

    // done with this table? calculate averages
    if (curTrial == numTrials) {
        for (i=1; i<numOffspring+1; i++) {
            finalscore = 0;
            scores = 0;
            for (t=0; t < numTrials; t++) {
                if (comboList[i]->fitness[t] != 0.0) {
                    scores++;
                    finalscore += comboList[i]->fitness[t];
                }
            }
            if (scores > 0) 
                finalscore /= scores;
            comboList[i]->fitness[numTrials]= finalscore;
        }
        [self printTableBoth: 1];
        return [self nextTable];
    }
    
    // not done--load next scenario
    printBoth("\n\nRunning Offspring Scenario # %2d\n\n", curOffspring);
    seed = NO_SEED;
    seed = InitRand(seed);
    printBoth("Evolution strategy random seed = %ld\n", seed);
    [comboList[curOffspring] apply];
    
    return 1;
}

-start
    // We've fiddled all parameters, now set up and begin.
{
    long seed = [[geckoControl getModelSwarm] getRandomSeed];
    InitRand(seed);

    // 0. echo starting world 
    [geckoControl echoInput];
    printBoth("\n\n\n");

    printBoth("Evolution strategy random seed = %ld\n", seed);

    // 1. make a table
    [self makeTable];
    if (initFitness != 0.0) 
        [self comboInitFitness: initFitness];

    // 2. report
    [self printTableBoth: 1];

    // counters ready...
    curTrial = -1;
    curOffspring = 0;
    [self nextScenario];
    return self;
}

-calcFitness
{
    double d;
    char scratch[9];
    IMP  mptr;

    // not the most logical spot, but skip this if didn't really need to 
    // calculate fitness.
    if (comboList[curOffspring]->fitness[curTrial] > 0.0)
        return self;
    
    mptr = objc_msg_lookup( fitFunctionOwner, fitFunction );
    if ( ! mptr )  {
        geckoError(InvalidCombination,
            "Invalid fitness function in evolution strategy\n");  
        return 0;
    }
    mptr( fitFunctionOwner, fitFunction, &d); 
    truncDouble(scratch, &d);
    comboList[curOffspring]->fitness[curTrial] = d;

    [self printTableBoth: 0];
    printArchiveD("    Fitness = %lf\n", d);

    return self;
}

-printRulerBoth: (BOOL) both
{
    int i;
    for (i=0; i < buflen - 1; i++)
        buffer[i] = '-';
    buffer[buflen - 1] = '\n';
    if (both) 
        printBoth(buffer); 
    else 
        printLog(buffer);
    return self;
}

-fmtParamVal: (id) p Column: (int) col
{
    char scratch[9];
    double d;
    int i;
    
    switch ([p getParamType]) {
        case DoubleParam: d = [p getDValue];
                        truncDouble(scratch, &d);
                        sprintf(&(buffer[col]), "%s", scratch);
                        break;
        case IntParam:    i = [p getIValue];
                        sprintf(&(buffer[col]), FMTI, i);
                        break;
        case StringParam: 
                        sprintf(&(buffer[col]), BLANK8);
                        break;
        default : 	sprintf(&(buffer[col]), BLANK8);
    }
    return self;
}


-printTableBoth: (BOOL) both 
    // Hard-coded for now.
    // Labels get 12 spaces including " :"
    // Numbers get 8 spaces including leading " "
    // Special entries:  stopday and sigma
{
    int col, combo, score, stopday;
    double fitness;
    static id* paramIndices = NULL;
    id param;
    
    // format into buffer, apply "both" test once per line.
    if (buffer == NULL) {
        buflen = 12 + (numOffspring + 1) * 8 + 1;
        buffer = (char*) [metaZone alloc: buflen];
    }
    
    stopday = [[geckoControl getModelSwarm] getStopDay];

    [self printRulerBoth: both];
    sprintf(buffer, "ES Generation %d, #Rounds = %d\n", 
        curGeneration, stopday); 
    if (both) printBoth(buffer); else printLog(buffer);
    [self printRulerBoth: both];
    
    sprintf(buffer, "Parameter  :  parent"); col = 20;
    for (combo = 1; combo < numOffspring + 1; combo++) {
        sprintf(&(buffer[col]), "   %2d   ", combo);
        col += 8;
    }
    buffer[col] = '\n';
    if (both) printBoth(buffer); else printLog(buffer);
    [self printRulerBoth: both];

    // need gob of indices to traverse all these lists, unfortunately.
    if (paramIndices == NULL) {
        paramIndices = (id*) [metaZone alloc: (numOffspring+1) * sizeof(id)];
    }
    for (combo = 0; combo < numOffspring + 1; combo++) {
        paramIndices[combo] = [[comboList[combo] getParams] begin: metaZone];
    }
    
    // first item on the list:  sigma
    sprintf(buffer, "%10s :", "Sigma");  col = 12;  
    for (combo = 0; combo < numOffspring + 1; combo++) {
        sprintf(&(buffer[col]), FMTD1, [comboList[combo] getSigma]);
        col += 8;
    }
    buffer[col] = '\n';
    if (both) printBoth(buffer); else printLog(buffer);

    // OK, print each line of parameter settings
    while ( (param = [paramIndices[0] next] ) ) {
        // one parent, label and value
        sprintf(buffer, "%10s", [param getName]);  col = 10; 
        if ([[param getOwner] isMemberOf: [OffsetParam class]])
            buffer[col] = '+';
        else
            buffer[col] = ' ';
        buffer[col+1]   = ':';
        col+=2;
     
        [self fmtParamVal: param Column: col];
        col += 8;
        
        // the offspring--params must be in same order all around....
        for (combo = 1; combo < numOffspring + 1; combo++) {
            param = [paramIndices[combo] next];
            [self fmtParamVal: param Column: col];
            col += 8;
        }

        buffer[col] = '\n';
        // was this a varying or invariant parameter?  
        if (([param getParamType] == LabelParam) || 
            ([param getParamType] == StringParam) ||
             [param getVary]) {
            if (both) printBoth(buffer); else printLog(buffer);
        } else 
            if (both) printBoth(buffer);
    }
    [self printRulerBoth: both];

    // now for the fitness scores
    for (score = 0; score < numTrials + 1; score++) {
        if (score == 0) 
            sprintf(buffer, "fitness %2d :", score);
        else if (score == numTrials) { 
            [self printRulerBoth: both];
            sprintf(buffer, "average    :");
        } else
            sprintf(buffer, "%10d :", score);
        col = 12;
        
        for (combo = 0; combo < numOffspring + 1; combo++) {
            fitness = comboList[combo]->fitness[score];
            if (fitness == 0.0)
                sprintf(&(buffer[col]), " -------");
            else
                truncDouble(&(buffer[col]), &fitness);
            col += 8;
        }
        buffer[col] = '\n';
        if (both) printBoth(buffer); else printLog(buffer);
    }
    [self printRulerBoth: both];
   
    return self;
}

-dump
{
    printBoth("@begin %s\n", [self getName]);
//    [super dump];
    [params forEach: M(dump)]; 
    printBoth("@end %s\n", [self getName]);
    return self;
}

@end
