// Original, vfb, 970424.

#import "Base.h"
#import "ResourceQuad.h"
#import "LandRoot.h"
#import "FeedQuad.h"
#import "random.h"
#import "Critter.h"
#import "GeckoControlSwarm.h"
#import <simtools.h>
#import <math.h>

@implementation FeedQuad

-init 
{
    [super init];
    prodfunc     = 0;
    feedfunc     = 0;
    reserveLimit = 0;
    return self;
}

+create 
{
    id result = [[FeedQuad createBegin: globalZone] createEnd];
    [result init];
    return result;					
}

-addSiteParams
{
    [siteDef addParam: "Color" Vary: 0 IValue: color
          Message: M(setColor:) Owner: self];
    [siteDef addParam: "McrLX" Vary: 0 DValue: mcr->lx
          Message: M(setMcrLX:) Owner: self];
    [siteDef addParam: "McrLY" Vary: 0 DValue: mcr->ly
          Message: M(setMcrLY:) Owner: self];
    [siteDef addParam: "McrUX" Vary: 0 DValue: mcr->ux
          Message: M(setMcrUX:) Owner: self];
    [siteDef addParam: "McrUY" Vary: 0 DValue: mcr->uy
          Message: M(setMcrUY:) Owner: self];
    [siteDef addParam: "Resource" Vary: 0 IValue: resource
          Message: M(setResource:) Owner: self];
    [siteDef addParam: "ProdFunc" Vary: 0 DValue: prodfunc
          Message: M(setProdFunc:) Owner: self];
    [siteDef addParam: "Limit" Vary: 0  DValue: reserveLimit
          Message: M(setLimit:) Owner: self];
    
    return self;
}

-landGrandma: (Site*) s
    // one default FeedQuad, 250x250, produces 
    // 1.0 sunwater/unit area, limit 1.2
{
    siteDef = s;
    [self setMcrLX: 0 LY: 0 UX: 250.0 UY: 250.0];
    [self produce: sunwater 
          Rate: 1.0
          Limit: 1.2];
    color = GroundMidMidColor;
    [self addSiteParams];

    return self;
}

-copymom
    // self=mom, returns copy
{
    FeedQuad* newbie;
    
    newbie           = [FeedQuad create];
    newbie->siteDef  = siteDef;
    newbie->color    = color;
    
    [newbie setMcrLX: mcr->lx LY: mcr->ly UX: mcr->ux UY: mcr->uy];
    [newbie produce: resource Rate: prodfunc Limit: reserveLimit];
    
    return newbie;
}

-setProdFunc: (GRate) rate {prodfunc = rate; return self;}
-setLimit: (GRate) lim {reserveLimit = lim; return self;}
-setColor: (int) c {color = c; return self;}
-(int) getColor {return color; }

-produce: (ResType) res Rate: (GRate) rate Limit: (GRate) lim
    // This production function replaces the feedquad production function.  
{
    // we won't add production of nullres
    if (res == nullres)
        return self;

    // whatever we've got, it's no longer valid
    reservoir = startres = 0;

    resource     = res;
    prodfunc     = rate;
    reserveLimit = lim;
    
    return self;
}

-addResourceAmount: (ResQuant) amount X: (GCoord) x Y: (GCoord) y
    // AMOUNT, not density
{
    if ([mcr ptInsideX: x Y: y])
        reservoir += amount;
    return self;
}

-(ResQuant) feedCritter: (Critter*) eater 
       Amount: (ResQuant) amount Area: (Mcr*) m
    // Edge of this quad qualifies as a competitor for setting direction
    // that eater wanna grow.  "Amount" interpreted as percentage.
    // "Resource" is ignored.  We feed what we got.
{
    ResQuant meal, lostmeal;
    GCoord wx, wy;
    McrRelation crossflag;
    static BOOL firsttime = 1;
    static Mcr* clipm;    

    ResQuant percent = amount;
    
    if (firsttime) {
        firsttime = 0;
        clipm = [Mcr create];
    }

    crossflag = [mcr intersectMcr: m Intersection: clipm];
    if ([eater eats: resource]) {
        meal = feedfunc * percent * [clipm areaMcr];
        reservoir -= meal;
        [eater feedRes: resource Amount: meal];
    }

    // set wanna if edge cost this critter resources
    if (crossflag == intersect) {
        lostmeal = [m areaMcr] - [clipm areaMcr];
        [clipm getMcrCenter: &wx Y: &wy];
        wx -= eater->x;
        wy -= eater->y;
        [eater addDirectionX: wx Y: wy Weight: lostmeal];
    }
   
    // [clipm drop];

    fed += meal;    
    return meal;
}

-stepResources 
{
    GArea area;

    area = [mcr areaMcr];
    if (reservoir < 0.0)
        reservoir = 0.0;

    reservoir += prodfunc * area;

    if (reservoir > (reserveLimit * area))
        reservoir = (reserveLimit * area);

    feedfunc = reservoir / area;
    startres = reservoir;
    
    return self;
}

-(BOOL) produces: (ResType) res {return (resource == res);}

-(ResQuant) curFeed {return feedfunc;}
-print
{
    printLog(" produces %d ", resource);
    printLogD("rate %lf, ", prodfunc);
    printLogD("limit %lf over ", reserveLimit);
    [mcr print];
    return self;
}

@end

