// 970512, vfb:  modified to uptake resources rather than be fed passively.

// 970311, original, vfb:  want to try crafting pseudobacteria a la question
// from Jan Kreft, U.Cardiff, on Swarm list.  Fission2D splits in two at
// breed_radius, and doesn't strew child.  Parameterized accordingly.

#import "random.h"
#import "Critter.h"
#import "CritterZoo.h"
#import "Fission2D.h"
#import "LandRoot.h"
#import "ResourceQuad.h"
#import "Diffuse.h"
#import "GeckoControlSwarm.h"
#import "Stat.h"
#import <stdlib.h>
#import <math.h>
#import <swarmobject.h>
#import <simtools.h>

@implementation Fission2D

Fission2D* recycleFission2Ds = 0;

-recycle
{
    [species tallyDeath: self];

    hook = recycleFission2Ds;
    recycleFission2Ds = self;
    [species decrementCensus];
    return nil;
}

-recycleClear
{
    Fission2D* togo;
    while (recycleFission2Ds) {
        togo = recycleFission2Ds;
        recycleFission2Ds = recycleFission2Ds->hook;
        [togo->energy drop];
        [togo drop];
    }
    return self;
}

-reuse
{
    Fission2D* result = recycleFission2Ds;
    recycleFission2Ds = recycleFission2Ds->hook;
    result->hook    = nil;
    return result;
}

+create 
{
    id result;
    
    if (recycleFission2Ds) {
        result = [recycleFission2Ds reuse];
    } else {
        result = [Fission2D createBegin: globalZone];
        [result allocEnergy];
    }
    [result init];
    
    return result;					
}

-copymom
    // create an agent matching mom in all heritable respects
{
    Fission2D* baby;
    Quad* newQuad;

    baby = [Fission2D create];
    baby->species   = species;
    [energy copyTo: baby->energy];
    baby->color     = color;

    [species incrementCensus];
    [species tallyBirth: self];
    
    baby->breedsize   = breedsize;
    baby->kidsize     = kidsize;
    baby->minsize     = minsize;
    baby->scaleradius = scaleradius;
    baby->scalev      = scalev;
    baby->shoveRadius = shoveRadius;
    baby->uptakeRate  = uptakeRate;
    baby->convertBack = convertBack;

    newQuad = [Quad create];
    [newQuad setData: baby Type: critter];
    [newQuad setSibling: nil];
    [newQuad setChild: nil];
    [newQuad setParent: nil];
    [baby setQuad: newQuad];

    return baby;
}

-setBreedSize: (GCoord) bs
{
    ResQuant breedvol, kidvol, minvol;
    
    breedsize = bs;
    
    breedvol = radius2volume(bs, scaleradius, scalev);
    kidvol   = breedvol / 2.0;
    kidsize  = volume2radius(kidvol, scaleradius, scalev);
    minvol   = 0.5 * kidvol; // probably not wise, but what the hey.
    minsize  = volume2radius(minvol, scaleradius, scalev);

    // OK, I want maximum overlap to be...40%?  set tax rate accordingly.
    
    return self;
}

-setUptakeRate: (GRate) ur {uptakeRate = ur; return self;}
-setShoveRadius: (GCoord) r {shoveRadius = r; return self;}
-setConvertBack: (GRate) cb {convertBack = cb; return self;}

-(GCoord) getMoveRadius 
    // wannax and wannay are unit direction vectors, weighted by how much
    // I wanna go that way, based on how much overlap cost me...not obviously
    // related to radius in any way....
{ 
    GCoord wannadist = sqrt(wannax * wannax + wannay * wannay);
    return wannadist;
}

-liveGrandma: (Species*) s
    // initialize/reinit a default Fission2D
{
    [energy addConvertFrom: substrate To: bodymass Rate: .1];
    [energy addTax: bodymass Rate: .0480 Exponent: .75];

    [self setBreedSize: 5.0];
    scaleradius = 1.0;
    scalev = 1.0/M_PI;
    [self setUptakeRate: 0.1];
    [self setShoveRadius: 1.2];
    species = s;
    [self setColor: [species getColor]];

    [self addSpeciesParams];
    
    return self;
}

-addSpeciesParams
{
    GRate rate;
    
    rate = [energy getTaxRate: [energy getIam]];
    [species addParam: "TaxRate" Vary: 0 DValue: rate 
          Message: M(setTaxRate:) Owner: self];
    rate = [energy getConvertFrom: [energy getIeat] To: [energy getIam]];
    [species addParam: "Convert" Vary: 0 DValue: rate 
          Message: M(setConvertRate:) Owner: self];
    [species addParam: "ConvertBack" Vary: 0 DValue: convertBack
          Message: M(setConvertBack:) Owner: self];
    [species addParam: "BreedSize" Vary: 0 DValue: breedsize
        Message: M(setBreedSize:) Owner: self];
    [species addParam: "UptakeRate" Vary: 0 DValue: uptakeRate
        Message: M(setUptakeRate:) Owner: self];
    [species addParam: "ShoveRadius" Vary: 0 DValue: shoveRadius
        Message: M(setShoveRadius:) Owner: self];
    [species addParam: "ScaleR" Vary: 0 DValue: scaleradius 
            Message: M(setScaleR:) Owner: self];
    [species addParam: "ScaleV" Vary: 0 DValue: scalev 
            Message: M(setScaleV:) Owner: self];
    return self;   
}

-(GCoord) getChildRadius
    // distance at which to place kid
{
    return 0.5 * kidsize;  
}

-die
{
    LandRoot* landRoot = [[geckoControl getModelSwarm] getLandRoot];
    ResList* taxes = [energy getStores];
    
    [species tallyLossStarve: self Amount: taxes->list[[energy getIam]]];

    [landRoot payTaxX: x Y: y Resource: [energy getIeat]
              Amount: taxes->list[[energy getIam]] * convertBack];
    // do NOT drop taxes, it's a pointer, not a copy

    // note that recycle, called by delData, does the census
    [myquad delData];

    return nil;
}

-move
    // OK, so I lied.  In theory, it shoves neighbors.  In practice,
    // each finds neighbors to be shoved by.  Would be more efficient
    // doing as a batch once-through process, but first do it right
    // then make it faster if and when.  As of the moment, the distance
    // shoved is simply the vector sum of overlap radii.
{
    Quad* myRoot;
    Critter* neighbor;
    Quad* nq;
    GCoord overlapRadius;
    
    static BOOL firsttime = 1;
    static Mcr* shovewin;		// equiv-area square
        
    if (firsttime) {
        firsttime = 0;
        shovewin = [Mcr create];
    }
    
    wannax = wannay = 0.0; 

    // pick around for neighbors inside shoveRadius * radius
    myquad->visible = 0;
    [shovewin circleToSquare: x Y: y Radius: radius * shoveRadius];
    myRoot = [self getRoot];
    nq = [myRoot pickWin: shovewin Root: myRoot Invisibles: 0]; 
    while (nq) {
        // get one shove
        neighbor = nq->data;
        overlapRadius = radius * shoveRadius;
        overlapRadius -= distance(x, y, neighbor->x, neighbor->y) -
                         neighbor->radius;
        if (overlapRadius > 0) {
            [self addDirectionX: x - neighbor->x Y: y - neighbor->y
                  Weight: overlapRadius];
        }
        
        // next
        nq = [nq pickWin: shovewin Root: myRoot Invisibles: 0]; 
    }
    
    myquad->visible = 1;

    // now do it.
    x += wannax;
    y += wannay;
    [self getMcr: myquad->mcr];
    [myquad growMcr];

    return self;
}

-uptake
    // The amount is added to self's gullet by the landscape Quad calling
    // Critter-feedRes.  The feeding amount is 
    // 		surface area * external density * uptakeRate
    // Since surface area is 4 * circle area, and feedCritter pays out
    // amount * density * area, we end up with the peculiar argument
    // 4.0 * uptakeRate being the Amount passed to feedCritter.
{
    DiffuseQuad* actualSite;
    double externalD;
    ResQuant fed;
    static Mcr* feedwin;		// equiv-area square
    static BOOL firsttime = 1;
    LandRoot* landRoot = [[geckoControl getModelSwarm] getLandRoot];
        
    if (firsttime) {
        firsttime = 0;
        feedwin = [Mcr create];
    }
    [feedwin circleToSquare: x Y: y Radius: radius];
    
actualSite = [landRoot firstQuadProducing: 1 Over: feedwin];
if (!actualSite)
    printLog("actualSite is nil, let's proceed\n");
externalD = [actualSite getDensityAtX: x Y: y];
printLogD("externalD = %f\n", externalD);  
    fed = [landRoot feedCritter: self Resource: substrate
                    Amount: 4.0 * uptakeRate 
                    Area: feedwin];
    return self;
}

-hunt
    // feeds itself 
{
    [self uptake];
    return self;
}

-(BOOL) reproduce
    // False if didn't.
{
    Fission2D* baby;
    GCoord distance;
    ResList* kidgets;
    double angle;
    Quad* myRoot;
    
    if (radius < breedsize)
        return 0;

    angle = 2.0 * M_PI * EvenRand();
    
    // record a birth--equivalent to generation counter, actually, and
    // firstbornon is the beginning of current generation....
    firstbornon = getCurrentTime();
    kidcount++;

    // make the kid
    baby         = [self copymom];
    distance     = [self getChildRadius];
    baby->x      = x + ( distance * cos(angle) );
    baby->y      = y + ( distance * sin(angle) );
    baby->wannax = cos(angle);
    baby->wannay = sin(angle);
    baby->radius = kidsize;

    // transfer resources 
    kidgets = [self kidshare];
    [energy debit: kidgets];
    [self resetRadius];
    baby->energy = [baby->energy setStores:  kidgets];

    // hang on quad tree
    [baby getMcr: baby->myquad->mcr];
    myRoot = [self getRoot];
    [myRoot addNode: baby->myquad];

    [kidgets drop];
    
    // now I move, too....
    x -= ( distance * cos(angle) );
    y -= ( distance * sin(angle) );
    [myquad growMcr];
    return 1;
}

@end
