// Copyright (C) 1996 Yale Institute for Biospheric Studies and Ginger Booth.
// This library is distributed without any warranty; without even the
// implied warranty of merchantability or fitness for a particular purpose.
// See file LICENSE for details and terms of copying.

// vfb, original, 960620:  copy from HeatbugBatchSwarm and fit to Gecko.

// The Gecko Control Swarm encapsulates the model swarm and batch and graphical 
// user interface swarms, and the user interface common to batch and graphics,
// and decides which run to run next.

#import <swarmobject.h>
#import "GeckoModelSwarm.h"
#import "GeckoDisplaySwarm.h"
#import "Colors.h"
#import <space.h>
#import <activity.h>
#import <collections.h>

@interface ControlParams : Parameterized {
  @public
    int displaycycle;		          
    int statcycle;		          
    int timestep;		          
    int geckoRuns;			  
    BOOL ESRun;				  
    char* fileStem;
}

+create;
-setStatCycle: (int) i;
-setTimeStep: (int) i;
-setDisplayCycle: (int) i ;
-setGeckoRuns: (int) i ;
-setESRun: (int) i ;
-setFileStem: (char*) s;
@end

@interface GeckoControlSwarm : Swarm {
    char* protocolFileName;		  
    FILE* protocolFile;
    
    ControlParams* settings;
    int displayFlag;			 

    id reportActions;			  // schedule data structs
    id reportSchedule;
    id stopSchedule;
    
    @public // for the moment...
    ControlPanel* controlPanel;		  // if displaying....
    GeckoModelSwarm *   geckoModelSwarm;  // the Swarm we're running
    GeckoDisplaySwarm * geckoDisplaySwarm;// if displaying
    
    // presently not optional
    TraceGraph* popGraph;
    TraceGraph* biomassGraph;

    FILE* logFile;			  // running reports
    FILE* archiveFile;			  // result reports
}

+createBegin: (id) aZone;
-setFileStem: (char*) stem;	// stem.log, stem.arc, stem.popGraph, etc.
-(char*) getFileStem;		// stem.log, stem.arc, stem.popGraph, etc.
-setDisplayFlag: (int) df;
-(BOOL) getDisplayFlag;
-setStatCycle: (int) sc;
-setGeckoRuns: (int) gr;
-(BOOL) getESRun;		
-setESRun;			// default off, turn on.

-setProtocol: (char*) s;	// default off, turn on.
-(FILE*) getProtocol;		// nil if no protocol file

-(int) getGeckoRuns;
-(GeckoModelSwarm*) getModelSwarm;
-(FILE*) getLogFile;
-(FILE*) getArchiveFile;
-(int) getStatCycle;

-buildObjects;
-buildActions;
-activateIn: (id) swarmContext;

-go;
-init;
-stopRunning;
-echoInput;	// not yet

-printv;	
-printCensus;	
-println;	// basically, calls all the printlns, so > 1 line...

@end

extern GeckoControlSwarm * geckoControl;

