// Copyright (C) 1996 Yale Institute for Biospheric Studies and Ginger Booth.
// This library is distributed without any warranty; without even the
// implied warranty of merchantability or fitness for a particular purpose.
// See file LICENSE for details and terms of copying.

// vfb, original, 960618:  copy from HeatbugModelSwarm and fit to Gecko.

// The Gecko Model swarm encapsulates all the objects used in the
// simulated gecko world itself (but not the user interface objects)

#import <swarmobject.h>
#import <space.h>
#import <activity.h>
#import <collections.h>
#import "Quad.h"
#import "Param.h"
#import "LandRoot.h"
#import "Species.h"
#import "Site.h"

@interface ModelParams : Parameterized {
  @public
    int stopday;	        // rounds per this run
    int reQuadRound;		// when Land should requad plants
    int season;			// on this cycle, call Site/Species-onSeason
}

+create;
-setStopDay: (int) i;
-setReQuad: (int) i;
-setSeason: (int) i;
-(long) getRandomSeed;
@end

@interface GeckoModelSwarm : Swarm {
  @public
    ModelParams* settings;	// Parameterized (protocol) portion
    
    int plantcycle;		// plants&land go plantcycleth round
    int carniecycle;		// carnies go carniecycleth round

    // scheduling data structures
    id modelActions;		
    id modelSchedule;

    // world data structures
    id	  species;		// list of all defined species
    id	  sites;		// list of all defined sites (feedquads)
    id    scenarios;		// ES tables
    Quad* worldRoot;		// root of all roots-all Gecko world/agent data
    LandRoot* landRoot;		// landscape manager
    Quad* mobileRoot;		// grazers and carnies
    Quad* plantRoot;		// plants (eat land)
}

// Expected methods to be provided for Swarms.
+createBegin: (id) aZone;		
// -createEnd is inherited from super.

// Access methods into the model swarm.  Allow model swarm to be observed.
-(Quad*) getWorldRoot;
-(LandRoot*) getLandRoot;
-(Quad*) getMobileRoot;
-(Quad*) getPlantRoot;
-(int) getStopDay;
-(long) getRandomSeed;
-setStopDay: (int) sd;
-(int) getReQuadRound;
-(int) getCarniecycle;
-(id) getSpecies;
-(id) getSites;

// This only hangs critters in site 1 for now, use site1's mcr if "where"=nil.
-hangCritters: (Critter*) grandma Num: (int) num 
         Root: (Quad*) root Where: (Mcr*) where;

-hangSites;
-onSeason;	// call Site-onSeason and Species-onSeason every seasonth time

-buildObjects;
-buildActions;
-activateIn: (id) swarmContext;
-resetClear;

-doStats;
-runFitness: (double*) fit;
-(BOOL) extinctCheck;		// any species gone extinct?

-printv;
-println;	// prints one line for each thing it prints...
-printCensus;
-dump; 		// executable

@end
