// vfb, 960116: rewrite.

// vfb, orig, 951020.

#import "random.h"
#import "Critter.h"
#import "Grazer.h"
#import "LandRoot.h"
#import "GeckoControlSwarm.h"
#import "Stat.h"
#import <stdlib.h>
#import <math.h>
#import <simtools.h>

@implementation Grazer

Grazer* recycleGrazers = 0;
Mcr* huntwin = nil;
Mcr* feedwin = nil;

-recycle
{
    [species tallyDeath: self];
    
    hook = recycleGrazers;
    recycleGrazers = self;
    [species decrementCensus];
    return nil;
}

-recycleClear
{
    Grazer* togo;
    while (recycleGrazers) {
        togo = recycleGrazers;
        recycleGrazers = recycleGrazers->hook;
        [togo->energy drop];
        [togo drop];
    }
    return self;
}


-reuse
{
    Grazer* result = recycleGrazers;
    recycleGrazers = recycleGrazers->hook;
    result->hook   = nil;
    return result;
}

-init
{
    double angle = 2 * M_PI * EvenRand();
    [super init];
    wannax = cos(angle);
    wannay = sin(angle);
    growrate = 0.0;
    return self;
}

+create 
{
    id result;
    if (!huntwin) huntwin = [Mcr create];
    if (!feedwin) feedwin = [Mcr create]; 
    
    if (recycleGrazers) {
        result = [recycleGrazers reuse];
    } else {
        result = [Grazer createBegin: globalZone];
        [result allocEnergy];
    }
    [result init];
    
    return result;					
}

-copymom
    // create an agent matching mom in all heritable respects
{
    Grazer* baby;
    Quad* newQuad;

    baby = [Grazer create];
    baby->species   = species;
    [energy copyTo: baby->energy];
    baby->color     = color;
    [species incrementCensus];
    [species tallyBirth: self];
    baby->breedsize = breedsize;
    baby->kidsize   = kidsize;
    baby->minsize   = minsize;
    baby->scaleradius = scaleradius;
    baby->scalev = scalev;
    baby->growrate  = growrate;
    baby->lastmealx = lastmealx;
    baby->lastmealy = lastmealy;
    baby->veer      = veer;
    baby->moverule  = moverule;

    newQuad = [Quad create];
    [newQuad setData: baby Type: critter];
    [newQuad setSibling: nil];
    [newQuad setChild: nil];
    [newQuad setParent: nil];
    [baby setQuad: newQuad];

    return baby;
}

-addSpeciesParams
{
    [super addSpeciesParams];

    [species addParam: "GrowRate" Vary: 0 DValue: growrate 
          Message: M(setGrowRate:) Owner: self];
    [species addParam: "Veer" Vary: 0 DValue: veer 
          Message: M(setVeer:) Owner: self];

    return self;   
}

-liveGrandma: (Species*) s
{
    
    [energy addConvertFrom: carbos To: proteins Rate: (1.0/3.0)];
    [energy addTax: proteins Rate: .365 Exponent: .75]; // .08
    breedsize   = 6.65;
    kidsize     = 3.2;
    minsize     = 3.0;
    scaleradius = 1.0;
    scalev = 1.0/M_PI;
    growrate    = .0306;
    veer        = 3.77; // M_PI;
    
    species = s;
    [self setColor: [species getColor]];
    [self addSpeciesParams];
    
    return self;
}

-resetGrowRate: (GRate) gr Veer: (double) v
    // use with Critter-resetBreedSize to reset grandma
{
    growrate = gr;
    veer = v;
    return self;
}
-setGrowRate: (GRate) gr {growrate = gr; return self;}
-setVeer: (double) v {veer = v; return self;}
-setConvertRate: (GRate) nr
    // cheats rather than parse what was there before.
{
    [energy setConvertFrom: carbos To: proteins Rate: nr];
    return self;
}

-(GCoord) getMoveRadius 
{ 
    if (moverule == 2)
        return ( radius * exp( EvenRand() ) ); 
    else if (moverule == 3)
        return ( 0.75 * radius * exp( EvenRand() ) ); 
    else if (moverule == 4)
        return ( 0.75 * radius); 
    else if (moverule == 5)
        return ( 0.5 * radius * exp( EvenRand() ) ); 
    else // includes moverule 1
        return ( radius );  // was for all before
}

-setMoveRule: (int) mr
{
    moverule = mr;
    return self;
}

-move
{
    GCoord scratchx, scratchy;
    GCoord delta = [self getMoveRadius];
    if (delta == 0.0)
        return self;
    
    [self getDirectionX: &scratchx Y: &scratchy]; // ignore, because reset wanna's
    x += fabs(delta) * wannax;
    y += fabs(delta) * wannay;
    // wannax = wannay = 0.0;	// done with this intention.
    [self getMcr: myquad->mcr];
    [myquad growMcr];
    return self;
}

-(Critter*) bestpreyMin: (GCoord) minradius
    // Choose a plant--kinda arbitrary logic, so localize.
    // Don't care how big plant is, so long as it's bigger than minradius
    // but will return plant smaller than minradius, if can't find bigger.
{
    Quad * pq;
    Critter* plant, *bestplant, *anyplant;
    Quad* plantRoot = [[geckoControl getModelSwarm] getPlantRoot];
    
    if (!feedwin) feedwin = [Mcr create];
    [feedwin circleToSquare: x Y: y Radius: radius];
    pq = [plantRoot pickWin: feedwin Root: plantRoot Invisibles: 0];

    bestplant = nil;	// any plant that matches minradius
    anyplant  = nil;	// some plant that doesn't match minradius
    
    if (pq) do {
        plant = pq->data;
        if ([self attacks: plant]) {
            if ([plant getRadius] > minradius) {
                if (EvenRand() > 0.5)
                    bestplant = plant;
                else if (!bestplant)
                    bestplant = plant;
            } else if (!(bestplant)) {
                if (!(anyplant))
                    anyplant = plant;
                else if (EvenRand() > 0.5)
                    anyplant = plant;
            }
        }
        pq = [pq pickWin: feedwin Root: plantRoot Invisibles: 0];
    } while ((pq));

    if (bestplant) 
        return bestplant;
    else
        return anyplant;
}

-hunt
    // How much I want to eat depends on -my- size & tax debt & growthrate.
    // Hunt for any plant that satisfies that desire, settle for smaller
    // one if can't have what we want.
{
    Critter* plant;
    ResType me = [energy getIam];
    ResType meat = [energy getIeat];
    ResQuant mybite;
    GCoord   mybiter;
    double angle, veerangle;
    ResQuant myvol;
    
    myvol   = [energy getStores: me];
    mybite  = ([energy getTaxRate: me] * pow(myvol, [energy getTaxExponent: me]))
              + growrate * myvol;
    mybite /= [energy getConvertFrom: [energy getIeat] To: me];
//printf("gopper myvol = %f, mybite = %f, tax = %f, exponent = %f\n",
//         myvol, mybite, [energy getTaxRate: me], [energy getTaxExponent: me]);
     mybiter = volume2radius(mybite, 1.0, scalev);

    plant = [self bestpreyMin: mybiter];
    
    if ((plant)) {
        if ([plant getRadius] > mybiter) {
            [plant stealRes: meat Amount: mybite];
        } else {
            mybite = [plant stealAllRes: meat];
        }
        [plant resetRadius];  // since another may choose to eat him.
        [self feedRes: meat Amount: mybite];

        // move at semi-random, save last place we ate
        // angle = (M_PI / 8.0) * EvenRand();
        // if (EvenRand() > 0.5) angle = -angle;
        angle = [self getDirectionAngle];
        veerangle = (veer/2.0) * EvenRand();
        if (EvenRand() < 0.5) 
            veerangle = - veerangle;
        angle += veerangle;
        [self setDirectionX: cos(angle) Y: sin(angle)];
        
        lastmealx = plant->x;
        lastmealy = plant->y;

    } else {
        // back up toward last place we found food
        [self setDirectionX: lastmealx - x Y: lastmealy - y];
    } 
    
    return self;    
}

-printv
{
    [super printv];
    printLogD("growrate = %7f, ", growrate);
    printLogD("veer = %7lf, ", veer);
    printLog("moverule = %d\n\n", moverule);
    return self;
}

@end
