// Copyright (C) 1996 Yale Institute for Biospheric Studies and Ginger Booth.
// This library is distributed without any warranty; without even the
// implied warranty of merchantability or fitness for a particular purpose.
// See file LICENSE for details and terms of copying.

// vfb, original, 960709: 
// 960829:  never did get this to work.  Can't get around bugs in gcc. 

#import "Base.h"
#import "GeckoModelSwarm.h"
#import "GeckoDisplaySwarm.h"
#import "GeckoControlSwarm.h"
#import <time.h>
#import <simtools.h>
#import <activity.h>
#import <tkobjc.h>
#import "QuadDisplay.h"
#import "OtherDisplay.h"

#define BUFSIZE 256
#define TOKSIZE 50
FILE* pf;              // protocol file
int  line;	       // the line number being parsed
int  lineposn;	       // character in the line being parsed
int  execposn;	       // character in the command being parsed
int  toklen;	       // token length
char linebuf[BUFSIZE]; // the line being parsed
char execbuf[BUFSIZE]; // an Obj-C command to execute, we think...
char token[TOKSIZE];   // current token (when useful)

typedef enum { tokError, tokEOF, tokComment, 
               tokSemicolon,    // ignored
               tokLSquare,      // '[' starts executable statement
               tokRSquare,      // ']' ends executable statement
               tokObj,          // who to pass message to
               tokMessage,      // message with no arguments
               tokArgLabel,     // message: 
               tokArg, 		// could be value or parm
               tokParm, tokVal  // specifically is value or parm
} tokType;

typedef enum { numberNot,
               numberInt,
               numberDouble,    // cannot tell whether double or float
               numberFloat      // according to somebody....default double
} numberType;

tokType expectTok;

void interpError(char* message, char* buf, int bufposn)
{
    int i;
    if (buf != linebuf)
        printBoth("%s\n", buf);
    for (i=0; i<(bufposn-1); i++)
        printBoth(" ");
    printBoth("^ %s\n", message);
    [InvalidArgument raiseEvent: "Error reading protocol file.\n"];
}

int nextline()
{
    if (!fgets(linebuf, BUFSIZE, pf))
        return 0;
    printBoth("%s\n", linebuf);
    line++;
    lineposn = 0;
    return 1;
}

numberType isnumber(char* token, int toklen)
{
    int i;
    int dotcount   = 0;
    int minuscount = 0;
    int digitcount = 0;

    token[toklen] = '\0';
    
    for (i=0; i<toklen; i++) {
        if (token[i] == '.') {
            if (dotcount)
                return numberNot;
            else
                dotcount++;            

        } else if (token[i] == '-') {
            if ((digitcount) || (minuscount))
                return numberNot;
            else
                minuscount++;

        } else if (isdigit(token[i])){
            digitcount++;

        } else {
            return numberNot;
        }
    }
    if (dotcount)
        return numberDouble;
    else if (digitcount)
        return numberInt;

    return numberNot;
}

int findnonblank()
{
    do {
        while (isspace(linebuf[lineposn])) {
            lineposn++;
            if (linebuf[lineposn] == '\0')
                if (!nextline()) 
                    return 0;
        }
        if (linebuf[lineposn] == '\0')
            if (!nextline()) 
                return 0;
    } while (isspace(linebuf[lineposn]));
    return 1;
}

tokType gettoken()
{
    int i;
    toklen = 0;
    
    if (!findnonblank()) 
        return tokEOF;

    // first let's look at the most obvious punctuation....    
    if ((linebuf[lineposn] == '/') && (linebuf[lineposn+1] == '/')) {
        lineposn += 2;
        return tokComment;
    }
    if (linebuf[lineposn] == ';') {
        lineposn += 1;
        return tokSemicolon;
    }
    if (linebuf[lineposn] == '[') {
        if (expectTok != tokLSquare) {
            interpError("Nested messages not allowed.", linebuf, lineposn);
            return tokError;
        } else {
            lineposn += 1;
            toklen = 1;
            token[0] = '[';
            return tokLSquare;
        }
    }
    if (linebuf[lineposn] == ']') {
        if (execposn == 0) {
            interpError("Unexpected ']'.", linebuf, lineposn);
            return tokError;
        } else {
            lineposn += 1;
            toklen = 1;
            token[0] = ']';
            return tokRSquare;
        }
    }
    
    // OK, it's a wordy token....
    do {
        token[toklen] = linebuf[lineposn];
        toklen++;
        lineposn++;
    } while ( ( !isspace(linebuf[lineposn]) ) && (linebuf[lineposn] != ']') );
    
    // see if it's a number or a named-thing...
    if (isnumber(token, toklen)) {
        if (expectTok != tokArg) {
            interpError("Unexpected number.", linebuf, lineposn);
            return tokError;
        } else {
            return tokVal;
        }
    }  
    // well, strings can be values, too, I guess.  Only "delimited" strings.
    if ((token[0]=='"') && (token[toklen-1] == '"'))
        return tokVal;
    
    // must be wordy....one of the : ones?
    if (token[toklen-1] == ':') {
        if ((expectTok == tokMessage) || (expectTok == tokArgLabel))
            return tokArgLabel;
        else {
            interpError("Misplaced ':'.", linebuf, lineposn);
            return tokError;
        }
    }
    // OK, it's just a named-thing.  Legal: alphanum _
    for (i=0; i<toklen; i++) {
        if (isalpha(token[i]) || (token[i] == '_')) {}
        else {
            interpError("Unexpected character.", linebuf, lineposn);
            return tokError;
        }
    }
    // OK, it's well-formed.  But what is it?
    if (expectTok == tokObj)
        return tokObj;
    else if (expectTok == tokArg)
        return tokParm;
    else if (expectTok == tokMessage)
        return tokMessage;

    interpError("Unexpected word.", linebuf, lineposn);
    return tokError;
} 

int getexec()
    // 0 return = no more to execute
{
    int i;
    tokType curTok;
    
    if (lineposn == 0)
        if (!nextline())
            return 0;
        
    // we expect to find a message to start 
    expectTok = tokLSquare; 
    execposn = 0;

    do { // this loop exits from the middle  
        curTok = gettoken();  
        switch (curTok) {
            // non-executable tokens
	    case tokError:    return 0; break;
	    case tokEOF:      return 0; break;
	    case tokComment:  if (!nextline()) return 0; break;
	    case tokSemicolon: break;
    
            // executable tokens -- set expectation for next token
	    case tokLSquare:  expectTok = tokObj;      break;
	    case tokObj:      expectTok = tokMessage;  break;
	    case tokMessage:  expectTok = tokRSquare;  break;
	    case tokArgLabel: expectTok = tokArg;      break;
	    case tokVal:      expectTok = tokArgLabel; break;
	    case tokParm:     expectTok = tokArgLabel; break;
	    // note gettoken returns tokVal or tokParm, never tokArg
	    
            // *** EXIT *** 
            // found end of executable
            //
	    case tokRSquare: 
	        if ((expectTok == tokArgLabel) || (expectTok == tokRSquare)) {
	            execbuf[execposn] = ']';
	            execposn++;
	            execbuf[execposn] = '\0';
	            return 1; 
	        } else {
                    interpError("Unexpected ']'.", linebuf, lineposn);
                    return tokError;
                }
	        break;
        
	    default: break;
        } 
        // copy token into execbuf, plus one blank
        for (i=0; i < toklen; i++) {
            execbuf[execposn] = token[i];
            execposn++;
        }
        if (toklen > 0) {
            execbuf[execposn] = ' ';
            execposn++;
        }
    } while (1);
    return 0;
}

void execute()
    // OK, a bit of error checking and formatting has already been
    // accomplished - in execbuf, we have:
    //
    // 		[ objname message ]
    // 		[ objname message: arg ] 
    // 	or	[ objname message: arg1 Arg2: arg2 ... ]
    //
    // Here we'll play monkey-see-monkey-do based on Swarm's MessageProbe
    // interface.
    //
    // NOTE: since we don't allow nested messages or assignment, we can
    // ignore all return types from message calls.  One drawback:  if class
    // has two messages that differ only in return type, there's nothing we
    // can do about it.    
{
    SEL selector;
    Critter* object;
    double arg1;
    id argh;
    IMP  mptr;
    
    execbuf[execposn] = '\0';
    printBoth("EXECUTE: %s, buflen = %d\n", execbuf, execposn);

    // using sel_get_any_uid instead of sel_get_any_typed_uid because we
    // have no idea what type it is, and can't find out.
    selector = sel_get_any_uid("setBreedSize:");
    printBoth("Selector d = %d, s = %s\n", selector, selector);
    
    // Oh, screw it.  Can't find the object id by name.  So...all objects
    // shall have been declared through registerObject.  Drat....
    // No, can't even do it that way.  Screw it.  We'll keep our own object id's.

    object = geckoControl->geckoModelSwarm->plantGrandma;
    printBoth("orig object, object =%d\n", object);
    [globalTkInterp registerObject: object 
        withName: "plantGrandma"];
    object = [globalTkInterp objectNamed: "plantGrandma"];
    printBoth("registered object, object =%d\n", object);
    printBoth("object says name =%s\n", [globalTkInterp nameForObject: object]);
    arg1 = 8.0;

    /*
    [globalTkInterp registerObject: arg1 withName: "argh"];
    argh = [globalTkInterp objectNamed: "argh"];
    printBoth("registered argh=%d\n", object);
    printBoth("argh says name =%s\n", [globalTkInterp nameForObject: argh]);
    mptr = objc_msg_lookup( object, selector );
    printBoth("call selector..., object= %d, selector = %d\n", object, selector);
    [object perform: selector 
        with: argh];
    printBoth("after perform\n");
    */
    
    // The following does not work, alas...grandma breedsize set to 0
    [globalTkInterp eval: "%s", execbuf] ;
    
    /* was thinking this over....
    // terminate the string in execbuf
    execbuf[execposn] = '\0';
    printBoth("EXECUTE: %s, buflen = %d\n", execbuf, execposn);
    selector = sel_get_any_typed_uid(execbuf) ;
    if (!selector) {
        interpError("Unable to execute.", execbuf, execposn);
    }
    
    if (!sel_get_type(selector)){
        interpError("Unable to execute.", execbuf, execposn);
    }
    selType = strdup(sel_get_type(selector)) ;
    argNum = get_number_of_arguments(selType) - 2 ;

    if(!argNum){
        switch(selType[0]){
            case _C_FLT:
            case _C_DBL: returnsDouble = 1 ; break ;                        
    
	    case _C_ID:
	    case _C_CLASS:
	    case _C_PTR:
	    case _C_UCHR:
            case _C_CHR:
            case _C_INT:
            case _C_UINT:
            case _C_CHARPTR:
            default:     returnsDouble = 0 ; break ;                        
        }
        argLabels = (char **) malloc(sizeof(char *)) ;
        argLabels[0] = probedMessage ;
    } else {
        probedSelector = 0 ;
        argLabels = (char **) malloc(argNum*sizeof(char *)) ;
        arguments = (char **) malloc(argNum*sizeof(char *)) ;
    
        for(i = 0 ; i < argNum ; i++){
            arguments[i] = NULL ;
            argLabels[i] = copy_to_nth_colon(probedMessage,i) ;
        }
    }
    */

}

void parsefile()
{
    line = 1;
    lineposn = 0;
    execposn = 0;
    while (getexec()) {
        execute();
    }
}

void interpret(char* protocolFile)
    // interpret an Obj-C-style protocol file, setting up a Gecko run
{
    time_t now;
    
    pf = fopen(protocolFile, "r");

    now = time(&now);
    printBoth("// %s\n", GeckoVersion); 
    printBoth("// %s", ctime(&now));
    printBoth("// Reading protocol file %s:\n", protocolFile);
    
    if (!pf) {
        [InvalidArgument raiseEvent: "Unable to open protocol file.\n"];
    }
    
    parsefile();
    
    printBoth("// Protocol file end.\n", protocolFile);
    
    fclose(pf);
}
