// The definition of a Land(scape) object. vfb rewrite, 951221.

#import "Base.h"
#import "ResourceQuad.h"
#import "LandRoot.h"
#import "random.h"
#import "Critter.h"
#import "GeckoControlSwarm.h"
#import <simtools.h>
#import <math.h>

@implementation LandRoot

-init 
{
    visible  	 = 1;
    color        = Black;
    type     	 = feed;	
    parent   	 = nil;
    sibling  	 = nil;
    child    	 = nil;
    mcr          = [[Mcr create] setNullMcr];
    startres     = [[ResList create] zero];
    reservoir    = [[ResList create] zero];
    taxTake      = [[ResList create] zero];
    convert      = 0;
    return self;
}

+create 
{
    id result = [[LandRoot createBegin: globalZone] createEnd];
    [result init];
    return result;					
}

-setConvertFrom: (ResType) rfrom To: (ResType) rto Rate: (GRate) newrate
    // not yet
{
    return self;
}

-setMcrLX: (GCoord) x1 LY: (GCoord) y1 UX: (GCoord) x2 UY: (GCoord) y2
{
    mcr->lx = x1;
    mcr->ux = x2;
    mcr->ly = y1;
    mcr->uy = y2;
    [self growMcr];
    
    return self;
}

-(ResQuant) feedCritter: (Critter*) eater Resource: (ResType) res
     Amount: (GRate) amount Area: (Mcr*) m;
{
    id site;
    ResQuant got = 0.0;
    
    site = [self pickWin: m Root: self Invisibles: 1];

    while (site) {
        if ([site produces: res]) {
            got += [site feedCritter: eater Amount: amount Area: m];
        }
        site = [site pickWin: m Root: self Invisibles: 1];
    }
    reservoir->list[res] -= got;
    
    return got;
}

-eatCompete: (Quad*) root
    // IF you don't want landRoot to use this algorithm, don't hang 
    // agents under plantRoot.  Then manage your eat/compete algorithm yourself
    // in the agents.

    // Feed the plantRoot tree of agents.  Tricky, in
    // that overlapping agents must share resources, as feed is an area
    // effect.  To do this:
    // 	a. Use each agent in turn as a pick window (eater).  Find its area.
    //  b. For each agent (competitor) picked with that window, subtract
    //     one half the intersecting area from the eater.
    //  c. When done picking competitors, we have the effective area of
    //     the eater.  Express this as a percent of the original.
    //  d. Send the FeedQuad the true eater mcr and %area, and request
    //     how much the eater gets fed.
    // Note:  this is not -correct-, in that if more than two agents
    // overlap, the effective area decreases faster than it should.  E.g.,
    // if three agents overlap, they should each get 1/3 of the resources
    // from the 3-way overlap, but in this scheme, they'll each get 1/4.
    // As of this writing, I weakly defend this as the price of competition,
    // and more strongly defend it as the most compute-effective way I could
    // think of to localize errors in an intuitively clear fashion.
    
    // This routine resets agent scratchpads and visibility across the tree,
    // as well as feeding the agents and debiting the landscape.
    
    // Note that competition is calculated separately for each resource.
    
{
    // eater and competitor can be considered two parsing pointers on tree
    ResType res;
    Quad* eater;
    Quad* competitor;
    Critter * edata, * cdata;
    GArea eaterarea;
    GArea eaterarea1;	// unmodified eater area.
    GArea contestedarea;
    GArea wx, wy;	// for calculating wannago direction
    int ecount;

static BOOL firsttime = 1;
static Mcr* feedwin;
static Mcr* eatersquare;
if (firsttime) {
    firsttime = 0;
    feedwin = [Mcr create];
    eatersquare = [Mcr create];
}

    // clip feeding area and tree area to each other    
    if ( [mcr intersectMcr: root->mcr Intersection: feedwin] == disjoint ) {
        return self;
    }
    
    // First, pick everyone in the area and set their scratchpads to zero.
    // All picking series go by same root.

    eater = [root pickWin: feedwin Root: root Invisibles: 0];
    if (eater == nil) {
        return self;
    }
    while (eater != nil) {
        edata = eater->data; 
        [edata setScratchpad: 0.0];
        eater = [eater pickWin: feedwin Root: root Invisibles: 0];
    }
        
    // now back to first eater
    eater  = [root pickWin: feedwin Root: root Invisibles: 0]; 
    ecount = 0;
    do {
        ecount++;
        edata = eater->data; 
        eater->visible = 0;

        // set eater's equivalent square now.
        [eatersquare circleToSquare: edata->x
                                  Y: edata->y
                                  Radius: edata->radius];
                                
        // subtract off results of prior picks
        eaterarea = eaterarea1 = [eatersquare areaMcr];
        eaterarea -= edata->scratchpad;
        

        // consider each competitor
        competitor = [root pickWin: eater->mcr Root: root Invisibles: 0];
        while (competitor) {
            if ([eater->data competesWith: competitor->data]) {
                cdata = competitor->data;
                contestedarea = [eatersquare crossCircles: edata->x
                                            Y1: 	  edata->y
                                            Radius1: 	  edata->radius
                                            X2:      cdata->x
                                            Y2:      cdata->y
                                            Radius2: cdata->radius];
    
                // NOTE:  because we're calculating debits on competitors as
                // well as debits on the eater, we must go on to the bitter end
                // even if eater's area becomes negative.  As it may because of
                // the error source noted above.
    
                contestedarea /= 2.0;
                eaterarea -= contestedarea;
                [cdata setScratchpad: (cdata->scratchpad + contestedarea)];
                
                // now, let's set wannas -- away from competitor
                wx = cdata->x - edata->x;
                wy = cdata->y - edata->y;
                [edata addDirectionX: -wx Y: -wy Weight: contestedarea];
                [cdata addDirectionX:  wx Y:  wy Weight: contestedarea];
                
            }
            competitor = [competitor pickWin: eatersquare Root: root Invisibles: 0];
        }
        
        if (eaterarea > 0.0) {
            for (res=1; res <=NumRes; res++) {
                if ([eater->data eats: res]) {
                    [self feedCritter: edata Resource: res
                        Amount: (eaterarea/eaterarea1) Area: eatersquare];
                }
            }       
        } 
        
        eater = [eater pickWin: feedwin Root: root Invisibles: 0];
    } while (eater != nil);
    

    // Now for cleanup.  Must set our invisibles back to visible. 
    eater = [root pickWin: feedwin Root: root Invisibles: 1];
    while (eater) {
        eater->visible = 1;
        eater = [eater pickWin: feedwin Root: root Invisibles: 1];
    }
    
    return self;
} 

// If you need to talk to the resource quad directly...returns first it
// finds that produces this resource over that area.  If trying to find for
// an agent, pass myquad->mcr as the area (or its equiv-area square).
-(id) firstQuadProducing: (ResType) res Over: (Mcr*) m
{
    id site;
    
    site = [self pickWin: m Root: self Invisibles: 1];

    while (site) {
        if ([site produces: res])
            return site;
        site = [site pickWin: m Root: self Invisibles: 1];
    }
    
    return 0;
}

-(id) nextQuadProducing: (ResType) res Over: (Mcr*) m Last: (id) resumefrom
{
    id site;
    
    site = [resumefrom pickWin: m Root: self Invisibles: 1];

    while (site) {
        if ([site produces: res])
            return site;
        site = [site pickWin: m Root: self Invisibles: 1];
    }
    
    return 0;
}

-feedTree: (Quad*) root
    // Should be called on landscape quad with sessile argument. 
{
    if (root->type != sessile)
        return nil;
    
    // sessile eaters compete for food
    
    [self eatCompete: root];

    return self;
} 

-feedTree: (Quad*) root On: (int) roundth
{
    if (getCurrentTime() % roundth == 0)
        [self feedTree: root];
    return self;
}


-payTaxX: (GCoord) x Y: (GCoord) y Resource: (ResType) res
  Amount: (ResQuant) amount
  
  // Later, when we convert, we use the coordinate to forward the
  // divvy up the result to the appropriate feeding quads.  For
  // now, we just tally up.
{
    id site;
    taxTake->list[res] += amount;
    
    // any takers? all goes to first volunteer
    site = [self pickWin: mcr Root: self Invisibles: 1];

    while (site) {
        if ([site produces: res] && [[site getMcr] ptInsideX: x Y: y]) {
            [site addResourceAmount: amount X: x Y: y];
            return self;
        }
        site = [site pickWin: mcr Root: self Invisibles: 1];
    }
    
    return self;
}

  
-payTaxX: (GCoord) x Y: (GCoord) y 
          List: (ResList*) list;
  // Tricky.  Need to split between all sites crossing x,y that produce res.
{
    int i;
    for (i=1; i <= NumRes; i++)
        if (list->list[i] > 0.0)
            [self payTaxX: x Y: y Resource: i Amount: list->list[i]];
    return self;
}

-(ResQuant) taxesPaid: (ResType) res
{
    return taxTake->list[res];
}	

-(ResList*) taxesPaid
{
    ResList* totals = [[ResList create] zero];
    
    [totals add: taxTake];
    return totals;
}	

-zeroTaxes {[taxTake zero]; return self;}

-stepResourcesOn: (int) roundth
{
    if ((getCurrentTime() % roundth) == 0)
        [self stepResources];
    return self;
} 

-stepResources
    // call on landscape collector
{
    ResourceQuad* asite;
    Quad* newborns;
    int reQuadRound  = [[geckoControl getModelSwarm] getReQuadRound];
    Quad* mobileRoot = [[geckoControl getModelSwarm] getMobileRoot];
    Quad* sessileRoot = [[geckoControl getModelSwarm] getPlantRoot];

    // squash and rezone trees regularly
    if ((getCurrentTime() > 0) && (getCurrentTime() % reQuadRound == 0)) {
        [sessileRoot squashQuad];
        [sessileRoot subQuad];
        [mobileRoot squashQuad];
        [mobileRoot subQuad];
    } else {
        newborns = [sessileRoot findNewbies];
        [newborns squashQuad];
        [newborns subQuad];
    } 
    
    [startres zero];

    asite = (ResourceQuad*) [self pickWin: mcr Root: self Invisibles: 1];
    while (asite) {
        [asite stepResources];
        startres->list[asite->resource] += asite->startres;
        asite = (ResourceQuad*) [asite pickWin: mcr Root: self Invisibles: 1];
    }

    [reservoir zero];
    [reservoir add: startres];
    
    return self;
}

-(ResList*) curFeed
    // feedfunc*area, whole tree
{
    ResList* totals = [[ResList create] zero];
    [totals add: reservoir];
    return totals;
}

-(ResList*) fedOut
    // startres - reservoir, whole tree
{
    ResList* totals = [[ResList create] zero];
    [totals add: reservoir];
    [totals subtract: startres];
    return totals;
}

-(ResList*) fedOutP
    // percent version of fedOut
{
    int i;
    ResList* totals = [[ResList create] zero];
    
    for (i=1; i <= NumRes; i++) {
        if (startres->list[i] != 0.0)
            totals->list[i] = reservoir->list[i] / startres->list[i];
    }
    
    return totals;
}

-print
{
    printLog("landRoot ");
    [mcr print];
    return self;
}

-println
{
    [self print];
    printLog("\n");
    return self;
}

-printv
{
    [self println];
    printLog("    startres : "); [startres println];
    printLog("    reservoir: "); [reservoir println];
    printLog("    taxTake  : "); [taxTake println];
    return self;
}
    
@end
