// vfb, orig, 951221:

#import <math.h>
#import "Res.h"
#import "Metab.h"
#import "Stat.h"

@implementation Metab

int metabcount = 0;

-zero
{
    int i, j;

    [gullet zero];
    [stores zero];
    if (convert) {
        for (i=0; i<=NumRes; i++) 
            for (j=0; j<=NumRes; j++) 
                convert[i][j] = 0.0;   
    }
    for (i=0; i<=NumRes; i++) {
        taxrate[i] = 0.0;   
        taxexp[i]  = 0.0;   
    }
    ieat 	   = 0;
    iam 	   = 0;
    efficiency 	   = 0.0; 
    return self;    
}

-init
{
    gullet  = [ResList create];
    stores  = [ResList create];
    taxrate = (GRate*) [globalZone alloc: (NumRes+1) * sizeof(GRate)];
    taxexp  = (GRate*) [globalZone alloc: (NumRes+1) * sizeof(GRate)];

    // not going to allocate convert matrix until we need one.
    convert = 0;

    [self zero];    
    return self;    
}

+create
{
    id result = [[Metab createBegin: globalZone] createEnd];
    metabcount++;
    [result init];
    return result;					
}

-addConvertFrom: (ResType) from To: (ResType) to Rate: (GRate) rate
{
    int i, j;
    
      //
     // Is this the first (only) conversion?
    // 
    if ( (!ieat) && (!convert) ){
        ieat       = from;
        iam        = to;
        efficiency = rate;
      //  
     // This is not the first (only) conversion.
    // 
    } else {
          //
         // Do we have a convert matrix going yet?
        // 
        if (!convert) {
              //
             // No, so allocate one.
            // 
            convert = (GRate**) [globalZone alloc: (NumRes+1) * sizeof(GRate*)];
            for (i=0; i<=NumRes; i++)
                convert[i] = (GRate*) [globalZone alloc: (NumRes+1) * sizeof(GRate)];
            for (i=0; i<=NumRes; i++)
                for (j=0; j<=NumRes; j++)
                    convert[i][j] = 0.0;
              //  
             // And transfer prior simple conversion into convert matrix.
            // 
            convert[ieat][iam] = efficiency;
            ieat = iam = nullres;
            efficiency = 0.0;
        }
          //
         // Add the current conversion into convert matrix. 
        //
        convert[from][to] = rate;       
    }
    return self;
}

-setConvertFrom: (ResType) rfrom To: (ResType) rto Rate: (GRate) newrate
{
    // Is this the first (only) conversion?
    if ( (!ieat) && (!convert) ) {
        [self addConvertFrom: rfrom To: rto Rate: newrate];
        return self;

    // This is not the first (only) conversion.
    } else {
        // Do we have a convert matrix going yet?
        if (!convert) {
            // Replace sole conversion?
            if ( (ieat == rfrom) && (iam == rto) ) {
                efficiency = newrate;
                return self;
            // No, second conversion needed--use add routine.
            } else {
                [self addConvertFrom: rfrom To: rto Rate: newrate];
                return self;
            }
        // Yes, have matrix - simply stuff this in. 
        } else {
            convert[rfrom][rto] = newrate;       
        }  
    }
    return self;
}

-addTax: (ResType) r Rate: (GRate) rate Exponent: (GRate) exponent
{
    taxrate[r] = rate;
    taxexp[r] = exponent;
    return self;
}

-replaceTax: (ResType) r Rate: (GRate) rate Exponent: (GRate) exponent
{
    [self addTax: r Rate: rate Exponent: exponent];
    return self;
}

-(GRate) getTaxRate: (ResType) r { return (taxrate[r]); }

-(GRate) getTaxExponent: (ResType) r { return (taxexp[r]); }

-(GRate) getConvertFrom: (ResType) rfrom To: (ResType) rto
{
    if ( (ieat == rfrom) && (iam == rto) ) 
        return efficiency;
    else if (ieat)
        return 0.0;
    else if (convert) 
        return convert[rfrom][rto];
    else
        return 0.0;
}

-copyTo: (Metab*) c
    // never copy gullet or stores--only the rules are transferable.    
{
    int i, j;

    for (i=0; i<= NumRes; i++) {
        c->taxrate[i] = taxrate[i];
        c->taxexp[i]  = taxexp[i];
    }
    c->ieat           = ieat;
    c->iam            = iam;
    c->efficiency     = efficiency;

    if (convert) {
        if (!c->convert) {
            c->convert = (GRate**) [globalZone alloc: (NumRes+1) * sizeof(GRate*)];
            for (i=0; i<=NumRes; i++)
                c->convert[i] = (GRate*) [globalZone alloc: (NumRes+1) * sizeof(GRate)];
            for (i=0; i<=NumRes; i++)
                for (j=0; j<=NumRes; j++)
                    c->convert[i][j] = 0.0;
        }
        for (i=0; i<=NumRes; i++)
            for (j=0; j<=NumRes; j++)
                c->convert[i][j] = convert[i][j];
    }
    return self;    
}

-newcopy
    // returns pointer to newly created copy
{
    Metab* copy = [Metab create];
    [self copyTo: copy];
    return copy;
}

-metabolize
    // metabolize gullet -> stores 
{
    int i, j;
    
    // simple
    if (ieat) {
        stores->list[iam] += gullet->list[ieat] * efficiency;
        gullet->list[ieat] = 0.0;
        
    // complex
    } else {
        for (i=1; i<=NumRes; i++) {
            for (j=1; j<= NumRes; j++) {
                if (convert[i][j] > 0) {
                    stores->list[j] += gullet->list[i] * convert[i][j];
                }
            }
        }
        for (i=1; i<=NumRes; i++) {
            gullet->list[i] = 0.0;
        }
    }
    return self;
}

-(ResList*) taxes
    // debit taxes and return list to pay
{
    int i;
    ResList* topay;
    
    topay = [[ResList create] zero];
    
    if (iam) {
        topay->list[iam] = taxrate[iam] * pow(stores->list[iam], taxexp[iam]);
    } else {
        for (i=1; i <= NumRes; i++) {
            topay->list[i] = taxrate[i] * pow(stores->list[i], taxexp[i]);
        } 
    }
    
    [stores subtract: topay];

    return topay;
}

-(ResQuant) getStores: (ResType) res { return stores->list[res]; } 

-(ResList*) getStores { return stores; } 

-setStores: (ResList*) list
{
    int i;
    for (i=0; i<= NumRes; i++)
        stores->list[i] = list->list[i];
    return self;
} 

-setStores: (ResType) what Amount: (ResQuant) quant
{
    stores->list[what] = quant;
    return self;
}

-(ResType) getIeat { return ieat; }

-(ResType) getIam  { return iam; }

-(BOOL) doIeat: (ResType) what
{
    int i;
    
    if (ieat) {
        if (ieat == what)
            return 1;
    } else {
        if (!convert)
            return 0;
        for (i=1; i<= NumRes; i++)
            if (convert[what][i] > 0.0)
                return 1;
    }
    return 0;
}

-(BOOL) amI: (ResType) what
{
    int i;
    
    if (iam) {
        if (iam == what)
            return 1;
    } else {
        if (!convert)
            return 0;
        for (i=1; i<= NumRes; i++)
            if (convert[i][what] > 0.0)
                return 1;
    }
    return 0;    
}

-feed: (ResList*) dinner
{
    int i;
    for (i=1; i<= NumRes; i++)
        gullet->list[i] += dinner->list[i];
    return self;
}

-feedRes: (ResType) res Amount: (ResQuant) quant
{
    gullet->list[res] += quant;
    return self;
}

-debit: (ResList*) payout
{
    int i;
    for (i=1; i<= NumRes; i++)
        stores->list[i] -= payout->list[i];
    return self;
}

-print
{
    if (iam) {
        printLog("converts %d to %d at ", ieat, iam);
        printLogD("%f%%", efficiency);
    } else {
        printLog("(conversion matrix)");
    }
    return self;
}

-println
{
    [self print]; 
    printLog("\n");
    return self;
}

-printv
{
    int i, j;
    
    printLog("Gullet:  "); [gullet println];
    printLog("Stores:  "); [stores println];
    printLog("Tax Rates    :  ");
    for (i=0; i <= NumRes; i++) printLogD("%7.3f ", taxrate[i]);
    printLog("\n");
    printLog("Tax Exponents:  ");
    for (i=0; i <= NumRes; i++) printLogD("%7.3f ", taxexp[i]);
    printLog("\n");
    
    if (iam)
        [self println];
    else {
        printLog("Conversion matrix, from vertical, to horizontal:\n");
        for (i=0; i <= NumRes; i++) {
            printLog("    ");
            for (j=0; j <= NumRes; j++) {
                printLogD("%7.3f ", convert[i][j]);
            }
            printLog("\n");
        }
    }
    return self;
}

@end
