// vfb, orig, 960909:

#import "Base.h"
#import "Param.h"
#import "Stat.h"

#define PARSEBUFSIZE 250
// parser "word" might be composite, as with Offset Gobbledygook:Pumpkin
#define PARSEWORDSIZE 100
char parseBuffer[PARSEBUFSIZE];
char wordBuffer[PARSEWORDSIZE];
int parsep;
int wordp;

@implementation Param;

-apply
{
    IMP  mptr;
    mptr = objc_msg_lookup( owner, message );
    if ( ! mptr )  {
        geckoError(InvalidCombination,
            "Parameter message error, applying '%s' to '%s'.\n",
            sel_get_name(message), [owner getName]);  
        return self;
    }
    switch (paramType) {
        case IntParam:    return mptr( owner, message, ival); break;
        case DoubleParam: return mptr( owner, message, dval); break;
        case StringParam: return mptr( owner, message, sval); break;
        default : break;
    }
    return self;
}

-copymomZone: (id) z
    // note: we never copy strings, just addresses
{
    Param* child    = [Param create: z];  
    child->label    = label;  
    child->paramType= paramType;  
    child->varyFlag = varyFlag;  
    child->ival     = ival;  
    child->dval     = dval;  
    child->sval     = sval;  
    child->message  = message;  
    child->owner    = owner;  
    
    return child;
}

-setLabel: (char*) l
    Vary: (BOOL) vflag
    Message: (SEL) m
    Owner: (id) o
{
    label    = l;
    varyFlag = vflag;
    message  = m;
    owner    = o;
    return self;
}

-setIValue: (int) i
{
    switch (paramType) {
        case IntParam: ival = i; break;
        case DoubleParam: dval = (double) i; break;
        default: break;
    }
    [self apply];
    return self;
}

-setDValue: (double) d 
{
    switch (paramType) {
        case IntParam: ival = (int) d; break;
        case DoubleParam: dval = d; break;
        default : break;
    }
    [self apply];
    return self;
}

-(double) getPartDValue    { return [self getDValue]; }
-setPartDValue: (double) d { return [self setDValue: d]; }
-(int) getPartIValue       { return [self getIValue]; }
-setPartIValue: (int) i    { return [self setIValue: i]; }

-setSValue: (char*) s
{
    sval = s;
    [self apply];
    return self;
}

-(id) getOwner { return owner; }
-(ParamType) getParamType { return paramType;}
-(BOOL) getVary { return varyFlag;}
-(int) getIValue { return ival; }
-(double) getDValue { return dval; }
-(char*) getName { return label; }
-(char*) getSValue { return sval; }

-print
{
    switch (paramType) {
        case IntParam:    printLog("%s:%d", label, ival); break;
        case DoubleParam: printLog("%s:", label);
                          printLogD("%f", dval); break;
        case StringParam: printLog("%s:%s", label, sval); break;
        default:          printLog("%s", label); break;
    }
    return self;
}

-println {[self print]; printLog("\n"); return self;}

-printv
{
    printLog("    "); 
    [self print];
    printLog("  varies:");
    if (varyFlag) printLog("1"); else printLog("0");
    printLog("  type:");
    switch (paramType) {
        case IntParam:    printLog("int value: %d", ival); break;
        case DoubleParam: printLogD("double value: %f", dval); break;
        case StringParam: printLog("string value: %s", sval); break;
        default:          printLog("label"); break;
    }

    printLog("\n");
    return self;
}

-dump
    // dump is reloadable, print is for looks
{
    printBoth("\t%s\t", label);
    if (strlen(label) < 8) 
        printBoth("\t");
    if (paramType == IntParam) 
        printBoth("%d\t", ival);
    else if (paramType == DoubleParam) 
        printBothD("%lf", dval);
    else if (paramType == StringParam) 
        printBoth("%s", sval);
    if (varyFlag)
        printBoth("\tVary");
    printBoth("\n");
    return self;
}

@end // Param
    
@implementation OffsetParam;

-apply
{
    IMP  mptr;
    mptr = objc_msg_lookup( owner, message );
    if ( ! mptr )  {
        geckoError(InvalidCombination,
            "Parameter message error applying '%s' to '%s'.\n",
            sel_get_name(message), [owner getName]);  
        return self;
    }
    switch (paramType) {
        case IntParam:    
            return mptr( owner, message, ival + [offsetFrom getIValue]); 
            break;
        case DoubleParam: 
            return mptr( owner, message, dval + [offsetFrom getDValue]); 
            break;
        default : break;
    }
    return self;
}

-copymomZone: (id) z
    // note: we never copy strings, just addresses.
{
    OffsetParam* child = [OffsetParam create: z];  
    child->label      = label;  
    child->paramType  = paramType;  
    child->varyFlag   = varyFlag;  
    child->ival       = ival;  
    child->dval       = dval;  
    child->sval       = sval;  
    child->message    = message;  
    child->owner      = owner;  
    child->offsetFrom = offsetFrom;  
    
    return child;
}

-(int) getOffsetIValue { return [offsetFrom getIValue]; }
-(double) getOffsetDValue { return [offsetFrom getDValue]; }

-(double) getPartDValue {return dval;}
-(int) getPartIValue {return ival;}

-setPartDValue: (double) d { return [super setDValue: d]; }
-setPartIValue: (int) i    { return [super setIValue: i]; }

-setIValue: (int) i {ival = i - [offsetFrom getIValue]; return self;}
-(int) getIValue {return [offsetFrom getIValue] + ival;}

-setDValue: (double) d {dval = d - [offsetFrom getDValue]; return self;}
-(double) getDValue {return [offsetFrom getDValue] + dval;}

-setOffsetFrom: (id) op {offsetFrom = op; return self;}
-(id) getOffsetFrom {return offsetFrom;}
-print
{
    switch (paramType) {
        case IntParam:    
            printLog("%s is %s + %d", label, [offsetFrom getName], ival); 
            break;
        case DoubleParam: printLog("%s is %s + ", label, [offsetFrom getName]);
                          printLogD("%f", dval); break;
        default:          printLog("%s", label); break;
    }
    return self;
}

-println {[self print]; printLog("\n"); return self;}

-printv
{
    printLog("    "); 
    [self print];
    printLog("  varies:");
    if (varyFlag) printLog("1"); else printLog("0");
    printLog("  type:");
    switch (paramType) {
        case IntParam:    printLog("int value: %d", ival); break;
        case DoubleParam: printLogD("double value: %f", dval); break;
        default:          printLog("label"); break;
    }
    printLog(", offset from %s\n", [offsetFrom getName]);
    return self;
}

-dump
    // dump is reloadable, print is for looks
{
    printBoth("\t%s\t", label);
    if (strlen(label) < 8) 
        printBoth("\t");
    if (paramType == IntParam) 
        printBoth("%d\t", ival);
    else if (paramType == DoubleParam) 
        printBothD("%lf", dval);
    else if (paramType == StringParam) 
        printBoth("%s", sval);
    if (varyFlag)
        printBoth("\tVary");
    printBoth(" Offset %s:%s\n", [offsetFrom->owner getName], [offsetFrom getName]);
    return self;
}

@end // OffsetParam

@implementation Parameterized

id listOfParameterizeds = nil;

-init
{
    params = [List create: [self getZone]];
    if (listOfParameterizeds == nil) 
        listOfParameterizeds = [List create: [self getZone]];
    [listOfParameterizeds addLast: self];
        
    return self;
}

+create
{
    id result = [[Parameterized createBegin: globalZone] createEnd];
    result = [result init];
    return result;					
}

-addLabelParam: (char*) label
{
    Param* param;
    
    param          = [Param create: [self getZone]];
    param->label   = label;
    param->paramType = LabelParam;

    [params addLast: param];
    
    return self;
}

-addParam: (char*) label 
    Vary: (BOOL) vflag
    DValue: (double) d
    Message: (SEL) m
    Owner: (id) o
{
    Param * param;
    
    if (![o respondsTo: m]) {
        geckoError(InvalidCombination,
            "Parameter message error applying '%s' to '%s'.\n",
            sel_get_name(m), [o getName]);  
        return self;
    } 
    
    param          = [Param create: [self getZone]];
    param->paramType = DoubleParam;
    param->dval    = d;
    [param setLabel: label Vary: vflag Message: m Owner: o];
    [params addLast: param];
    
    return self;
}

-addParam: (char*) label 
    Vary: (BOOL) vflag
    DValue: (double) d
    Message: (SEL) m
    Owner: (id) o
    Offset:   (id) op
{
    OffsetParam * param;

    if (![o respondsTo: m]) {
        geckoError(InvalidCombination,
            "Parameter message error applying '%s' to '%s'.\n",
            sel_get_name(m), [o getName]);  
        return self;
    } 
    if ([op getParamType] != DoubleParam) {
        geckoError(InvalidCombination,
            "Offset parameter type mismatch.\n"); 
        return self;
    }
    
    param             = [OffsetParam create: [self getZone]];
    param->paramType  = DoubleParam;
    param->dval       = d - [op getDValue];
    param->offsetFrom = op;
    [param setLabel: label Vary: vflag Message: m Owner: o];
    [params addLast: param];
    
    return self;
}

-addParam: (char*) label 	
    Vary: (BOOL) vflag
    IValue: (int) i
    Message: (SEL) m
    Owner: (id) o
{
    Param * param;
    
    if (![o respondsTo: m]) {
        geckoError(InvalidCombination,
            "Parameter message error '%s' doesn't respond to '%s'.\n",
            [o getName], sel_get_name(m));  
        return self;
    } 
    param          = [Param create: [self getZone]];
    param->paramType = IntParam;
    param->ival    = i;
    [param setLabel: label Vary: vflag Message: m Owner: o];
    [params addLast: param];

    return self;
}

-addParam: (char*) label 	
    Vary: (BOOL) vflag
    SValue: (char*) s
    Message: (SEL) m
    Owner: (id) o
{
    Param * param;
    if (![o respondsTo: m]) {
        geckoError(InvalidCombination,
            "Parameter message error '%s' doesn't respond to '%s'.\n",
            [o getName], sel_get_name(m));  
        return self;
    } 
    param          = [Param create: [self getZone]];
    param->paramType = StringParam;
    param->sval    = s;
    [param setLabel: label Vary: vflag Message: m Owner: o];
    [params addLast: param];
    
    return self;
}


-addParam: (char*) label 	
    Vary: (BOOL) vflag
    IValue: (int) i
    Message: (SEL) m
    Owner: (id) o
    Offset:   (id) op
{
    OffsetParam * param;
    
    if (![o respondsTo: m]) {
        geckoError(InvalidCombination,
            "Parameter message error '%s' doesn't respond to '%s'.\n",
            [o getName], sel_get_name(m));  
        return self;
    } 
    if ([op getParamType] != DoubleParam) {
        geckoError(InvalidCombination,
            "Offset parameter type mismatch.\n"); 
        return self;
    }
    param             = [OffsetParam create: [self getZone]];
    param->paramType  = IntParam;
    param->ival       = i - [op getIValue];
    param->offsetFrom = op;
    [param setLabel: label Vary: vflag Message: m Owner: o];
    [params addLast: param];
    
    return self;
}

-offsetParam: (char*) label Off: (id) op
    // Offset a parameter you've already added.  Necessary for offsetting from
    // parameter of a different parameter list or overriding what the species
    // added as default parameters.
    // This version takes subtracts op's value from current value.
{
    Param* origp;
    OffsetParam* offp;
    id     index;
    
    index = [params begin: [self getZone]];
    while ( (origp = [index next]) && (strcmp(origp->label, label) ) ) {
    }
    if (!origp) {
        geckoError(InvalidCombination,
            "Parameter '%s' not found.\n", label); 
        return self;
    }
    
    offp             = [OffsetParam create: [self getZone]];
    offp->paramType  = origp->paramType;
    offp->offsetFrom = op;

    // from now on, my value is added to op's value, so subtract to start
    if (origp->paramType == IntParam) 
        offp->ival   = origp->ival - [op getIValue];
    else if (origp->paramType == DoubleParam) 
        offp->dval   = origp->dval - [op getDValue];

    [offp setLabel: origp->label Vary: origp->varyFlag 
          Message: origp->message Owner: origp->owner];
    [index put: offp];
    [origp drop];
    return self;
}

-offsetParam: (char*) label AbsOff: (id) op
    // Offset a parameter you've already added.  Necessary for offsetting from
    // parameter of a different parameter list or overriding what the species
    // added as default parameters.
{
    id origp;
    OffsetParam* offp;
    Param* oldp;
    id     index;
    
    // find parameter to offset
    index = [params begin: [self getZone]];
    while ( (origp = [index next]) && (strcmp([origp getName], label) ) ) {
    }
    if (!origp) {
        geckoError(InvalidCombination,
            "Parameter '%s' not found.\n", label); 
        return self;
    }
    
    // if it's already an offset, only change what it's offset from
    if ( [ origp isMemberOf: [OffsetParam class] ] ) {
        offp = origp;
        if (op == offp->offsetFrom) {
            return self;
        } else {
            offp->offsetFrom = op;
            return self;
        }
    }
    
    // not presently offset from anything
    oldp             = origp;
    offp             = [OffsetParam create: [self getZone]];
    offp->paramType  = oldp->paramType;
    offp->offsetFrom = op;
    offp->ival       = oldp->ival;
    offp->dval       = oldp->dval;

    [offp setLabel: oldp->label Vary: oldp->varyFlag 
          Message: oldp->message Owner: oldp->owner];
    [index put: offp];
    [origp drop];
    return self;
}

-setParam: (char*) label IValue: (int) i
{
    Param* p = [self getParam: label];
    [p setIValue: i];
    return self;
}

-setParam: (char*) label DValue: (double) d
{
    Param* p = [self getParam: label];
    [p setDValue: d];
    return self;
}

-setParam: (char*) label SValue: (char*) s
{
    Param* p = [self getParam: label];
    [p setSValue: s];
    return self;
}

-varyParam: (char*) label
{
    Param* p = [self getParam: label];
    p->varyFlag = 1;
    return self;
}

-(Param*) getParam: (char*) label
{
    id onep;
    id     index = [params begin: [self getZone]];

    while ( (onep = [index next] ) ) {
        if ( ! strcmp( [onep getName], label ) ) { 
            [index drop];
            return onep;
        }
    }
    [index drop];
    return nil;
}

// This routine is intended for override downstream.
-(Param*) getOptParam: (char*) label { return [self getParam: label]; } 

-(id) getParams  {return params;}
-(char*) getName {return name;}

-setName: (char*) label
{
    name = label; 
    return self;
}

-apply 
{
    [params forEach: M(apply)];
    return self;   
}

-print
{
    printLog("%s", name); return self;    
}

-println
{
    [params forEach: M(println)];return self;   
}

-printv
{
    printLog("Name = %s\n", name);
    [params forEach: M(println)];return self;   
}

-dump
{
    [params forEach: M(dump)]; return self;
}


// The following routines are subroutines of -loadToNextSection.  Not of much
// use out of context -- just divide and conquer.

int isLinuxSpace(char c)
{
    if (c == '\n')
        return 0;
    return (isspace(c));
}

-parseWord
{
    while (isLinuxSpace(parseBuffer[parsep])) parsep++;
    wordp = 0;
    while ( (!isLinuxSpace(parseBuffer[parsep])) && 
           (parseBuffer[parsep] != '\n') && (parseBuffer[parsep] != '\0') ) {
        wordBuffer[wordp] = parseBuffer[parsep];
        parsep++;
        wordp++;
    }
    wordBuffer[wordp] = '\0';
    return self;
}

-(Parameterized*) getParameterized: (char*) label
{
    Parameterized* onep;
    id     index = [listOfParameterizeds begin: [self getZone]];
    
    while ( (onep = [index next] ) ) {
        if ( ! strcmp( onep->name, label ) ) { 
            return onep;
        }
    }
    [index drop];
    return nil;
}


-(Param*) parseOffsetParam 
{
    int paramp;
    Param* offsetParam;
    Parameterized* otherSelf;
    
    [self parseWord]; 
    // now wordBuffer is owner:parameter--replace ':' with '\0'
    paramp = 0;
    while ((wordBuffer[paramp] != ':') && (wordBuffer[paramp] != '\0')) paramp++;

    // if no ':' found, must be referring to current thang.
    if (wordBuffer[paramp] == '\0') {
        offsetParam = [self getParam: wordBuffer];

    // found a ':', so replace it, and find the Parameterized referred to.
    } else {
        wordBuffer[paramp] = '\0';
        paramp++;
        if (!(strcmp(wordBuffer, [self getName]))) {
            offsetParam = [self getParam: &(wordBuffer[paramp])];
        } else {
            otherSelf = [self getParameterized: wordBuffer];
            if (!otherSelf)
                offsetParam = nil;
            else {
                offsetParam = [otherSelf getParam: &(wordBuffer[paramp])];
            }
        }
    }
    
    return offsetParam;
}

-loadParam: (Param*) p
{
    int    i;
    double d;
    Param* offp;
    char*  string;
    
    switch ([p getParamType]) {
        case IntParam:    
                          [self parseWord];
                          if (wordp == 0) {
                              geckoError(InvalidCombination,
                                "Error in protocol:  parameter incomplete.\n%s",
                                parseBuffer);
                              return self;
                          }
                          sscanf(wordBuffer, "%d", &i); 
                          [p setPartIValue: i];
                          break;
        case DoubleParam: 
                          [self parseWord];
                          if (wordp == 0) {
                              geckoError(InvalidCombination,
                                "Error in protocol:  parameter incomplete.\n%s",
                                parseBuffer);
                              return self;
                          }
                          sscanf(wordBuffer, "%lf", &d); 
                          [p setPartDValue: d];
                          break;
        case StringParam: 
                          [self parseWord];
                          if (wordp == 0) {
                              geckoError(InvalidCombination,
                                "Error in protocol:  parameter incomplete.\n%s",
                                parseBuffer);
                          }	
                          string = [[self getZone] alloc: wordp+1];
                          for (i=0; i <= wordp; i++) 
                              string[i] = wordBuffer[i];
                          [p setSValue: string];
                          break;
        default:          
                          break;
    }
    
    // rest of line, if any....
    
    [self parseWord];
    while (wordp > 1) {
        i=0;
/*
        while (wordBuffer[i] != '\0') {
            toupper(wordBuffer[i]); 
            i++;
        }
*/
                
        if (!strcmp(wordBuffer, "Vary")) {
            p->varyFlag = 1;
            
        } else if (!strcmp(wordBuffer, "Offset")) {
            if (!(offp = [self parseOffsetParam])) {
                geckoError(InvalidCombination,
                    "Error in protocol:  offset target not found.\n%s",
                    parseBuffer);
            } else {
                [self offsetParam: p->label AbsOff: offp];
            }
        }
            
        [self parseWord];
    }

    return self;
}

-loadCommand
    // OK, try to execute it.  For now, only no-arg commands allowed.
    // GINGER:  generalize this--can find the arguments (a few, at least).
{
    SEL sel;

    sel = sel_get_uid( wordBuffer );
    if ( sel ) {
        if ( [self respondsTo: sel] ) {
	    [self perform: sel];
        } else {
            geckoError(InvalidCombination, 
                "Protocol error, command '%s' not recognized.\n", wordBuffer);
        }
    } else {
        geckoError(InvalidCombination,
            "Protocol error, command '%s' not recognized.\n", wordBuffer);
    }
    return self;
}

-loadln
{
    Param* param;
    
    [self parseWord];
    // is it a parameter? 
    param = [self getOptParam: (char*) wordBuffer];
    if (param) {
        [self loadParam: param];
    // is it a command to execute?
    } else {
        [self loadCommand];
    }
    return self;
}

-(char*) reloadln: (FILE*) fp
{
    parsep = 0;
    [self loadln];
    return [self loadToNextSection: fp];
}

-(char*) loadToNextSection: (FILE*) fp
    // parses/executes on current parameterized until next @begin / EOF.
    // if @begin, string returned is rest of line after @begin
    // if EOF, returns nil for string.
{
    do {
        parsep = 0;
        if (!fgets(parseBuffer, PARSEBUFSIZE, fp)) {
            [self apply];
            return (char*) 0;
        }
        
        while (isLinuxSpace(parseBuffer[parsep])) parsep++;
        if ( (parseBuffer[parsep] == '#') || (parseBuffer[parsep] == '\0') ||
             (parseBuffer[parsep] == '\n')) {
            // next....
        } else if (parseBuffer[parsep] == '@') {
            parsep++;
            [self parseWord];
            if ( (!strcmp(wordBuffer, "begin")) ||
                 (!strcmp(wordBuffer, "Begin")) ||
                 (!strcmp(wordBuffer, "BEGIN")) ) {
                while (isLinuxSpace(parseBuffer[parsep])) parsep++;
                [self apply];
                return(&parseBuffer[parsep]);
                 
            } else if ( (!strcmp(wordBuffer, "end")) ||
                 (!strcmp(wordBuffer, "End")) ||
                 (!strcmp(wordBuffer, "END")) ) {
                 // next....
            }
        } else {
            [self loadln];
        }
    } while (1);

    [self apply];
    return (char*) 0;       
}

-(char*) loadWhere;
    // reiterates what was returned at last loadToNextSection
{
    return &parseBuffer[parsep];
}

-(char*) loadNextWord: (FILE*) fp
    // single-stepping through words now--like loadToNext, except executes
    // nothing, just parses.
{

    while (isLinuxSpace(parseBuffer[parsep])) parsep++;
    if ( (parseBuffer[parsep] == '#') || (parseBuffer[parsep] == '\0') ||
            (parseBuffer[parsep] == '\n')) {
        // fall through
    } else {
        [self parseWord];
        return wordBuffer;
    }
    
    do {
        parsep = 0;
        if (!fgets(parseBuffer, PARSEBUFSIZE, fp)) {
            return (char*) 0;
        }
        
        while (isLinuxSpace(parseBuffer[parsep])) parsep++;
        if ( (parseBuffer[parsep] == '#') || (parseBuffer[parsep] == '\0') ||
             (parseBuffer[parsep] == '\n')) {
            // next....
        } else if (parseBuffer[parsep] == '@') {
            parsep++;
            [self parseWord];
            if ( (!strcmp(wordBuffer, "begin")) ||
                 (!strcmp(wordBuffer, "Begin")) ||
                 (!strcmp(wordBuffer, "BEGIN")) ) {
                while (isLinuxSpace(parseBuffer[parsep])) parsep++;
                return(&parseBuffer[parsep]);
                 
            } else if ( (!strcmp(wordBuffer, "end")) ||
                 (!strcmp(wordBuffer, "End")) ||
                 (!strcmp(wordBuffer, "END")) ) {
                 // next....
            }
        } else {
            [self parseWord];
            return wordBuffer;
        }
    } while (1);
}

 
@end
