// orig, vfb, 951018
// rewrite, vfb, 951221

// The purpose of the quadding tree is 
// to keep things sorted by area to support faster picks, draws, etc.  This
// is low-level stuff:  a triply-linked sideways binary sort tree.

// All things on the tree must have mcr's="minimum containing rectangles".  
// In other words, they must have area.  The area of a quad or any other 
// collector with substructure, is the mcr of all of its substructure. 
// In general, "the tree" has a root quad, whose children are logical 
// collectors (uptakers, grazers, predators, or whatever), whose children
// are quadding trees of agents.

#import <swarmobject.h>			// we inherit swarm objects
#import "Base.h"	
#import "Mcr.h"	

// quads of type "aquad" are transparent and subject to deletion
// whenever devoid of children.  The others have meaning.
typedef enum { invalid, aquad, newbies, critter, 
               landscape, feed, diffuse, sessile, mobile } QuadType;
            
#define isagent(x) ( x==critter )
#define ispickable(x) ((x==critter) || (x==feed) || (x==diffuse)) 

@interface Quad: SwarmObject {
  @public
    QuadType 	type;
    BOOL        visible;
    Mcr* 	mcr;
    id	 	data;		
    int		color;			// used sometimes
    Quad*	parent;
    Quad*	sibling;
    Quad*	child;
}

// Gimme a new/recycled quad.
+create;
-recycle;
-recycleClear;                  // clear out recycling list
-reuse;

// low-level setting/getting operations--these affect only one node
-setParent: (Quad *) p;
-setSibling: (Quad *) s;
-setChild: (Quad *) c;
-setData: d Type: (QuadType) t;
-setColor: (int) c;
-(GCoord) getRadius;

// high-level setting operations--these have ripple effects on other nodes
-growMcr;			// grows mcr's above this node, as needed
-(Mcr *) getMcr;

// Root: determines extent of subtree to hunt across
// self says where to pick up the search.
-(Quad *) pickWin: (Mcr*) pickwin 
          Root: (Quad*) rootPick Invisibles: (BOOL) pickinvis;

// calls other with invisibles = false
-(Quad *) pickWin: (Mcr*) pickwin Root: (Quad*) rootPick;

// root = self, pick window = self mcr, pick something visible
-(Quad *) pick;

// Root: determines extent of subtree to hunt across
// self says where to pick up the search.  Pick circles with circles, and
// make sure distance < r1 + r2
-(Quad *) pickRadius: (GCoord) pickR X: (GCoord) pickX Y: (GCoord) pickY
          Root: (Quad*) pickRoot;

-delNode;			// unplug only this quad tree node--no drop
-delData;			// node and its data
-delTree;			// node, data, and substructure nodes & data

-addNode: (Quad *) p;		// add node to newbies quad under this parent
-findNewbies;                   // find newbies subcollector--make if none

-(void) subQuad;		// recursively subquad the children of this quad
-squashQuad;			// flatten subtree into simple child list
-diagnose;			// check parent/child match & agent has data

-(void) printQuad;		// just this quad, no eol.
-println;		        // just this quad, plus eol.
-(void) printTree;		// call printTree, not the one with arguments.
-(void) printTree: (int) depth Quads: (int*) qcount Leaves: (int*) lcount;
-(void) headcount: (char*) message; // just print number of agents in tree

// Collect some stats about the tree.  Note that only "aquad" goes in qcount,
// all else in lcount.
-(int) maxDepth: (int) depth Quads: (int*) qcount Leaves: (int*) lcount;

// call -step on each leaf node on the tree. 
-step;
// calls step only on roundth round
-stepOn: (int) roundth;

@end
extern int pickTrace;
