// vfb, orig, 960909:

// Site and Species really should be ParamMaster subclasses.  Too similar.
// Let's not bother overhauling sites now, though....

#import "Param.h"
#import "Site.h"
#import "LandRoot.h"
#import "ResourceQuad.h"
#import "FeedQuad.h"
#import "Diffuse.h"
#import "Stat.h"
#import "Colors.h"
#import "GeckoControlSwarm.h"

@implementation Site;

-init
{
    grandma = nil;
    bioscale = 1;
    [super init];
    return self;
}

+create
{
    id result = [[Site createBegin: globalZone] createEnd];
    result = [result init];
    return result;					
}

-setType: (char*) t { return self; } // does nothing--can't reset

-make: (char*) n WithGrandma: (id) fq 
    // "create" a grandma of correct type, then this
    // rely on class of grandma to dynamically select correct message
{
    name     = n;
 
    [self addParam: "Type" Vary: 0 SValue: object_get_class_name(fq)
          Message: M(setType:) Owner: self];
    [self addParam: "BioScale" Vary: 0 DValue: bioscale
          Message: M(setBioScale:) Owner: self];

    // remaining parameters added by grandma
    grandma = [fq landGrandma: self];

    return self;
}

-setGrandma: (id) g { grandma = g; return self; }

-(id) getGrandma {return grandma;}

-(Param*) getOptParam: (char*) label
{
    id result;
    char* opt1="InitCellLX";
    char* opt2="InitCellLY";
    char* opt3="InitCellUX";
    char* opt4="InitCellUY";

    // simple case--one of the required parameters
    if ((result = [self getParam: label])) {
        return result;
    }
    
    // OK, now consider whether it's one of our optional params --  
    if ( !strcmp(label, opt1) ||
         !strcmp(label, opt2) ||
         !strcmp(label, opt3) ||
         !strcmp(label, opt4) ) {
        
        [grandma setHaveInitMcr];
        [self addParam: opt1 Vary: 0 DValue: 0 
            Message: M(setICLX:) Owner: grandma];
        [self addParam: opt2 Vary: 0 DValue: 0 
            Message: M(setICLY:) Owner: grandma];
        [self addParam: opt3 Vary: 0 DValue: 0 
            Message: M(setICUX:) Owner: grandma];
        [self addParam: opt4 Vary: 0 DValue: 0 
            Message: M(setICUY:) Owner: grandma];
    } else {
        return nil; 
    }
    
    return [self getParam: label];
} 

-observeBiomass
{
    bioFlag = 1;
    [geckoControl->biomassGraph trackDouble: &biomass Key: name Color: "brown"];
    return self;
}

-observeBiomassFed
{
    char* namebuf = [globalZone alloc: 50];
    sprintf(namebuf, "%s %s", name, "fed");
    bioFedFlag = 1;
    
    [geckoControl->biomassGraph trackDouble: &biomassFed 
        Key: namebuf Color: "brown3"];
    return self;
}

-setBioScale: (double) d {bioscale = d; return self;}

-(double) getBiomass {return biomass;}
-setBiomass: (double) b { dbiomass= b; return self;}
-addBiomass: (double) b { dbiomass += b; return self;}

-(double) getBiomassFed {return biomassFed;}
-setBiomassFed: (double) b { dbiomassFed= b; return self;}
-addBiomassFed: (double) b { dbiomassFed += b; return self;}

-zeroBiomass { dbiomass = biomass = dbiomassFed = biomassFed = 0; return self;}
-scaleBiomass {
    biomass = dbiomass / bioscale; 
    biomassFed = dbiomassFed / bioscale; 
    return self;
}

-onSeason { return self; }

-print
{
    printLog("%s", name); return self;    
}

-println
{
    printLog("Site %s ", name); 
    [grandma println];
    return self;    
}

-printv
{
    [self println];
    printLog("Site params: \n"); 
    [super printv];
    printLog("Site grandma: \n"); 
    [grandma printv];
    return self;
}

-dump
    // reloadable
{
    printBoth("@begin Site %s\n", name);
    [super dump];
    if (bioFlag)
        printBoth("\tobserveBiomass\n");
    if (bioFedFlag)
        printBoth("\tobserveBiomassFed\n");
    printBoth("@end Site %s\n\n", name);
    return self;
}

@end
