// vfb, orig, 960909:
// 	Need a species minder object.
// vfb, 961112: 
// 	Composite species--basically, for multiple species in a class.

#import <swarmobject.h>		
#import <activity.h>				  
#import "Param.h"				  
#import "Stat.h"				  

// Place to keep the books for a species as a whole.  Note that agents are
// instances, members of a species.  The archetypal member is the grandma,
// kept here.  You can't get from the Species record to all the members of
// a species.  That info's not duplicated here--must go out into the world
// tree and look.

// I do wish Obj-C had multiple inheritance...

@interface Species: Parameterized {
    int          color;		// orig. display color (can override in agent)
    id 		 grandma;	// archetypal species member
    unsigned     census;	// members currently alive--kept up to date 
    unsigned     lastalive;	// last time we noticed they were still alive
    double       biomass;	// scaled total biomass of members
    double       dbiomass;	// total biomass of members--set by Stats.
    unsigned     birthday;	// when to introduce
    unsigned     saveBirthday;	// when to introduce--nobody resets
    unsigned     deathday;	// when to make extinct (if nonzero)
    unsigned     initNum;	// how many to introduce
    unsigned     saveInitNum;	// how many to introduce--nobody resets
    Mcr*	 birthArea;	// optional--where to introduce
    SEL          resetAreaMessage; // reset birthArea
    id           birthAreaOwner;   // who to send message to reset area.
    SEL          seasonMessage; // reset rigamarole onSeason
    BOOL         mobile;	// mobile (or plant) root
    double       bioscale;	// divide biomass by this for scaling
    AgeStructure* ageStats;	// statistics on age structure of population
    TraceGraph*   ageGraph;	// if want ageStats and have display, show it.
    SizeStructure* sizeStats;	// statistics on size structure of population
    TraceGraph*   sizeGraph;	// if want sizeStats and have display, show it.
    StageStructure* stageStats;	// statistics on stage structure of population
    TraceGraph*   stageGraph;	// if want stageStats and have display, show it.
    SpecificStructure* specificStats;// species-implemented statistics
    TraceGraph*   specificGraph;// if want specificStats and have display, show it.
    BiomassRates* bioRateStats;	// biomassRates 
    TraceGraph*   bioRateGraph;	// and graph
    BiomassTransfers* bioTransStats;    // abs. biomass transfers
    TraceGraph*   bioTransGraph;	// and graph
    TangleGraph*  bioTheftGraph;	// losses (e.g. herbivory) v biomass graph
    TangleGraph*  bioGrowthGraph;	// growth v biomass graph
    TraceGraph*   fertGraph;	
    TraceGraph*   mortGraph;	
    BOOL          popFlag;	// if want trace on popGraph
    BOOL          bioFlag;	// if want trace on biomassGraph
    // popup menu for user to intervene on params...not yet.
}

+create;
-createEnd;	// after loading *everything*, finish up the bookkeeping

// "create" a grandma of correct type, then call this
-make: (char*) n WithGrandma: (id) g Color:   (int) c
                 Birthday:   (int) b InitNum: (int) i
                 Mobile: (BOOL) m    BioScale: (double) bs;

// All the "observeXXX" routines may be used as a protocol directive.  Some
// have defaulted optional parameters that ought to be set (Ages).
// NOTE:  to get l(x)m(x) type stats, you must set statcycle and the length
// of an age on observeAges to match, because the rates are per-statcycle
// (i.e., fertility/age*statcycle).
-observeAges;			// AgesBuckets 5, age len = statcycle
-observeAgesOnly;		// skip mortality & fertility traces (still calc)
-observeSizes;			// species determines the size def'ns
-observeSizesOnly;		// skip mortality & fertility traces (still calc)
-observeStages;			// species determines the stage def'ns
-observeStagesOnly;		// skip mortality & fertility traces (still calc)
    // observeSpecific stats are an arbitrary stage structure, in effect.
-observeSpecific;		// species defined

-observePop;
-observeBiomass;
-observeBiomassRates;		// growth, losses, netGrowth rates
-observeBiomassTransfers;       // growth, losses, netGrowth absolutes
-observeBiomassTransfersSilent; // no graphs or output--this is a substat
-observeBiomassTheft;           // theft snanglegram
-observeBiomassGrowth;          // growth snanglegram

-setBirthArea: (Mcr*) where
    ResetMessage: (SEL) m
    Owner: (id) o;
-setOnSeason: (char*) s; 
-setDeathDay: (int) d;		// all die at time d

// these are for use by protocol loader
-setBALX: (GCoord) d;
-setBALY: (GCoord) d;
-setBAUX: (GCoord) d;
-setBAUY: (GCoord) d;

-setAgesBuckets: (int) i; 	
-setStagesBuckets: (int) i; 
-setSizesBuckets: (int) i; 
-setSpecificBuckets; 

-birthdayCheck;			
-deathdayCheck;			

-resetClear;			// prepare for next round

// birthplace, etc. are optional parameters, so override from Parameterized
-(Param*) getOptParam: (char*) label; 

-(unsigned) getCensus;
-(unsigned) getLastAlive;
-(unsigned) getBirthday;
-incrementCensus;
-decrementCensus;
-onSeason; /* if any seasonMessage, call on each member, then do birthday */

-(double) getBiomass;
-setBiomass: (double) b;
-addBiomass: (double) b;
-zeroBiomass;
-scaleBiomass;

// rates & transfers
-tallyGrowth: (id) instance Amount: (ResQuant) b;
-tallyLossSteal: (id) instance Amount: (ResQuant) b;
-tallyLossStarve: (id) instance Amount: (ResQuant) b;
-zeroBiomassRates;		

// don't do anything unless you asked to observe something
-tallyAlive: (id) instance;
-tallyDeath: (id) instance;
-tallyBirth: (id) instance;
-biomassRateUpdate;
-biomassTransferUpdate;		// called by biomassRateUpdate
-allStatsReset;

-setBirthday: (unsigned) b;
-setBothBirthday: (unsigned) b; // current & save
-setInitNum:  (unsigned) i;
-setBothInitNum:  (unsigned) i; // current & save
-(unsigned) getInitNum;

-(int) getColor;
-setColor: (int) c;

// these are for use by CompositeSpecies...I want multiple inheritance...
-(id) getAgeStats;
-(id) getSizeStats;
-(id) getSpecificStats;
-(id) getStageStats;
-(BiomassRates*) getBioRateStats;	// biomassRates 
-(BiomassTransfers*) getBioTransStats;	// biomassRates 
-(double) getBioscale;
-(id) getGrandma;
// these hooks are for observeSpecific, where Species may wish to ensure a
// mortality/fertility graph is available to add to
-makeMortAndFertMax: (int) max ;
-(TraceGraph*) getMortGraph;
-(TraceGraph*) getFertGraph;
-(TraceGraph*) getSpecificGraph;
-setSpecificGraph:  (TraceGraph*) sGraph;
-setSpecificStats:  (id) specific;

// the following are used in creating species from protocol--no error checking
// for arbitrary use.
-setType: (char*) s; // dummy--type can't be set.
-setName: (char*) s;
-setMobile: (int) i; 
-setBioScale: (double) d ;

-print;
-println;
-printv;
-printCensus; // just "species:count ", no eoln.
-dump; // to file controlled by (geckoControl getObjectSaver)

@end

// Note, I'm not going to bother with a lot of error checking here.  If the
// species are not summable, this won't make them so.  Wish I could remember
// why this isn't an extension of Species....
@interface CompositeSpecies : Parameterized {
    int          color;		// display color 
    id 		 members;	// list of Species
    unsigned     census;	// members alive
    unsigned     lastalive;	// last time we noticed they were still alive
    double       biomass;	// total biomass of members
    double       bioscale;	// divide biomass by this for scaling
    AgeStructure* ageStats;	// statistics on age structure of population
    TraceGraph*   ageGraph;	// if want ageStats and have display, show it.
    SizeStructure* sizeStats;	// statistics on size structure of population
    TraceGraph*   sizeGraph;	// if want sizeStats and have display, show it.
    StageStructure* stageStats;	// statistics on stage structure of population
    TraceGraph*   stageGraph;	// if want stageStats and have display, show it.
    StageStructure* specificStats;// species-implemented statistics
    TraceGraph*   specificGraph;// if want specificStats and have display, show it.
    TraceGraph*   mortGraph;	// if want stageStats and have display, show it.
    TraceGraph*   fertGraph;	
    BiomassRates* bioRateStats;	// biomassRates 
    TraceGraph*   bioRateGraph;	// and graph
    BiomassTransfers* bioTransStats;	
    TraceGraph*   bioTransGraph;	// and graph
    TangleGraph*  bioTheftGraph;	// losses (e.g. herbivory) v biomass graph
    TangleGraph*  bioGrowthGraph;	// growth v biomass graph
    BOOL          popFlag;	// if want trace on popGraph
    BOOL          bioFlag;	// if want trace on biomassGraph
}
+create;
-createEnd;

// call once for each member of the composite; obviously, need to exist first.
// call all the addSpecies before any of the observe's.
-addSpecies:  (Species*) s;

-(char*) getName;
-setName: (char*) s;
-(int) getColor;
-setColor: (int) c;
-(unsigned) getLastAlive;
-(unsigned) getBirthday;

// to use these, it's best to only use them on the CompositeSpecies, and
// not on the individual species.
-observeAges;
-observeSizes;
-observeStages;
-observeAgesOnly;		// skip mortality & fertility traces (still calc)
-observeStagesOnly;		// skip mortality & fertility traces (still calc)
-observeSizesOnly;		// skip mortality & fertility traces (still calc)

-observeBiomassRates;		// growth, losses, netGrowth rates
-observeBiomassTransfers;       // growth, losses, netGrowth 
-observePop;
-observeBiomass;
-observeBiomassTransfersSilent; // no graphs or output--this is a substat
-observeBiomassTheft;
-observeBiomassGrowth;

-setAgesBuckets: (int) i; 	
-setStagesBuckets: (int) i; 
-setSizesBuckets: (int) i; 
-setSpecificBuckets: (id) grandma; 

-resetClear;			// prepare for next round
-allStatsReset;			// noop, stats reset at beginning for composite

-(unsigned) getCensus;
-(double) getBiomass;
-setBiomass: (unsigned) b;
-addBiomass: (unsigned) b;
-zeroBiomass;
-scaleBiomass;

-ageUpdate;			// noops
-sizeUpdate;
-stageUpdate;
-specificUpdate;
-biomassRateUpdate;
-biomassTransferUpdate;

// no-ops
-birthdayCheck;
-deathdayCheck;
-onSeason;

// these are for use by specific stats...
-(id) getAgeStats;
-(id) getSizeStats;
-(id) getSpecificStats;
-(id) getStageStats;
-(BiomassRates*) getBioRateStats;	// biomassRates 
-(BiomassTransfers*) getBioTransStats;	// biomassRates 
-(double) getBioscale;
-makeMortAndFertMax: (int) max ;
-(TraceGraph*) getMortGraph;
-(TraceGraph*) getFertGraph;
-setSpecificGraph:  (TraceGraph*) sGraph;
-setSpecificStats:  (id) specific;

-print;
-println;
-printv;
-printCensus; // just "species:count ", no eoln.
-dump; 

@end
