// 960102, prototype, vfb.

#import "Stat.h"
#import "Quad.h"
#import "LandRoot.h"
#import "GeckoControlSwarm.h"
#import "Evolve.h"
#import "Critter.h"

@implementation GenericBucket

+createCounters: (int) nc Sums: (int) ns Derived: (int) nd
{
    GenericBucket* obj = [GenericBucket createBegin: globalZone];
    obj->numCounters = nc;
    obj->numSummers  = ns;
    obj->numDerived  = nd;
    if (nc > 0)
        obj->counters = (int*) [globalZone alloc: nc * sizeof(int)];
    if (ns > 0)
        obj->summers  = (double*) [globalZone alloc: ns * sizeof(double)];
    if (nd > 0)
        obj->derived  = (double*) [globalZone alloc: nd * sizeof(double)];
    return obj;
}

-clear
{
    int i;
    for (i=0; i<numCounters; i++)
        counters[i] = 0;
    for (i=0; i<numSummers; i++)
        summers[i] = 0;
    for (i=0; i<numDerived; i++)
        derived[i] = 0;
    return self;
}
-calc { return self; }

-addData: (GenericBucket*) substat
    // add summers and counters, ignore derived.
{
    int i, which;
    for (i=0; i<substat->numCounters; i++) {
        if (i > numCounters - 1)
            which = numCounters - 1;
        else
            which = i;
        counters[which] += substat->counters[i];
    }
    for (i=0; i<numSummers; i++) {
        if (i > numSummers - 1)
            which = numSummers - 1;
        else
            which = i;
        summers[which]  += substat->summers[i];
    }
    return self;
}

-(int) getTotal 
    // of counters
{
    int i, sum;
    sum = 0;
    for (i=0; i<numCounters; i++)
        sum += counters[i];
    return sum;
}

-(double) getDoubleTotal
    // of summers
{
    int i;
    double sum;
    sum = 0;
    for (i=0; i<numSummers; i++)
        sum += summers[i];
    return sum;
}

-(int) getNumCounters { return numCounters;}		
-(int) getNumSummers {return numSummers;}		
-(int) getNumDerived {return numDerived;}		
-(int) getCount: (int) i { return counters[i];}		
-(double) getDerived: (int) i { return derived[i];}		
-(double) getSum: (int) i {return summers[i];}		
-setCount: (int) i Value: (int) num {counters[i] = num; return self;}		
-incCount: (int) i { counters[i]+= 1; return self;}	
-setSum: (int) i Value: (double) quant {summers[i] = quant; return self;}			
-incSum: (int) i Value: (double) quant {summers[i] += quant; return self;}	
-setDerived: (int) i Value: (double) d {derived[i] = d; return self;}	
-println {
    int i;
    for (i=0; i<numCounters; i++)
        printLog ("  %5d  ", counters[i]);
    for (i=0; i<numSummers; i++)
        printLogD("  %7.4f", summers[i]);
    for (i=0; i<numDerived; i++)
        printLogD("  %7.4f", derived[i]);
    printLog("\n");
    return self;
}
@end

@implementation BucketDistrib

-init: (int) buckets Type: (char*) classname
    Counters: (int) nc Sums: (int) ns Derived: (int) nd
{
    int i;
    maxBucket = buckets;
    total     = 0;
    dtotal    = 0;
    
    if (maxBucket < 1) {
        geckoError(InvalidArgument, 
            "Cannot have bucket stat with no buckets.\n");
            return self;
    }
    if (!objc_lookup_class(classname)) {
        geckoError(InvalidArgument, 
            "Invalid stat bucket type = '%s'.\n", classname );
            return self;
    }
    totalBucket = [objc_lookup_class(classname) createCounters: nc 
            Sums: ns Derived: nd];
    [totalBucket clear];
    bucket = [globalZone alloc: buckets * sizeof(totalBucket)];
    for (i=0; i<maxBucket; i++) {
        bucket[i] = [objc_lookup_class(classname) createCounters: nc 
            Sums: ns Derived: nd];
        [bucket[i] clear];
    }
    return self;
}

+createMax: (int) buckets Type: (char*) classname
    Counters: (int) nc Sums: (int) ns Derived: (int) nd
{
    BucketDistrib* obj = [BucketDistrib createBegin: globalZone];
    [obj init: buckets Type: classname Counters: nc Sums: ns Derived: nd];
    return obj;
}

-(int) getMaxBucket { return maxBucket; }		
-(int) getTotal { return total;}
-(double) getDoubleTotal { return dtotal;}
-calc
    // a very rudimentary default--tally totals
{
    int i;
    [totalBucket clear];
    total = 0;
    dtotal = 0;
    for (i=0; i<maxBucket; i++) {
        [totalBucket addData: bucket[i]];
        total  += [bucket[i] getTotal];
        dtotal += [bucket[i] getDoubleTotal];
    }
    return self;
}

-addBucket: (BucketDistrib*) substat
    // add substat data (not derived) to this.  Don't do anything that calc
    // will do later.
{
    int i, which;
    for (i=0; i < [substat getMaxBucket]; i++) {
        if (i > maxBucket - 1)
            which = maxBucket - 1;
        else
            which = i;
        [bucket[which] addData: substat->bucket[i]];
    }
    return self;
}

-clear
{
    int i;
    for (i=0; i<maxBucket; i++) {
        [bucket[i] clear];
    }
    [totalBucket clear];
    total  = 0;
    dtotal = 0;
    return self;
}

-tallyBucket: (int) b Which: (int) which
     // add one to a counter
{
    [bucket[b] incCount: which];
    return self;
}
-tallyBucket: (int) b Which: (int) which Quant: (double) quant
    // add quant to a summer
{
    double sum = [bucket[b] getSum: which] + quant;
    [bucket[b] setSum: which Value: sum];
    return self;
}
-(double) getBucket: (int) b Derived: (int) which
{
    return [bucket[b] getDerived: which];
}
-(double) getBucket: (int) b Sum: (int) which
{
    return [bucket[b] getSum: which];
}
-(int) getBucket: (int) b Counter: (int) which
{
    return [bucket[b] getCount: which];
}
-printv
{
    int i;
    for (i=0; i < maxBucket; i++) {
        [bucket[i] println];
    }
    printLog("totals:");
    [totalBucket println];
    return self;
}			
@end

@implementation AgeCounters

+create {
    return [super createCounters: AGE_COUNTERS 
           Sums: 0 Derived: AGE_DERIVED];
}

@implementation AgeStructure

-init
{
    int i;
    if (maxBucket < 1) {
        geckoError(InvalidArgument, 
            "Cannot have bucket stat with no buckets.\n");
            return self;
    }
    bucket = [globalZone alloc: maxBucket * sizeof(AgeCounters)];
    for (i=0; i<maxBucket; i++) {
        bucket[i] = [AgeCounters create];
        [bucket[i] clear];
    }
    totalBucket = [AgeCounters create];
    [totalBucket clear];
    return self;
}

+createAges: (int) buckets
{
    AgeStructure* obj = [AgeStructure createBegin: globalZone];
    obj->maxBucket = buckets;
    [obj init];
    return obj;
}

-(int) getMaxAge
{
    return maxBucket;
}

-calc
    // find the derived stats
{
    int a;
    int alive, dead, members0;
    AgeCounters* age;
    double derive;
    
    alive = 0;
    dead  = 0;
    
    // simple sums -- total includes births   
    for (a=0; a < maxBucket; a++) {
        alive += [bucket[a] getCount: AGE_COUNT_MEMBERS];
        dead  += [bucket[a] getCount: AGE_COUNT_DEATHS];
    }
    total = alive + dead;
    
    // probabilities
    for (a=0; a < maxBucket; a++) {
        age = (AgeCounters*) bucket[a];
        // those who started in this age...
        members0 = [age getCount: AGE_COUNT_MEMBERS]
                   - [age getCount: AGE_COUNT_BIRTHS]
                   + [age getCount: AGE_COUNT_DEATHS];
        if (members0 > 0) {
            derive = [age getCount: AGE_COUNT_KIDS];
            derive /= (double)members0;
            [age setDerived: AGE_DERIVE_MX Value: derive];
        }
        // those ever alive in this age...
        members0 = [age getCount: AGE_COUNT_MEMBERS]
                   + [age getCount: AGE_COUNT_DEATHS];
        if (members0 > 0) {
            derive  = [age getCount: AGE_COUNT_DEATHS];
            derive /= (double) members0;
            [age setDerived: AGE_DERIVE_PX Value: derive];
        }
        if (total > 0) {
            derive = (double) members0 / (double) total;
            [age setDerived: AGE_DERIVE_AGEPERCENT Value: derive];
        }
    }
    return self;    
}

-tallyBirthToMom: (Critter*) mom
    // add one
{
    int momAge = [mom getAgeBucket];
    if (momAge >= [self getMaxAge])
        momAge = [self getMaxAge] - 1;
    [bucket[momAge] incCount: AGE_COUNT_KIDS];
    [bucket[0] incCount: AGE_COUNT_BIRTHS];
    return self;
}

-tallyDeath: (Critter*) goner
{
    int gonerAge = [goner getAgeBucket];
    if (gonerAge >= [self getMaxAge])
        gonerAge = [self getMaxAge] - 1;
    [bucket[gonerAge] incCount: AGE_COUNT_DEATHS];
    return self;
}

-tallyAlive: (Critter*) one
{
    int oneAge = [one getAgeBucket];
    if (oneAge >= [self getMaxAge])
        oneAge = [self getMaxAge] - 1;
    [bucket[oneAge] incCount: AGE_COUNT_MEMBERS];
    return self;
}

-printv
{
    int s;
    printLog("           members   kids   births   deaths    m(x)      p(x)  popPercent\n");
    for (s=0; s < maxBucket; s++) {
        if ( ([bucket[s] getCount: AGE_COUNT_MEMBERS] > 0) || 
             ([bucket[s] getCount: AGE_COUNT_DEATHS] > 0) ) {
            printLog("age %2d :", s);
            [bucket[s] println];
        }
    }
    return self;
}

@end // of AgeStructure

@implementation SizeStructure

+createSizes: (int) buckets
{
    SizeStructure* obj = [SizeStructure createBegin: globalZone];
    obj->maxBucket = buckets;
    [obj init];
    return obj;
}


-tallyBirthToMom: (Critter*) mom
    // add one
{
    int momAge = [mom getSizeBucket];
    if (momAge >= [self getMaxAge])
        momAge = [self getMaxAge] - 1;
    [bucket[momAge] incCount: AGE_COUNT_KIDS];
    [bucket[0] incCount: AGE_COUNT_BIRTHS];
    return self;
}

-tallyDeath: (Critter*) goner
{
    int gonerAge = [goner getSizeBucket];
    if (gonerAge >= [self getMaxAge])
        gonerAge = [self getMaxAge] - 1;
    [bucket[gonerAge] incCount: AGE_COUNT_DEATHS];
    return self;
}

-tallyAlive: (Critter*) one
{
    int oneAge = [one getSizeBucket];
    if (oneAge >= [self getMaxAge])
        oneAge = [self getMaxAge] - 1;
    [bucket[oneAge] incCount: AGE_COUNT_MEMBERS];
    return self;
}

-printv
{
    int s;
    printLog("           members   kids   births   deaths    m(x)      p(x)  popPercent\n");
    for (s=0; s < maxBucket; s++) {
        if ( ([bucket[s] getCount: AGE_COUNT_MEMBERS] > 0) || 
             ([bucket[s] getCount: AGE_COUNT_DEATHS] > 0) ) {
            printLog("size %2d:", s);
            [bucket[s] println];
        }
    }
    return self;
}

@end // of SizeStructure

@implementation StageStructure

+createStages: (int) buckets
{
    StageStructure* obj = [StageStructure createBegin: globalZone];
    obj->maxBucket = buckets;
    [obj init];
    return obj;
}


-tallyBirthToMom: (Critter*) mom
    // add one
{
    int momAge = [mom getStageBucket];
    if (momAge >= [self getMaxAge])
        momAge = [self getMaxAge] - 1;
    [bucket[momAge] incCount: AGE_COUNT_KIDS];
    [bucket[0] incCount: AGE_COUNT_BIRTHS];
    return self;
}

-tallyDeath: (Critter*) goner
{
    int gonerAge = [goner getStageBucket];
    if (gonerAge >= [self getMaxAge])
        gonerAge = [self getMaxAge] - 1;
    [bucket[gonerAge] incCount: AGE_COUNT_DEATHS];
    return self;
}

-tallyAlive: (Critter*) one
{
    int oneAge = [one getStageBucket];
    if (oneAge >= [self getMaxAge])
        oneAge = [self getMaxAge] - 1;
    [bucket[oneAge] incCount: AGE_COUNT_MEMBERS];
    return self;
}

-printv
{
    int s;
    printLog("           members   kids   births   deaths    m(x)      p(x)  popPercent\n");
    for (s=0; s < maxBucket; s++) {
        if ( ([bucket[s] getCount: AGE_COUNT_MEMBERS] > 0) || 
             ([bucket[s] getCount: AGE_COUNT_DEATHS] > 0) ) {
            printLog("stage%2d:", s);
            [bucket[s] println];
        }
    }
    return self;
}

@end // of StageStructure

@implementation SpecificStructure

+createBuckets: (int) buckets Counters: (int) nc 
          Sums: (int) ns Derived: (int) nd;
{
    SpecificStructure* obj = [SpecificStructure createBegin: globalZone];
    [obj init: buckets Type: "GenericBucket" 
        Counters: nc Sums: ns Derived: nd];
    return obj;
}

-tallyBirthToMom: (Critter*) mom
    // add one.  Births always go in bucket 0.
{
    int momi, kidi, momd, kidd;
    double howmuch;
    
    int momAge = [mom getSpecificBucket];
    if (momAge < 0)
        return self;
    
    if (momAge >= [self getMaxBucket])
        momAge = [self getMaxBucket] - 1;
    
    momi = [mom getSpecificCounter: SPECIFIC_EVENT_HADKID];
    kidi = [mom getSpecificCounter: SPECIFIC_EVENT_BORN];
    momd = [mom getSpecificSum: SPECIFIC_EVENT_HADKID];
    kidd = [mom getSpecificSum: SPECIFIC_EVENT_BORN];
    
    if (momi >= 0) {
        [bucket[momAge] incCount: momi];
    }
    if (kidi >= 0) {
        [bucket[0] incCount: kidi];
    }
    if (momd >= 0) {
        howmuch = [mom getSpecificAmount: SPECIFIC_EVENT_HADKID];
        [bucket[momAge] incSum: momd Value: howmuch];
    }
    if (kidd >= 0) {
        howmuch = [mom getSpecificAmount: SPECIFIC_EVENT_BORN];
        [bucket[0] incSum: kidd Value: howmuch];
    }
    return self;
}

-tallyDeath: (Critter*) goner
{
    int deadi, deadd;
    double howmuch;
    
    int gonerAge = [goner getSpecificBucket];
    if (gonerAge < 0)
        return self;
    
    if (gonerAge >= [self getMaxBucket])
        gonerAge = [self getMaxBucket] - 1;
    
    deadi = [goner getSpecificCounter: SPECIFIC_EVENT_DIED];
    deadd = [goner getSpecificSum: SPECIFIC_EVENT_DIED];
    
    if (deadi >= 0) {
        [bucket[gonerAge] incCount: deadi];
    }
    if (deadd >= 0) {
        howmuch = [goner getSpecificAmount: SPECIFIC_EVENT_DIED];
        [bucket[gonerAge] incSum: deadd Value: howmuch];
    }
    return self;
}

-tallyAlive: (Critter*) one
{
    int alivei, alived;
    double howmuch;
    
    int oneAge = [one getSpecificBucket];
    if (oneAge < 0)
        return self;
    
    if (oneAge >= [self getMaxBucket])
        oneAge = [self getMaxBucket] - 1;
    
    alivei = [one getSpecificCounter: SPECIFIC_EVENT_ISALIVE];
    alived = [one getSpecificSum: SPECIFIC_EVENT_ISALIVE];
    
    if (alivei >= 0) {
        [bucket[oneAge] incCount: alivei];
    }
    if (alived >= 0) {
        howmuch = [one getSpecificAmount: SPECIFIC_EVENT_ISALIVE];
        [bucket[oneAge] incSum: alived Value: howmuch];
    }
    return self;
}

-tally: (Critter*) c Amount: (ResQuant) b Event: (int) which
{
    int summer;
    
    int abucket = [c getSpecificBucket];
    if (abucket < 0)
        return self;
    
    if (abucket >= [self getMaxBucket])
        abucket = [self getMaxBucket] - 1;

    summer = [c getSpecificSum: which];
    if (summer >= 0) {
        [bucket[abucket] incSum: summer Value: b];
    }
    return self;
}
-tallyGrowth: (Critter*) c Amount: (ResQuant) b
{
    [self tally: c Amount: b Event: SPECIFIC_EVENT_GREW];
    return self;
}

-tallyLossSteal: (Critter*) c Amount: (ResQuant) b
{
    [self tally: c Amount: b Event: SPECIFIC_EVENT_THEFT];
    return self;
}
-tallyLossStarve: (Critter*) c Amount: (ResQuant) b
{
    [self tally: c Amount: b Event: SPECIFIC_EVENT_STARVED];
    return self;
}
-tallySpecific: (id) aCritter EventType: (int) anEvent
{
    int abucket;
    int summer, counter;
    double howmuch;
    
    abucket = [aCritter getSpecificBucket];
    if (abucket < 0)
        return self;
    
    if (abucket >= [self getMaxBucket])
        abucket = [self getMaxBucket] - 1;

    summer = [aCritter getSpecificSum: anEvent];
    counter = [aCritter getSpecificCounter: anEvent];
    if (summer >= 0) {
        howmuch = [aCritter getSpecificAmount: anEvent];
        [bucket[abucket] incSum: summer Value: howmuch];
    }
    if (counter >= 0) {
        [bucket[abucket] incCount: counter];
    }
    return self;
}

@end

@implementation BiomassRates

+create
{
    BiomassRates* obj = [BiomassRates createBegin: globalZone];
    [obj clear];
    obj->lastNetGrowth = 0;
    return obj;
}

-clear
{
    lastNetGrowth = netGrowth;
    growth = lossSteal = lossStarve = 0;
    netGrowth = losses = netRatio = 0;
    return self;
}

-calcNet: (ResQuant) b 
    // b = current biomass
{
    double statcycle = [geckoControl getStatCycle];

    if (b==0) {
        [self clear];    
        return self;
    }
    growth     /= statcycle;
    growth     /= b;
    lossSteal  /= statcycle;
    lossSteal  /= b;
    lossStarve /= statcycle;
    lossStarve /= b;
    
    losses = lossSteal + lossStarve;
    netGrowth = growth - losses;
    netRatio = netGrowth - lastNetGrowth;   
    return self;
}

-tallyGrowth: (Critter*) c Amount: (ResQuant) b 
{ growth += b; return self; }
-tallyLossSteal: (Critter*) c Amount: (ResQuant) b 
{ lossSteal += b; return self; }
-tallyLossStarve: (Critter*) c Amount: (ResQuant) b
{ lossStarve += b; return self; }

-printv
{
    printLogD("  growth = %f, ", growth);
    printLogD("  lossStarve = %f, ", lossStarve);
    printLogD("  lossSteal = %f\n", lossSteal);
    printLogD("  netGrowth = %f, ", netGrowth);
    printLogD("  netRatio = %f\n", netRatio);
    return self;
}
@end

@implementation BiomassTransfers
+create
{
    BiomassTransfers* obj = [BiomassTransfers createBegin: globalZone];
    [obj clear];
    obj->biomass = 0;
    obj->lastNetGrowth = 0;
    return obj;
}

-calcNet: (ResQuant) b 
{
    double statcycle = [geckoControl getStatCycle];
    
    biomass     = b;
    growth     /= statcycle;
    lossSteal  /= statcycle;
    lossStarve /= statcycle;
    
    losses = lossSteal + lossStarve;
    netGrowth = growth - losses;
    netRatio = netGrowth - lastNetGrowth;   
    return self;
}

-clear
{
    [super clear];
    biomass = 0;
    return self;
}

-printv
{
    [super printv];
    printLogD("  biomass = %f, ", biomass);
    return self;
}

@end


// i/o manager -- not sure how to do this Obj-Cwise, but need it Now.
char printBuffer[255];

void printLogD(char* fmt, double D)
{
    FILE* logFile = [geckoControl getLogFile];
    printf(fmt, D); 
    fprintf(logFile, fmt, D); 
}

void printArchiveD(char* fmt, double D)
{
    FILE* archiveFile = [geckoControl getArchiveFile];
    printf(fmt, D); 
    fprintf(archiveFile, fmt, D); 
}

void printBothD(char* fmt, double D)
{
    FILE* logFile = [geckoControl getLogFile];
    FILE* archiveFile = [geckoControl getArchiveFile];
    printf(fmt, D); 
    fprintf(logFile, fmt, D); 
    fprintf(archiveFile, fmt, D); 
}

void printLog(fmt, a, b, c, d, e, f, g, h)
char *fmt, *a, *b, *c, *d, *e, *f, *g, *h;
{
    FILE* logFile = [geckoControl getLogFile];
    sprintf(printBuffer, fmt, a, b, c, d, e, f, g, h); 
    printf(printBuffer); 
    fprintf(logFile, printBuffer); 
}

void printArchive(fmt, a, b, c, d, e, f, g, h)
char *fmt, *a, *b, *c, *d, *e, *f, *g, *h;
{
    FILE* archiveFile = [geckoControl getArchiveFile];
    printf(fmt, a, b, c, d, e, f, g, h); 
    fprintf(archiveFile, fmt, a, b, c, d, e, f, g, h); 
}

void printBoth(fmt, a, b, c, d, e, f, g, h)
char *fmt, *a, *b, *c, *d, *e, *f, *g, *h;
{
    FILE* logFile = [geckoControl getLogFile];
    FILE* archiveFile = [geckoControl getArchiveFile];
    printf(fmt, a, b, c, d, e, f, g, h); 
    fprintf(logFile, fmt, a, b, c, d, e, f, g, h); 
    fprintf(archiveFile, fmt, a, b, c, d, e, f, g, h); 
}

void flushBoth()
{
    FILE* logFile = [geckoControl getLogFile];
    FILE* archiveFile = [geckoControl getArchiveFile];
    fflush(logFile);
    fflush(archiveFile);
}

