#include <sys/time.h>
#include <values.h>
#include <stdlib.h>
#include <math.h>
/* #include "types.h" */
#include "random.h"

/*
 *  History:
 *     Ported from C++ implementation for Popsim, 940719, vfb.
 *     Replaced original Echo random functions, which didn't appear to be
 *     working, and using mine seemed simpler than figuring out his.
 *     
 * // -------------------------------------------------------------------------
 * // Random number functions
 * // -------------------------------------------------------------------------
 * // These functions are based on NUMERICAL RECIPES IN C:  The Art of Scientific
 * // Computing, by William Press, et al, Cambridge University Press, 1988.
 * // Please see Chapter 7 for discussion.  Note that the Sun's and HP7xx's 
 * // algorithm for producing pseudo-random numbers in the rand48 family is
 * // documented as a linear congruential generator of the same type
 * // described in section 7.1 of the book, using 48-bit integer arithmetic.
 * // -------------------------------------------------------------------------
 */

/* "The exact number 98 is unimportant." */
#define	RandTableSize    98 
#define MaxRand          (1.0 + (double) MAXLONG)
double	LastRandD = -1;
double 	RandTable[RandTableSize];


long
InitRand(RandomSeed)
long RandomSeed;
{
    long junk;
    int	 j;

    /* get seed if don't have one yet */
    if (RandomSeed <= 0){
	extern int gettimeofday();

	struct timeval tp;
	if (gettimeofday(&tp, (struct timezone *)0) == -1){
	    printf("Could not gettimeofday in InitRand().");
	}
	
	RandomSeed = tp.tv_sec;
    }
    
    /* initialize standard library */
    srand48(RandomSeed);
    
    /* initialize our routines */		
    for (j=0; j < RandTableSize; j++)	/* just exercise sysroutine */
        junk = lrand48();		
    for (j=0; j < RandTableSize; j++)	/* now save values in table */
	RandTable[j] = lrand48();
    LastRandD = lrand48();	/* and one more */

    return(RandomSeed);
};

double EvenRand()
/* uniformly distributed over [0..1] */
{
    int	j;
    double result;

    /*    
    // Use saved random number LastRandD to get an index j between 1 and
    // RandTableSize - 1.  Then use RandTable[j] for both the next j and 
    // for the output random number.
    */
    	    
    j = (int) ((1 + (1.0 * (RandTableSize - 1)) * LastRandD) / MaxRand);
    LastRandD = RandTable[j];

    /* fill up hole */
    RandTable[j] = lrand48();	
    result = (LastRandD / MaxRand);
    return(result);
};


double GaussRand()
/* Normal distrib, mean=0., variance=1. */
{
    static long	spare_f = 0;	/* flag--have a value already? */
    static double spare;	/* saved value */
    double radius2, r1, r2, factor;		

    /* don't have a value handy, so generate... */
    if (spare_f == 0) {	

        /*
	// pick two uniformly distrib numbers in square extending
	// from -1 to + 1 in each direction, check if they're inside
	// the unit circle, and if not, pick again.
	*/

	do {
	    r1 = 2.0 * EvenRand() - 1.0;
	    r2 = 2.0 * EvenRand() - 1.0;
	    radius2 = r1*r1 + r2*r2;				
	} while (radius2 >= 1.0);
	
	factor = sqrt(-2.0*log(radius2)/radius2);

	/*
	// Make Box-Muller transformation to get two normally
	// distributed numbers.  Return one, save other for next time.
	*/

	spare = r1 * factor;
	spare_f = 1;
	return r2 * factor;
    } else {
    	/* do have a value handy, so use it. */
	spare_f = 0;
	return spare;
    }
}; 

double ExpRand()
/* Exponential distrib, mean = 1.0 */
{
    float dum;
    do
        dum = EvenRand();
    while (dum == 0.0);
    return -log(dum);
}
